#include <iostream>
#include <cassert>
#include <cstring>
#include <thread>
#include <chrono>
#include "mpi.h"

#include "rdma/fabric.h"
#include "rdma/fi_domain.h"
#include "rdma/fi_cm.h"
#include "rdma/fi_errno.h"
#include "rdma/fi_endpoint.h"


int main(int argc, char* argv[]) {
	MPI_Init(&argc, &argv);

	int rank, size;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	std::cout << "rank: " << rank << std::endl;


	fi_info* hints = fi_allocinfo();
	fi_info* info;
	hints->mode = FI_CONTEXT | FI_LOCAL_MR;
	hints->caps = FI_TAGGED | FI_MSG | FI_MULTI_RECV | FI_RMA;
	hints->domain_attr->threading = FI_THREAD_ENDPOINT;
	hints->domain_attr->control_progress = FI_PROGRESS_AUTO;
	hints->domain_attr->data_progress = FI_PROGRESS_AUTO;
	hints->domain_attr->resource_mgmt = FI_RM_ENABLED;
	hints->domain_attr->mr_mode = FI_MR_SCALABLE;
	hints->tx_attr->op_flags = FI_DELIVERY_COMPLETE | FI_COMPLETION;

	int res;
	res = fi_getinfo(FI_VERSION(1, 3), nullptr, nullptr, 0, hints, &info);
	assert(res == 0);

	// just take the first
	fi_info* prov = info->next->next;
	
	// open fabric
	fid_fabric* fabric;
	res = fi_fabric(prov->fabric_attr, &fabric, nullptr);
	assert(res == 0);

	// open domain
	fid_domain* domain;
	res = fi_domain(fabric, prov, &domain, nullptr);
	assert(res == 0);

	// open completion queue
	fi_cq_attr cq_attr;
	memset(&cq_attr, 0, sizeof(cq_attr));
	cq_attr.wait_obj = FI_WAIT_NONE;
	cq_attr.format = FI_CQ_FORMAT_CONTEXT;
	cq_attr.size = prov->tx_attr->size;
	fid_cq* completion_queue;
	res = fi_cq_open(domain, &cq_attr, &completion_queue, nullptr);
	assert(res == 0);

	fid_ep* endpoint;
	res = fi_endpoint(domain, prov, &endpoint, nullptr);

	fi_av_attr av_attr;
	memset(&av_attr, 0, sizeof(av_attr));
	av_attr.type = FI_AV_MAP;
	av_attr.count = 3;
	fid_av* av;
	res = fi_av_open(domain, &av_attr, &av, nullptr);
	assert(res == 0);

	res = fi_ep_bind(endpoint, reinterpret_cast<fid*>(av), 0);
	assert(res == 0);

	res = fi_ep_bind(endpoint, reinterpret_cast<fid*>(completion_queue), FI_TRANSMIT | FI_RECV);
	assert(res == 0);

	res = fi_enable(endpoint);

	assert(res == 0);
	char addrname[50];
	size_t addrnamelen = 50;
	res = fi_getname(reinterpret_cast<fid*>(endpoint), addrname, &addrnamelen);
	assert(res == 0);

	char* alladdrnames = (char*) malloc(16 * size);
	res = MPI_Allgather(addrname, 16, MPI_BYTE, alladdrnames, 16, MPI_BYTE, MPI_COMM_WORLD);

	fi_addr_t* addr = new fi_addr_t[size];
	res = fi_av_insert(av, alladdrnames, size, addr, 0, nullptr);
	assert(res == size);

	free(alladdrnames);

	if (rank < 2) {
		int buf = rank*10 + 123;
		res = fi_send(endpoint, &buf, sizeof(buf), nullptr, addr[2], nullptr);
		assert(res == 0);

		do {
			fi_cq_entry entry;
			res = fi_cq_read(completion_queue, &entry, 1);
		} while (res == -FI_EAGAIN);
		assert(res > 0);
	} else {
		int buf1 = 0;
		int buf2 = 0;
		res = fi_recv(endpoint, &buf1, sizeof(buf1), nullptr, addr[0], nullptr);
		assert(res == 0);
		res = fi_recv(endpoint, &buf2, sizeof(buf2), nullptr, addr[1], nullptr);
		assert(res == 0);

		do {
			fi_cq_entry entry;
			res = fi_cq_read(completion_queue, &entry, 1);
		} while (res == -FI_EAGAIN);
		do {
			fi_cq_entry entry;
			res = fi_cq_read(completion_queue, &entry, 1);
		} while (res == -FI_EAGAIN);
		assert(res > 0);
		std::cout << buf1 << " " << buf2 << std::endl;
	}
	MPI_Barrier(MPI_COMM_WORLD);

	res = fi_close(reinterpret_cast<fid*>(endpoint));
	assert(res == 0);

	res = fi_close(reinterpret_cast<fid*>(av));
	assert(res == 0);

	res = fi_close(reinterpret_cast<fid*>(completion_queue));
	assert(res == 0);

	res = fi_close(reinterpret_cast<fid*>(domain));
	assert(res == 0);

	res = fi_close(reinterpret_cast<fid*>(fabric));
	assert(res == 0);

	fi_freeinfo(hints);
	fi_freeinfo(info);

	MPI_Finalize();
}
