#include <infiniband/verbs.h>
#include "mpi.h"
#include <cstring>
#include <cmath>

#include <sys/unistd.h>

#include "utility/error.h"

#include "../tmpi/include/cache_flush.h"

constexpr double freq = 2.7; //GHz
using namespace tinympi;

int main(int argc, char* argv[]) {

	char hostname[128];
	gethostname(hostname, sizeof(hostname));
	std::cout << hostname << std::endl;

	CacheFlush cf(CACHESIZE);

	MPI_Init(&argc, &argv);

	int res;
	int numprocs, rank;
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	Debugger::of.open(std::string("out") + std::to_string(rank));

	MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
	TMPI_ASSERT_FATAL(numprocs == 2, "too many processors", __FILE__, __LINE__);
	TMPI_LOG("numprocs", numprocs);
	TMPI_LOG("rank", rank);


	// open device list
	int num_devices;
	ibv_device** dev_list = ibv_get_device_list(&num_devices);
	TMPI_ASSERT_FATAL(dev_list != nullptr, "ibv_get_device_list() returned 0", __FILE__, __LINE__);
	TMPI_ASSERT_FATAL(num_devices > 0, "ibv_get_device_list() did not find anz device!", __FILE__,
			__LINE__);

	// use first device and open
	ibv_device* device = dev_list[0];
	TMPI_ASSERT_FATAL(dev_list != nullptr, "detected device does not exist!", __FILE__, __LINE__);
	ibv_context* context = ibv_open_device(device);
	TMPI_ASSERT_FATAL(context != nullptr, "ibv_open_device() returned 0", __FILE__, __LINE__);

	// close device list
	ibv_free_device_list(dev_list);

	// get device name and print to log
	const char* dev_name = ibv_get_device_name(device);
	TMPI_ASSERT_FATAL(dev_name != nullptr, "couldn't get device name", __FILE__, __LINE__);
	TMPI_LOG("device name: ", dev_name);

	// allocate protection domain (for AH, QP, MR, MW and SRQ)
	ibv_pd* pd = ibv_alloc_pd(context);
	TMPI_ASSERT_FATAL(pd != nullptr, "couldn't create protection domain", __FILE__,
			__LINE__);

//////////////////////////////


	ibv_port_attr port_attr;
	res = ibv_query_port(context, 1, &port_attr);
	TMPI_ASSERT_FATAL(res == 0, "ibv_query_port() returned " << res, __FILE__, __LINE__);

	ibv_cq* cq = ibv_create_cq(context, 1000, nullptr, nullptr, 0);
	TMPI_ASSERT_FATAL(cq != nullptr, "could not create completion queue", __FILE__, __LINE__);
	ibv_qp_init_attr qp_init_attr = {};
	qp_init_attr.qp_context = context;
	qp_init_attr.send_cq = cq;
	qp_init_attr.recv_cq = cq;
	qp_init_attr.cap.max_send_wr = 10000;
	qp_init_attr.cap.max_recv_wr = 10000;
	qp_init_attr.cap.max_send_sge = 1;
	qp_init_attr.cap.max_recv_sge = 1;
	qp_init_attr.cap.max_inline_data = 256;
	qp_init_attr.qp_type = IBV_QPT_RC;
	qp_init_attr.sq_sig_all = 1;

	ibv_qp* qp = ibv_create_qp(pd, &qp_init_attr);
	TMPI_ASSERT_FATAL(qp != nullptr, "could not create qp", __FILE__, __LINE__);
	uint32_t qp_num = qp->qp_num;
	uint16_t lid = port_attr.lid;


////// regmr
	constexpr unsigned N = 512*1024;
	int* data = new int[N];
	ibv_mr* data_mr = ibv_reg_mr(pd, data, N * sizeof(int),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(data_mr != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);
	uint32_t rkey = data_mr->rkey;
	uint64_t addr = reinterpret_cast<uint64_t>(data);

	TMPI_LOG("ibv_reg_mr", "done");

	int dest = 1 - rank;
	MPI_Request req[8];
	MPI_Status status[8];
	uint16_t recv_lid;
	uint32_t recv_qpnum;
	uint32_t recv_rkey;
	uint64_t recv_addr;
	TMPI_LOG("MPI_Send", lid << " " << qp_num << " " << rkey << " " << addr);
	MPI_Isend(&lid   , 1, MPI_SHORT, dest, 0, MPI_COMM_WORLD, &req[0]);
	MPI_Isend(&qp_num, 1, MPI_INT  , dest, 0, MPI_COMM_WORLD, &req[1]);
	MPI_Isend(&rkey  , 1, MPI_INT  , dest, 0, MPI_COMM_WORLD, &req[2]);
	MPI_Isend(&addr  , 1, MPI_LONG , dest, 0, MPI_COMM_WORLD, &req[3]);
	MPI_Irecv(&recv_lid  , 1, MPI_SHORT, dest, 0, MPI_COMM_WORLD, &req[4]);
	MPI_Irecv(&recv_qpnum, 1, MPI_INT  , dest, 0, MPI_COMM_WORLD, &req[5]);
	MPI_Irecv(&recv_rkey , 1, MPI_INT  , dest, 0, MPI_COMM_WORLD, &req[6]);
	MPI_Irecv(&recv_addr , 1, MPI_LONG , dest, 0, MPI_COMM_WORLD, &req[7]);
	
	MPI_Waitall(8, req, status);
	TMPI_LOG("MPI_Recv", recv_lid << " " << recv_qpnum << " " << recv_rkey << " " << recv_addr);



//
	// to init state
	ibv_qp_attr qp_attr_to_init = {};
	qp_attr_to_init.qp_state = IBV_QPS_INIT;
	qp_attr_to_init.pkey_index = 0;
	qp_attr_to_init.qp_access_flags = IBV_ACCESS_REMOTE_READ | IBV_ACCESS_REMOTE_WRITE |
			IBV_ACCESS_REMOTE_ATOMIC;
	qp_attr_to_init.port_num = 1;
	res = ibv_modify_qp(qp, &qp_attr_to_init,
			IBV_QP_STATE | IBV_QP_PKEY_INDEX | IBV_QP_ACCESS_FLAGS | IBV_QP_PORT);
	TMPI_ASSERT_FATAL(res == 0, "ibv_modify_qp() returned " << res, __FILE__, __LINE__);

	// to rtr state
	ibv_qp_attr qp_attr_to_rtr = {};
	qp_attr_to_rtr.qp_state = IBV_QPS_RTR;
	qp_attr_to_rtr.ah_attr.dlid = recv_lid;
	qp_attr_to_rtr.ah_attr.sl = 0;
	qp_attr_to_rtr.ah_attr.src_path_bits = 0;
	qp_attr_to_rtr.ah_attr.static_rate = 0;
	qp_attr_to_rtr.ah_attr.is_global = 0;
	qp_attr_to_rtr.ah_attr.port_num = 1;
	qp_attr_to_rtr.rq_psn = 0;
	qp_attr_to_rtr.max_dest_rd_atomic = 4;
	qp_attr_to_rtr.min_rnr_timer = 3;
	qp_attr_to_rtr.dest_qp_num = recv_qpnum;
	qp_attr_to_rtr.path_mtu = IBV_MTU_2048;
	res = ibv_modify_qp(qp, &qp_attr_to_rtr,
		IBV_QP_STATE | IBV_QP_AV | IBV_QP_RQ_PSN | IBV_QP_MAX_DEST_RD_ATOMIC
		| IBV_QP_MIN_RNR_TIMER | IBV_QP_DEST_QPN | IBV_QP_PATH_MTU);
	TMPI_ASSERT_FATAL(res == 0, "ibv_modify_qp() returned " << res, __FILE__, __LINE__);

	ibv_qp_attr qp_attr_to_rts = {};
	qp_attr_to_rts.qp_state = IBV_QPS_RTS;
	qp_attr_to_rts.sq_psn = 0;
	qp_attr_to_rts.timeout = 1;
	qp_attr_to_rts.retry_cnt = 7;
	qp_attr_to_rts.rnr_retry = 7;
	qp_attr_to_rts.max_rd_atomic = 4;
	res = ibv_modify_qp(qp, &qp_attr_to_rts,
		IBV_QP_STATE | IBV_QP_SQ_PSN | IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT
		| IBV_QP_RNR_RETRY | IBV_QP_MAX_QP_RD_ATOMIC);
	TMPI_ASSERT_FATAL(res == 0, "ibv_modify_qp() returned " << res, __FILE__, __LINE__);

	TMPI_LOG("ibv_modify_qp", "done");

///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////


	auto send_function = [&](int sz) {
		// now fill the data to be sent
		ibv_sge sr_sg_lst = {
			reinterpret_cast<uint64_t>(data),
			sz * sizeof(int),
			data_mr->lkey
		};

		// prepare eager send request
		ibv_send_wr sr_wr;
		sr_wr.wr_id = 1;
		sr_wr.opcode = IBV_WR_RDMA_WRITE_WITH_IMM; // we write directly into rdma buffer
		sr_wr.imm_data = 1234;
		if (sz * sizeof(int) <= 256)
			sr_wr.send_flags = IBV_SEND_SIGNALED | IBV_SEND_INLINE;
		else
			sr_wr.send_flags = IBV_SEND_SIGNALED;

		sr_wr.sg_list = &sr_sg_lst;
		sr_wr.num_sge = 1; // only one message
		sr_wr.next = nullptr;

		// set remote data with peer info retrieved from channel
		sr_wr.wr.rdma.rkey = recv_rkey;
		sr_wr.wr.rdma.remote_addr = recv_addr;
		TMPI_LOG("ibv_post_send", "send to " << recv_rkey << " " << recv_addr << " from "
			<< sr_wr.sg_list->addr << " with size " << sr_wr.sg_list->length);
		// now we can submit the message to the channel
		ibv_send_wr* bad_wr;
		res = ibv_post_send(qp, &sr_wr, &bad_wr);
		TMPI_ASSERT_FATAL(res == 0, "ibv_post_send returned " << res, __FILE__, __LINE__);

		while (true) {
			ibv_wc wc;
			res = ibv_poll_cq(cq, 1, &wc);
			TMPI_ASSERT_FATAL(res != -1, "ibv_poll_cq() returned " << res, __FILE__, __LINE__);

			if (res > 0) {
				TMPI_ASSERT_FATAL(wc.status == IBV_WC_SUCCESS, "wc.wr_id "
						<< reinterpret_cast<void*>(wc.wr_id) << " failed with status \""
						<< ibv_wc_status_str(wc.status) << "\" (" << wc.status << ")"
						, __FILE__, __LINE__);
				TMPI_ASSERT_FATAL(wc.wr_id == 1, "wrong wr_id! " << wc.wr_id << " != 1", __FILE__, __LINE__);

				break;

			}
		}
	};
	auto recv_function = [&]() {
		ibv_recv_wr dummy_recv_ = {};
		ibv_recv_wr* dummy_bad_recv_;
		std::memset(&dummy_recv_, 0, sizeof(ibv_recv_wr));
		std::memset(&dummy_bad_recv_, 0, sizeof(ibv_recv_wr));
		dummy_recv_.wr_id = 123;
		res = ibv_post_recv(qp, &dummy_recv_, &dummy_bad_recv_);
		TMPI_ASSERT_FATAL(res == 0, "ibv_post_recv() returned " << res, __FILE__, __LINE__);


		while (true) {
			ibv_wc wc;

			res = ibv_poll_cq(cq, 1, &wc);
			TMPI_ASSERT_FATAL(res != -1, "ibv_poll_cq() returned " << res, __FILE__, __LINE__);

			if (res > 0) {
				TMPI_ASSERT_FATAL(wc.status == IBV_WC_SUCCESS, "wc.wr_id "
						<< reinterpret_cast<void*>(wc.wr_id) << " failed with status \""
						<< ibv_wc_status_str(wc.status) << "\" (" << wc.status << ")"
						, __FILE__, __LINE__);
				TMPI_ASSERT_FATAL(wc.wr_id == 123, "wrong wr_id! " << wc.wr_id << " != 123", __FILE__, __LINE__);

				break;

			}
		}
	};

	std::vector<unsigned> sizes;
	for (unsigned i = 1; i < 512*1024; i *= 2) {
		double k = i;
// 		for (int j = 0; j < 4; ++j) {
			sizes.push_back(k);
// 			k *= 1.189207115;
// 		}
	}
	for (auto length : sizes) {
		cf.flush(); // clear cache

		for (int i = 0; i < 500; ++i) {
			if (rank == 0) {
				send_function(length);
				recv_function();

				send_function(length);
				recv_function();

			} else {
				recv_function();
				send_function(length);

				recv_function();
				send_function(length);
			}
		}
		for (int i = 0; i < 300; ++i) {
			if (rank == 0) {
				send_function(length); //1
				recv_function();
			} else {
				recv_function();
				send_function(length); //6
			}
		}
	}


///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////

	res = ibv_destroy_qp(qp);
	TMPI_ASSERT_FATAL(res == 0, "ibv_destroy_qp returned " << res, __FILE__, __LINE__);

	res = ibv_destroy_cq(cq);
	TMPI_ASSERT_FATAL(res == 0, "ibv_destroy_cq returned " << res, __FILE__, __LINE__);

	res = ibv_dereg_mr(data_mr);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);

	res = ibv_dealloc_pd(pd);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dealloc_pd returned " << res, __FILE__, __LINE__);

	res = ibv_close_device(context);
	TMPI_ASSERT_FATAL(res == 0, "ibv_close_device returned " << res, __FILE__, __LINE__);

	TMPI_LOG("ibv_close_device", "done");

	delete[] data;
	
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize();

	TMPI_LOG("MPI_Finalize", "done");


	return 0;

}

namespace tinympi {
	std::ofstream Debugger::of { };
	std::string Debugger::color = "";
}
