/// \file
/// Performance timer functions.
#ifndef __PERFORMANCE_TIMERS_H_
#define __PERFORMANCE_TIMERS_H_

#include <stdio.h>
#include <stdint.h>

/// Timer handles
enum TimerHandle{
   totalTimer,
   loopTimer,
   timestepTimer,
   positionTimer,
   velocityTimer,
   redistributeTimer,
   atomHaloTimer,
   computeForceTimer,
   eamHaloTimer,
   commHaloTimer,
   commReduceTimer,
   externalTimer,
   numberOfTimers};

/// Use the startTimer and stopTimer macros for timers in code regions
/// that may be performance sensitive.  These can be compiled away by
/// defining NTIMING.  If you are placing a timer anywere outside of a
/// tight loop, consider calling profileStart and profileStop instead.
///
/// Place calls as follows to collect time for code pieces.
/// Time is collected everytime this portion of code is executed. 
///
///     ...
///     startTimer(computeForceTimer);
///     computeForce(sim);
///     stopTimer(computeForceTimer);
///     ...
///
#ifndef NTIMING
#define startTimer2(handle)    \
   do                         \
{                          \
   profileStart(handle, &myPerfTimers[tinympi::tasks::LocalTask::get_current_task()->get_identifier().local.task_id*numberOfTimers]); \
} while(0)
#define stopTimer2(handle)     \
   do                         \
{                          \
   profileStop(handle, &myPerfTimers[tinympi::tasks::LocalTask::get_current_task()->get_identifier().local.task_id*numberOfTimers]); \
} while(0)
#define startTimer(handle)    \
   do                         \
{                          \
   profileStart(handle, perfTimer);   \
} while(0)
#define stopTimer(handle)     \
   do                         \
{                          \
   profileStop(handle, perfTimer);    \
} while(0)
#else
#define startTimer2(handle)
#define stopTimer2(handle)
#define startTimer(handle)
#define stopTimer(handle)
#endif

struct Timers
{
   uint64_t start;     //!< call start time
   uint64_t total;     //!< current total time
   uint64_t count;     //!< current call count
   uint64_t elapsed;   //!< lap time
 
   int minRank;        //!< rank with min value
   int maxRank;        //!< rank with max value

   double minValue;    //!< min over ranks
   double maxValue;    //!< max over ranks
   double average;     //!< average over ranks
   double stdev;       //!< stdev across ranks
} ;

/// Global timing data collected.
struct TimerGlobal
{
   double atomRate;       //!< average time (us) per atom per rank 
   double atomAllRate;    //!< average time (us) per atom
   double atomsPerUSec;   //!< average atoms per time (us)
} ;

extern thread_local Timers* myPerfTimers;

/// Use profileStart and profileStop only for timers that should *never*
/// be turned off.  Typically this means they are outside the main
/// simulation loop.  If the timer is inside the main loop use
/// startTimer and stopTimer instead.
void profileStart(const enum TimerHandle handle, Timers* perfTimer);
void profileStop(const enum TimerHandle handle, Timers* perfTimer);

/// Use to get elapsed time (lap timer).
double getElapsedTime(const enum TimerHandle handle, Timers* perfTimer);

/// Print timing results.
void printPerformanceResults(int nGlobalAtoms, int printRate, TimerGlobal& perfGlobal, Timers* perfTimer);

/// Print timing results to Yaml file
void printPerformanceResultsYaml(FILE* file, TimerGlobal& perfGlobal, Timers* perfTimer);

#endif
