// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _COLLECTIVES_H_635106f3226c46d3d114158785752466
#define _COLLECTIVES_H_635106f3226c46d3d114158785752466

#include <functional>

#include "communicator.h"
#include "debug/debugger.h"
#include "pt2pt.h"

namespace tinympi {
namespace collectives {
	inline int bcast(void* buffer, int length, rank_type root, const Communicator* comm);

	using reduction_function_t = std::function<void(void*, const void*)>;

	template<typename T>
	inline int allreduce(const void* sendbuf, void* recvbuf, int length, reduction_function_t op,
			const Communicator* comm);

	inline reduction_function_t get_reduction_function(TMPI_Op op, TMPI_Datatype type, int count);

	inline int barrier(const Communicator* comm);

//////
}
}

// include implementation files
#include "collectives_impl/barrier.h"
#include "collectives_impl/allreduce.h"
#include "collectives_impl/bcast.h"


namespace tinympi {
namespace collectives {

inline int bcast(void* buffer, int length, rank_type root, const Communicator* comm) {
	TMPI_LOG_LEVEL(-1, "Collectives", "bcast....");
	using namespace impl::bcast;

	if (comm->get_size() == 1)
		return TMPI_SUCCESS;

	return bcast_binomial(buffer, length, root, comm);
}

inline int allreduce(void* sendbuf, void* recvbuf, int length, reduction_function_t op,
		const Communicator* comm) {
	TMPI_LOG_LEVEL(-1, "Collectives", "allreduce ....");
	using namespace impl::allreduce;

	// if the communicator is only this task, just copy to the recv buffer
	if (comm->get_size() == 1) {
		memcpy(recvbuf, sendbuf, length);
		return TMPI_SUCCESS;
	}

	return allreduce_impl(sendbuf, recvbuf, length, op, comm);
}
inline reduction_function_t get_reduction_function(TMPI_Op op, TMPI_Datatype type, int count) {
	using namespace impl::allreduce;

	switch (op) {
	case TMPI_MAX:
		return operations::datatype_dispatch<operations::max>(type, count);
	case TMPI_MIN:
		return operations::datatype_dispatch<operations::min>(type, count);
	case TMPI_SUM:
		return operations::datatype_dispatch<operations::sum>(type, count);
	case TMPI_PROD:
		return operations::datatype_dispatch<operations::prod>(type, count);
	case TMPI_BAND:
		return operations::datatype_dispatch<operations::band>(type, count);
	case TMPI_LAND:
		return operations::datatype_dispatch<operations::land>(type, count);
	case TMPI_BOR:
		return operations::datatype_dispatch<operations::bor>(type, count);
	case TMPI_LOR:
		return operations::datatype_dispatch<operations::lor>(type, count);
	case TMPI_BXOR:
		return operations::datatype_dispatch<operations::bxor>(type, count);
	case TMPI_LXOR:
		return operations::datatype_dispatch<operations::lxor>(type, count);
	case TMPI_MINLOC:
		return operations::datatype_dispatch<operations::minloc>(type, count);
	case TMPI_MAXLOC:
		return operations::datatype_dispatch<operations::maxloc>(type, count);
	default:
		TMPI_FATAL_ERROR("Undefined operation", __FILE__, __LINE__);
		return [](void*,const void*){};
	}
}

inline int barrier(const Communicator* comm) {
	using namespace impl::barrier;
	TMPI_LOG_LEVEL(-1, "Collectives", "barrier....");

	int res = barrier_impl(comm);
	TMPI_LOG_LEVEL(-1, "Collectives", "...done");
	return res;
}

} // collectives
} // tinympi

#endif
