// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#include <functional>
#include <cstring>
#include <type_traits>

#include "global.h"
#include "pt2pt.h"
#include "debug/debugger.h"

namespace tinympi {
namespace collectives {
namespace impl {
	using namespace tinympi::impl;

/////////////////////////////////////
// ALLREDUCE
namespace allreduce {
inline int allreduce_impl(void* sendbuf, void* recvbuf, int length,
		reduction_function_t op, const Communicator* comm) {
	int res;

	rank_type my_rank = comm->get_rank();
	rank_type size = comm->get_size();

	// we assume associativity

	// allocate temporary buffer where we recv temporary data
	void* tmpbuf = malloc(length);
	// copy send buffer to recv buffer where we accumulate the results
	memcpy(recvbuf, sendbuf, length);
	TMPI_LOG_LEVEL(-1, "Collectives", "our send buffer: " << ((int*)sendbuf)[0]);

	// get max_n 2^n <= numprocs
	rank_type newsize = 1;
	while (newsize <= size) newsize <<= 1;
	newsize >>= 1;

	rank_type rem = size - newsize;

	// reduce data to 2^n procs
	int rel_rank;
	if (my_rank < 2*rem) {

		// if we have more than 2^n procs, we exchange data between certain two processors
		if (my_rank % 2 == 0) {
			rel_rank = my_rank / 2;

			rank_type src = my_rank + 1;

			TMPI_LOG_LEVEL(-1, "Collectives", "recv from " << src);
			res = pt2pt::recv(tmpbuf, length, src, TMPI_ALLREDUCE_TAG, comm->get_id(), comm,
					TMPI_STATUS_IGNORE);
			TMPI_ASSERT_FATAL(res == TMPI_SUCCESS, "recv() failed", __FILE__, __LINE__);

			// apply reduction
			op(recvbuf, tmpbuf);

		} else {
			rel_rank = -1;

			rank_type dest = my_rank - 1;

			TMPI_LOG_LEVEL(-1, "Collectives", "send to " << dest);
			res = pt2pt::send(sendbuf, length, dest, TMPI_ALLREDUCE_TAG, comm->get_id(), comm);
			TMPI_ASSERT_FATAL(res == TMPI_SUCCESS, "send() failed", __FILE__, __LINE__);
		}
	} else {
		rel_rank = my_rank - rem;
	}
	TMPI_LOG_LEVEL(-1, "Collectives", "rel_rank = " << rel_rank);

	if (rel_rank != -1) {
		// now we have 2^n procs that do the reduction.
		int mask = 1;
		while (mask < newsize) {
			int dest = rel_rank ^ mask;
			dest = (dest < rem) ? dest * 2 : dest + rem;

			TMPI_LOG_LEVEL(-1, "Collectives", "sendrecv with " << dest);
			res = pt2pt::sendrecv(recvbuf, length, dest, TMPI_ALLREDUCE_TAG, tmpbuf, length, dest,
					TMPI_ALLREDUCE_TAG, comm->get_id(), comm, TMPI_STATUS_IGNORE);
			TMPI_ASSERT_FATAL(res == TMPI_SUCCESS, "sendrecv() failed", __FILE__, __LINE__);
			TMPI_LOG_LEVEL(-1, "Collectives", "sendbuf = " << ((int*)sendbuf)[0]
					<< "; tmpbuf = " << ((int*)tmpbuf)[0] << "; recvbuf = "
					<< ((int*)recvbuf)[0]);
			op(recvbuf, tmpbuf);

			mask <<= 1;

		}

	}

	if (my_rank < 2*rem) {
		// the remaining processes need to exchange the results
		if (my_rank % 2 == 0) {
			rank_type dest = my_rank + 1;

			TMPI_LOG_LEVEL(-1, "Collectives", "send to " << dest);
			res = pt2pt::send(recvbuf, length, dest, TMPI_ALLREDUCE_TAG, comm->get_id(),
					comm);
			TMPI_ASSERT_FATAL(res == TMPI_SUCCESS, "send() failed", __FILE__, __LINE__);

		} else {
			rank_type src = my_rank - 1;

			TMPI_LOG_LEVEL(-1, "Collectives", "recv from " << src);
			res = pt2pt::recv(recvbuf, length, src, TMPI_ALLREDUCE_TAG, comm->get_id(), comm,
					TMPI_STATUS_IGNORE);
			TMPI_ASSERT_FATAL(res == TMPI_SUCCESS, "recv() failed", __FILE__, __LINE__);
		}
	}

	free(tmpbuf);

	return TMPI_SUCCESS;

}
namespace operations {
	// in this namespace we define all needed operators

	// takes an element operator and a type, and applies this operator to count elements
	template<typename T, template<typename> class element_op>
	inline void apply_elementwise(void* recv, const void* send, int count) {
		T* recv_ = reinterpret_cast<T*>(recv);
		const T* send_ = reinterpret_cast<const T*>(send);

		for (int i = 0; i < count; ++i) {
			element_op<T>::Do(recv_[i], send_[i]);
		}
	}

	// create an reduce operation function from an element operator by using apply_elementwise
	template<typename T, template<typename> class element_op>
	inline reduction_function_t get_op_function(int count) {
		return [count](void* recv, const void* send) {
			apply_elementwise<T, element_op>(recv, send, count);
		};
	}

	// depending on the datatype >type< we create a reduce operation function
	template<template<typename> class element_op>
	inline reduction_function_t datatype_dispatch(TMPI_Datatype type, int count) {
		switch (type) {
		case TMPI_CHAR:
			return get_op_function<char, element_op>(count);
		case TMPI_SHORT:
			return get_op_function<short, element_op>(count);
		case TMPI_INT:
			return get_op_function<int, element_op>(count);
		case TMPI_LONG:
			return get_op_function<long, element_op>(count);
		case TMPI_UNSIGNED_CHAR:
			return get_op_function<unsigned char, element_op>(count);
		case TMPI_UNSIGNED_SHORT:
			return get_op_function<unsigned short, element_op>(count);
		case TMPI_UNSIGNED:
			return get_op_function<unsigned, element_op>(count);
		case TMPI_UNSIGNED_LONG:
			return get_op_function<unsigned long, element_op>(count);
		case TMPI_FLOAT:
			return get_op_function<float, element_op>(count);
		case TMPI_DOUBLE:
			return get_op_function<double, element_op>(count);
		case TMPI_LONG_DOUBLE:
			return get_op_function<long double, element_op>(count);
		case TMPI_BYTE:
			return get_op_function<uint8_t*, element_op>(count);

		case TMPI_FLOAT_INT:
			return get_op_function<pair_type<float, int>, element_op>(count);
		case TMPI_DOUBLE_INT:
			return get_op_function<pair_type<double, int>, element_op>(count);
		case TMPI_LONG_INT:
			return get_op_function<pair_type<long, int>, element_op>(count);
		case TMPI_2INT:
			return get_op_function<pair_type<int, int>, element_op>(count);
		case TMPI_SHORT_INT:
			return get_op_function<pair_type<short, int>, element_op>(count);
		case TMPI_LONG_DOUBLE_INT:
			return get_op_function<pair_type<long double, int>, element_op>(count);
		default:
			TMPI_FATAL_ERROR("Undefined datatype (" << type << ")", __FILE__, __LINE__);
			return [](void*, const void*) {};
		}
	}

// NOW LETS DEFINE ALL REDUCTION FUNCTIONS WE NEED

	// standard reduction functions. For certain types of operators (COND) we create an operator
	// NAME which executes the instructions FUNC
	#define REDUCTION_FUNCTION(NAME, COND, FUNC) \
		template<typename T, bool B = \
				COND \
			> \
		struct NAME ## _impl { \
			inline static void Do(T& out, const T& in) { \
				FUNC \
			} \
		}; \
		template<typename T> \
		struct NAME ## _impl<T, false> { \
			inline static void Do(T&, const T&) { \
				TMPI_FATAL_ERROR("Operation not defined for datatype" \
						<< datatype_traits<T>::name, __FILE__, __LINE__); \
			} \
		}; \
		template<typename T> \
		struct NAME { \
			inline static void Do(T& out, const T& in) { \
				NAME ## _impl<T>::Do(out, in); \
			} \
		};

	// we may require certain classes
	#define REQUIRE_DATATYPE_INT_CLASS     datatype_traits<T>::reduction_class == reduction_class_t::datatype_int_class
	#define REQUIRE_DATATYPE_FLOAT_CLASS   datatype_traits<T>::reduction_class == reduction_class_t::datatype_float_class
	#define REQUIRE_DATATYPE_COMPLEX_CLASS datatype_traits<T>::reduction_class == reduction_class_t::datatype_complex_class
	#define REQUIRE_DATATYPE_BYTE_CLASS    datatype_traits<T>::reduction_class == reduction_class_t::datatype_byte_class
	#define REQUIRE_DATATYPE_OTHER_CLASS   datatype_traits<T>::reduction_class == reduction_class_t::datatype_other_class
	#define REQUIRE_DATATYPE_LOC_CLASS     datatype_traits<T>::reduction_class == reduction_class_t::datatype_loc_class

	// e.g., we create a max operator, which is specified for int and float classes, and executes
	// the command "if (in > out) out = in"
	REDUCTION_FUNCTION( \
			max, \
			REQUIRE_DATATYPE_INT_CLASS || REQUIRE_DATATYPE_FLOAT_CLASS, \
			if (in > out) out = in; \
	)
	REDUCTION_FUNCTION( \
			min, \
			REQUIRE_DATATYPE_INT_CLASS || REQUIRE_DATATYPE_FLOAT_CLASS, \
			if (in < out) out = in; \
	)
	REDUCTION_FUNCTION( \
			sum, \
			REQUIRE_DATATYPE_INT_CLASS || REQUIRE_DATATYPE_FLOAT_CLASS || REQUIRE_DATATYPE_COMPLEX_CLASS, \
			out += in;
	)
	REDUCTION_FUNCTION( \
			prod, \
			REQUIRE_DATATYPE_INT_CLASS || REQUIRE_DATATYPE_FLOAT_CLASS || REQUIRE_DATATYPE_COMPLEX_CLASS, \
			out *= in;
	)
	REDUCTION_FUNCTION( \
			land, \
			REQUIRE_DATATYPE_INT_CLASS, \
			out = out && in;
	)
	REDUCTION_FUNCTION( \
			lor, \
			REQUIRE_DATATYPE_INT_CLASS, \
			out = out || in;
	)
	REDUCTION_FUNCTION( \
			lxor, \
			REQUIRE_DATATYPE_INT_CLASS, \
			out = (out && !in) || (!out && in);
	)

	REDUCTION_FUNCTION( \
			band, \
			REQUIRE_DATATYPE_INT_CLASS, \
			out = out & in;
	)
	REDUCTION_FUNCTION( \
			bor, \
			REQUIRE_DATATYPE_INT_CLASS, \
			out = out | in;
	)
	REDUCTION_FUNCTION( \
			bxor, \
			REQUIRE_DATATYPE_INT_CLASS, \
			out = out ^ in;
	)

	// minloc and maxloc are special, because we require a pair_type. So we need an additional level
	// of indirection to extract the two elements of the pair
	template<typename T>
	struct minloc_spec {
		inline static void Do(T& out, const T& in) {
			TMPI_FATAL_ERROR("not implemented function for datatype "
					<< datatype_traits<T>::name, __FILE__, __LINE__);
		}
	};
	template<typename T1>
	struct minloc_spec<pair_type<T1, int> > {
		using T = pair_type<T1, int>;
		inline static void Do(T& out, const T& in) {
			if (out.first > in.first) {
				out.first = in.first;
				out.second = in.second;
			}
		}
	};
	REDUCTION_FUNCTION( \
			minloc, \
			REQUIRE_DATATYPE_LOC_CLASS, \
			minloc_spec<T>::Do(out, in);
	)

	template<typename T>
	struct maxloc_spec {
		inline static void Do(T& out, const T& in) {
			TMPI_FATAL_ERROR("not implemented function for datatype "
					<< datatype_traits<T>::name, __FILE__, __LINE__);
		}
	};
	template<typename T1>
	struct maxloc_spec<pair_type<T1, int> > {
		using T = pair_type<T1, int>;
		inline static void Do(T& out, const T& in) {
			if (out.first < in.first) {
				out.first = in.first;
				out.second = in.second;
			}
		}
	};
	REDUCTION_FUNCTION( \
			maxloc, \
			REQUIRE_DATATYPE_LOC_CLASS, \
			maxloc_spec<T>::Do(out, in);
	)

} // END operations

} // END allreduce

}
}
}
