// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _COMM_COMMUNICATOR_H_635106f3226c46d3d114158785752466
#define _COMM_COMMUNICATOR_H_635106f3226c46d3d114158785752466

#include "global.h"
#include "tasks.h"

#include "utility/error.h"

#include "global_communicator.h"

namespace tinympi {

// There is one Communicator per Thread; one GlobalCommunicator per Node. Each Communicator (e.g.,
// comm_world) has one Communicator per Thread and one GlobalCommunicator per Node.
class Communicator {
public:

	// return the communicator id (e.g., for pt2pt communication)
	inline comm_id_t get_id() const      { return global_communicator_->id_; }
	rank_type get_rank() const;
	inline rank_type get_size() const    { return global_communicator_->num_tasks_; }
	rank_type get_commrank(global_id_t) const;
	tasks::Task& get_task(rank_type rank) const;

private:
	// each communicator has a global communicator attached which is shared among all tasks on node
	GlobalCommunicator* global_communicator_;
	// a communicator is shared between all tasks of one thread. Therefore we have a mapping
	// task_id -> commrank
	rank_type* task_ranks_;

////////////////////////////////////
// STATIC FUNCTIONS
public:
	static void post_init();
	static void finalize();
};

////////////////////////////////////
// FUNCTIONS
inline rank_type Communicator::get_rank() const {
	return task_ranks_[tasks::get_current()->get_identifier().local.task_id];
}
inline rank_type Communicator::get_commrank(global_id_t global_id) const {
	// TODO inefficient
	for (int i = 0; i < get_size(); ++i) {
		TMPI_LOG_LEVEL(1, "compare", global_communicator_->task_mapping_[i]->get_identifier().global_identifier
				<< " " << global_id);
		if (global_communicator_->task_mapping_[i]->get_identifier().global_identifier == global_id) {
			return i;
		}
	}
	TMPI_FATAL_ERROR("did not find rank.", __FILE__, __LINE__);
	return -1;
}
inline tasks::Task& Communicator::get_task(rank_type rank) const {
	TMPI_ASSERT_FATAL_EFF(rank >= 0 && rank < get_size(), "Invalid rank.", __FILE__, __LINE__);
	return *(global_communicator_->task_mapping_[rank]);
}

extern thread_local Communicator comm_world_;

} // END NAMESPACE tinympi

#endif //_COMM_COMMUNICATOR_H_
