// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _DATATYPES_H_635106f3226c46d3d114158785752466
#define _DATATYPES_H_635106f3226c46d3d114158785752466

#include <cstdint>

// all supported datatypes in public namespace
enum TMPI_Datatype {
	TMPI_CHAR,
	TMPI_SHORT,
	TMPI_INT,
	TMPI_LONG,
	TMPI_UNSIGNED_CHAR,
	TMPI_UNSIGNED_SHORT,
	TMPI_UNSIGNED,
	TMPI_UNSIGNED_LONG,
	TMPI_FLOAT,
	TMPI_DOUBLE,
	TMPI_LONG_DOUBLE,
	TMPI_BYTE,

	TMPI_FLOAT_INT,
	TMPI_DOUBLE_INT,
	TMPI_LONG_INT,
	TMPI_2INT,
	TMPI_SHORT_INT,
	TMPI_LONG_DOUBLE_INT
};

namespace tinympi {
namespace impl {

// pair_type is the internal storage format for combined types, e.g. TMPI_FLOAT_INT
template<typename T1, typename T2>
struct pair_type {
	T1 first;
	T2 second;
};

// all datatypes belong to different reduction classes (see section 4.9 in MPI-Standard)
enum class reduction_class_t {
	datatype_int_class,
	datatype_float_class,
	datatype_complex_class,
	datatype_byte_class,
	datatype_other_class,
	datatype_loc_class // used for minloc / maxloc
};

template<typename T>
struct datatype_traits;
template<>
struct datatype_traits<char> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_other_class;
	static constexpr char const * name = "TMPI_CHAR";
};
template<>
struct datatype_traits<short> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_int_class;
	static constexpr char const * name = "TMPI_SHORT";
};
template<>
struct datatype_traits<int> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_int_class;
	static constexpr char const * name = "TMPI_INT";
};
template<>
struct datatype_traits<long> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_int_class;
	static constexpr char const * name = "TMPI_LONG";
};
template<>
struct datatype_traits<unsigned char> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_other_class;
	static constexpr char const * name = "TMPI_UNSIGNED_CHAR";
};
template<>
struct datatype_traits<unsigned short> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_int_class;
	static constexpr char const * name = "TMPI_UNSIGNED_SHORT";
};
template<>
struct datatype_traits<unsigned> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_int_class;
	static constexpr char const * name = "TMPI_UNSIGNED";
};
template<>
struct datatype_traits<unsigned long> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_int_class;
	static constexpr char const * name = "TMPI_UNSIGEND_LONG";
};
template<>
struct datatype_traits<float> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_float_class;
	static constexpr char const * name = "TMPI_FLOAT";
};
template<>
struct datatype_traits<double> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_float_class;
	static constexpr char const * name = "TMPI_DOUBLE";
};
template<>
struct datatype_traits<long double> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_float_class;
	static constexpr char const * name = "TMPI_LONG_DOUBLE";
};
template<>
struct datatype_traits<uint8_t*> {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_byte_class;
	static constexpr char const * name = "TMPI_BYTE";
};
template<>
struct datatype_traits<pair_type<float, int> > {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_loc_class;
	static constexpr char const * name = "TMPI_FLOAT_INT";
};
template<>
struct datatype_traits<pair_type<double, int> > {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_loc_class;
	static constexpr char const * name = "TMPI_DOUBLE_INT";
};
template<>
struct datatype_traits<pair_type<long, int> > {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_loc_class;
	static constexpr char const * name = "TMPI_LONG_INT";
};
template<>
struct datatype_traits<pair_type<int, int> > {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_loc_class;
	static constexpr char const * name = "TMPI_2INT";
};
template<>
struct datatype_traits<pair_type<short, int> > {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_loc_class;
	static constexpr char const * name = "TMPI_SHORT_INT";
};
template<>
struct datatype_traits<pair_type<long double, int> > {
	static constexpr reduction_class_t reduction_class = reduction_class_t::datatype_loc_class;
	static constexpr char const * name = "TMPI_LONG_DOUBLE_INT";
};

}
}

#endif
