// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _PROVIDERS_H_635106f3226c46d3d114158785752466
#define _PROVIDERS_H_635106f3226c46d3d114158785752466

#include "provider_impl/baserequest.h"
#include "provider_impl/progress.h"
#include "provider_impl/config.h"

namespace tinympi {
namespace impl {

constexpr unsigned max_request_size = max(
		NodeWrapper::max_request_size,
		ThreadWrapper::max_request_size
);

static_assert(max_request_size <= BASE_REQUEST_SIZE && max_request_size * 2 > BASE_REQUEST_SIZE,
		"We require max_request_size <= BASE_REQUEST_SIZE < 2*max_request_size");

// depending on the provider type, forwards the arguments to the right wrapper
template<dispatch_methods dispatch_method, typename... Args>
inline int request_dispatch(BaseRequest** reqptr, Args... args) {
	BaseRequest* req = *reqptr;
	while (req != nullptr) {
		switch (req->provider_type) {
		case provider_types::node2node:
			TMPI_LOG_LEVEL(2, "request_dispatch", "forward to node2node");
			return NodeWrapper::request_dispatch<dispatch_method>(
					reinterpret_cast<NodeWrapper::ProviderBaseRequest*>(req), args...);

		case provider_types::thread2thread:
			TMPI_LOG_LEVEL(2, "request_dispatch", "forward to thread2thread");
			return ThreadWrapper::request_dispatch<dispatch_method>(
					reinterpret_cast<ThreadWrapper::ProviderBaseRequest*>(req), args...);

		default:
			TMPI_FATAL_ERROR("unknown provider type", __FILE__, __LINE__);
			return -1;
		}
		req = req->next_anysource_request;
	}
	return TMPI_SUCCESS;
}
inline bool other_request_already_retrieved(BaseRequest* req) {
	// LOCK REQUIRED FOR thread and node queue, for try_cancel_request (OK)

	// we test all requests linked by next_anysource_request whether they have already been retrieved
	// until we are back to req
	BaseRequest* head = req;
	bool other_retrieved = false;
	// the first next anysource request
	req = req->next_anysource_request;
	while (req != nullptr && req != head) {
		other_retrieved |= (!request_dispatch<dispatch_methods::try_cancel_request>(&req));
		req = req->next_anysource_request;
	}
	return other_retrieved;
}
// delete all other requests. note that in contradiction to other_request_already_retrieved,
// this function enforces all other requests to be cancelled, while for the function above,
// each other provider can decide whether it wants to retrieved the request.
inline void other_matched_anysource(TMPI_Request req, TMPI_Request tail_req) {
	// LOCK REQUIRED for node and thread queue, for try_cancel_request (OK)

	using namespace tinympi;
	using namespace tinympi::impl;
	// remove all other requests. req ist the first request to be removed, tail_req is the request
	// which does not have to be removed
	while (req != tail_req) {
		TMPI_LOG_LEVEL(2, "other_matched_anysource", "cancel " << req);
		if (!request_dispatch<dispatch_methods::try_cancel_request>(&req)) {
			// we MUST be able to cancel the request, otherwise data would have been retrieved
			// twice for this request
			TMPI_FATAL_ERROR("could not cancel request", __FILE__, __LINE__);
		}
		// then delete the request because it is no more required
		TMPI_Request next_req = req->next_anysource_request;
		request_dispatch<dispatch_methods::delete_request>(&req);
		req = next_req;
	}
	// when we have cancelled all other requests, we are a normal receive request
	req->next_anysource_request = nullptr;

	TMPI_LOG_LEVEL(3, "other_matched_anysource", "req = " << req << "; "
			<< "req->next_anysource_request = " << req->next_anysource_request);
}

// since a request can be an anysource, we need to get the terminated request and fill the status
// from this request
inline int fill_status(BaseRequest** reqptr, TMPI_Status* status) {
	// no lock required
	using namespace tinympi;

	if (status != TMPI_STATUS_IGNORE) {
		TMPI_LOG_LEVEL(3, "fill_status", "req = " << *reqptr);
		BaseRequest* req = *reqptr;
		if (req->next_anysource_request != nullptr) {
			// if we are an anysource, we look for one request that is not cancelled.
			while (req->atomic_state.load().data.completion_status
					== completion_status_type::REQUEST_CANCELLED) {
				req = req->next_anysource_request;

				// if we are back to the first request, all requests are cancelled
				if (req == *reqptr)
					return false;
			}

			// for this request we then fill the status
			return request_dispatch<dispatch_methods::fill_status>(&req, status);
		} else {
			// if it is not an anysource, we can just forward the request
			return request_dispatch<dispatch_methods::fill_status>(reqptr, status);
		}
	} else {
		// if we ignore the status, then return false
		return false;
	}
}

} // END NAMESPACE impl
} // END NAMESPACE tinympi

#endif // _PROVIDERS_H_
