// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _OFI_H_635106f3226c46d3d114158785752466
#define _OFI_H_635106f3226c46d3d114158785752466

#include <vector>

#include "rdma/fabric.h"
#include "rdma/fi_domain.h"
#include "rdma/fi_cm.h"
#include "rdma/fi_errno.h"
#include "rdma/fi_endpoint.h"
#include "rdma/fi_tagged.h"

#include "status.h"
#include "global.h"
#include "utility/error.h"

#include "ofi_recvrequest.h"
#include "ofi_sendrequest.h"

namespace tinympi {

namespace ofi {

namespace impl {

//tinympi::ofi::impl::handle_request
inline int handle_wc(OfiRequest* req, fi_cq_tagged_entry& wc) {
	// dispatch wc to the right request
	switch (req->ofi_request_type) {
	case ofi_request_types::OFI_SEND:
		return ofiSendRequest::handle_poll(TO_OfiSendRequest(req), wc);
	case ofi_request_types::OFI_RECV:
		return ofiRecvRequest::handle_poll(TO_OfiRecvRequest(req), wc);

	default:
		TMPI_FATAL_ERROR("Unknown request type", __FILE__, __LINE__);
		exit(1);
	}
	return 0;
}

} // namespace impl

//tinympi::ofi::ofi_wrapper
struct ofi_wrapper {
private:
	// import some classes from the impl class to the namespace of the struct
	// we should not move them outside of impl, because they are not directly
	// availbale from the outside
	using OfiEndpoint = ofi::impl::OfiEndpoint;
	using OfiRequest = ofi::impl::OfiRequest;
	using BaseRequest = tinympi::impl::BaseRequest;
	using dispatch_methods = tinympi::impl::dispatch_methods;

	static int progress(const OfiEndpoint*);

public:
	using ProviderBaseRequest = ofi::impl::OfiRequest;

	static int progress();
	static int isend(void*, buflen_t, tasks::Task*, tag_t, context_id_t, BaseRequest**);
	static int posted_request(void*, buflen_t, const tasks::Task*, tag_t, context_id_t,
			BaseRequest**);
	static bool recv_or_post_anysource(void*, buflen_t, tag_t, context_id_t, BaseRequest**);

	static void init();
	static void pre_init();
	static void finalize();
	static void post_finalize();

	static constexpr int max_request_size = max(
			sizeof(impl::OfiSendRequest), sizeof(impl::OfiRecvRequest)
	);

	template<dispatch_methods dispatch_method, typename... Args>
	static int request_dispatch(OfiRequest* req, Args... args) {
		TMPI_FATAL_ERROR("Invalid call!", __FILE__, __LINE__);
	}
};
// just dispatch to the right function.
template<>
inline int ofi_wrapper::request_dispatch<ofi_wrapper::dispatch_methods::fill_status>(OfiRequest* req,
		TMPI_Status* status) {
	using namespace impl;

	switch (req->ofi_request_type) {
	case ofi_request_types::OFI_SEND:
		return ofiSendRequest::fill_status(TO_OfiSendRequest(req), status);
	case ofi_request_types::OFI_RECV:
		return ofiRecvRequest::fill_status(TO_OfiRecvRequest(req), status);

	default:
		TMPI_FATAL_ERROR("unknown request type", __FILE__, __LINE__);
		return -1;
	}
}
template<>
inline int ofi_wrapper::request_dispatch<ofi_wrapper::dispatch_methods::delete_request>(
		OfiRequest* req) {

	free(req);
	return TMPI_SUCCESS;
}
template<>
inline int ofi_wrapper::request_dispatch<ofi_wrapper::dispatch_methods::try_cancel_request>(
		OfiRequest* oreq) {
	using namespace impl;
	int res;

	// submit to cancel
	res = fi_cancel(reinterpret_cast<fid_t>(provider_.get_ep()), &oreq->fabric_context);
	TMPI_ASSERT_FATAL(res == 0, "cancel request was not submitted", __FILE__, __LINE__);

	// we now progress as usually and wait for either occuring something in the error queue, or for
	// getting the data. Note that this is nearly the same as for progress()
	fi_cq_tagged_entry wc;
	while (true) {
		int res = fi_cq_read(provider_.get_cq(), &wc, 1);
		if (res > 0) {
			// got something
			TMPI_LOG_LEVEL(3, "ofi_wrapper", "found something in cq!");
			fi_context* ctx = static_cast<fi_context*>(wc.op_context);
			OfiRequest* req = ofiRequest::from_context(ctx);
			handle_wc(req, wc);
			if (req == oreq)
				return false; // could not abort!
		} else if (res == 0) {
			// if we have no answer, we may wait until we get what we expect...
			continue;
		} else if (res == -FI_EAVAIL) {
			// if an error is available this will hopefully be the answer to our cancellation request

			// we read it
			fi_cq_err_entry ec;
			res = fi_cq_readerr(provider_.get_cq(), &ec, 0);

			// verify if the error matches our expectations
			TMPI_ASSERT_FATAL(res > 0, "There should be an error available, but there is not ...",
					__FILE__, __LINE__);
			TMPI_ASSERT_FATAL(ec.op_context == &oreq->fabric_context,
					"error does not match expected request", __FILE__, __LINE__);
			TMPI_ASSERT_FATAL(ec.err == FI_ECANCELED, "Wrong reason for error", __FILE__, __LINE__);

			// then set the request to cancelled
			baseRequest::request_cancelled(TO_BaseRequest(oreq));

			// then return true, because the request was cancelled
			return true;
		} else {
			// other errors are real errors!
			TMPI_FATAL_ERROR("fi_cq_read() returned " << res, __FILE__, __LINE__);
		}
	}
	return 0;

}
// this function is called when a request was posted for a given source.
inline int ofi_wrapper::posted_request(void* buf, buflen_t buflen, const tasks::Task* src_descr,
		tag_t tag, context_id_t context_id, BaseRequest** req) {
	using namespace impl;

	// in this case we just generate a new recv request.
	OfiEndpoint* endpoint = endpoints_[src_descr->get_identifier().local.node_id];
	*req = TO_BaseRequest(ofiRecvRequest::generate(buf, buflen, src_descr, tag, context_id,
			endpoint));
	return TMPI_SUCCESS;
}
// this function is called when a request was posted with ANY_SOURCE as source
inline bool ofi_wrapper::recv_or_post_anysource(void* buf, buflen_t buflen, tag_t tag,
		context_id_t context_id, BaseRequest** req) {
	using namespace impl;

	// in this case we just generate a new recv request.
	*req = TO_BaseRequest(ofiRecvRequest::generate_anysource(buf, buflen, tag, context_id));
	return false;
}
inline int ofi_wrapper::progress(const OfiEndpoint*) {
	return progress();
}
inline int ofi_wrapper::progress() {
	using namespace impl;

	fi_cq_tagged_entry wc;
	while (true) {
		int res = fi_cq_read(provider_.get_cq(), &wc, 1);
		if (res > 0) {
			// got something
			TMPI_LOG_LEVEL(3, "ofi_wrapper", "found something in cq!");
			fi_context* ctx = static_cast<fi_context*>(wc.op_context);
			handle_wc(ofiRequest::from_context(ctx), wc);
		} else if (res == -FI_EAGAIN) {
			break;
		} else if (res == -FI_EAVAIL) {
			TMPI_FATAL_ERROR("error available but not handled" << res, __FILE__, __LINE__);
		} else {
			TMPI_FATAL_ERROR("fi_cq_read() returned " << res, __FILE__, __LINE__);
		}
	}
	return 0;
}
inline int ofi_wrapper::isend(void* buf, buflen_t buflen, tasks::Task* dest_pos, tag_t tag,
		context_id_t context_id, BaseRequest** reqptr) {
	using namespace impl;

	// get the endpoint of the respective node
	OfiEndpoint* endpoint = endpoints_[dest_pos->get_identifier().local.node_id];

	OfiSendRequest* req = ofiSendRequest::generate(buf, buflen, tag, context_id, *dest_pos,
			endpoint);
	*reqptr = reinterpret_cast<BaseRequest*>(req);
	return TMPI_SUCCESS;
}

inline void ofi_wrapper::pre_init() {
	using namespace impl;

	provider_.init();
	OfiProvider::setup_endpoints();
}
inline void ofi_wrapper::init() {
}
inline void ofi_wrapper::finalize() {
}
inline void ofi_wrapper::post_finalize() {
	using namespace impl;
	OfiProvider::destroy_endpoints();

	provider_.finalize();
}

} // END NAMESPACE ofi
} // END NAMESPACE tinympi

#endif //_OFI_H_
