// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#include <cstring>

#include "ofi.h"
#include "utility/external.h"
#include "utility/error.h"

namespace tinympi {
namespace ofi {
namespace impl {

OfiProvider provider_;
OfiEndpoint** endpoints_;
global_id_t num_endpoints_;

void OfiProvider::init() {

	fi_info* hints = fi_allocinfo();
	fi_info* info;
	hints->mode                       = FI_CONTEXT;
	hints->caps                       = FI_TAGGED | FI_DIRECTED_RECV;

	hints->ep_attr->type              = FI_EP_RDM;
	hints->tx_attr->msg_order         = FI_ORDER_SAS;
	hints->rx_attr->msg_order         = FI_ORDER_SAS;

	hints->domain_attr->threading        = FI_THREAD_ENDPOINT;
	hints->domain_attr->control_progress = FI_PROGRESS_MANUAL;
	hints->domain_attr->resource_mgmt    = FI_RM_ENABLED;
	hints->domain_attr->av_type          = FI_AV_MAP;

	int res;
	res = fi_getinfo(FI_VERSION(1, 3), nullptr, nullptr, 0, hints, &info);
	TMPI_ASSERT_FATAL(res == 0, "fi_getinfo() returned " << res, __FILE__, __LINE__);

	// just take the first
	fi_info* prov = info;
	TMPI_LOG_LEVEL(-1, "provider name", prov->domain_attr->name);
	max_inject_size_ = prov->tx_attr->inject_size; // store max inject size

	// open fabric
	fabric_ = 0;
	res = fi_fabric(prov->fabric_attr, &fabric_, nullptr);
	TMPI_ASSERT_FATAL(res == 0, "fi_fabric() returned " << res, __FILE__, __LINE__);

	// open domain
	res = fi_domain(fabric_, prov, &domain_, nullptr);
	TMPI_ASSERT_FATAL(res == 0, "fi_domain() returned " << res, __FILE__, __LINE__);

	// open completion queue
	fi_cq_attr cq_attr;
	std::memset(&cq_attr, 0, sizeof(cq_attr));
	cq_attr.format = FI_CQ_FORMAT_TAGGED;
	res = fi_cq_open(domain_, &cq_attr, &cq_, nullptr);
	TMPI_ASSERT_FATAL(res == 0, "fi_cq_open() returned " << res, __FILE__, __LINE__);

	res = fi_endpoint(domain_, prov, &ep_, nullptr);
	TMPI_ASSERT_FATAL(res == 0, "fi_endpoint() returned " << res, __FILE__, __LINE__);

	fi_av_attr av_attr;
	std::memset(&av_attr, 0, sizeof(av_attr));
	av_attr.type = FI_AV_MAP;
	av_attr.count = env::num_nodes;
	res = fi_av_open(domain_, &av_attr, &av_, nullptr);
	TMPI_ASSERT_FATAL(res == 0, "fi_av() returned " << res, __FILE__, __LINE__);

	res = fi_ep_bind(ep_, reinterpret_cast<fid*>(av_), 0);
	TMPI_ASSERT_FATAL(res == 0, "fi_ep_bind(ep, av) returned " << res, __FILE__, __LINE__);

	res = fi_ep_bind(ep_, reinterpret_cast<fid*>(cq_), FI_TRANSMIT | FI_RECV);
	TMPI_ASSERT_FATAL(res == 0, "fi_ep_bind(ep, cq) returned " << res, __FILE__, __LINE__);

	res = fi_enable(ep_);
	TMPI_ASSERT_FATAL(res == 0, "fi_enable() returned " << res, __FILE__, __LINE__);

	fi_freeinfo(hints);
	fi_freeinfo(info);

}
void OfiProvider::finalize() {
	int res;

	res = fi_close(reinterpret_cast<fid*>(ep_));
	TMPI_ASSERT_FATAL(res == 0, "fi_close(ep) returned " << res, __FILE__, __LINE__);

	res = fi_close(reinterpret_cast<fid*>(av_));
	TMPI_ASSERT_FATAL(res == 0, "fi_close(av) returned " << res, __FILE__, __LINE__);

	res = fi_close(reinterpret_cast<fid*>(cq_));
	TMPI_ASSERT_FATAL(res == 0, "fi_close(cq) returned " << res, __FILE__, __LINE__);

	res = fi_close(reinterpret_cast<fid*>(domain_));
	TMPI_ASSERT_FATAL(res == 0, "fi_close(domain) returned " << res, __FILE__, __LINE__);

	res = fi_close(reinterpret_cast<fid*>(fabric_));
	TMPI_ASSERT_FATAL(res == 0, "fi_close(fabric) returned " << res, __FILE__, __LINE__);

}

void OfiProvider::setup_endpoints() {
	int res;

	num_endpoints_ = env::num_nodes;
	global_id_t node_rank = env::node_id;

	endpoints_ = new OfiEndpoint*[num_endpoints_];

	char addrname[FI_NAME_MAX];
	size_t addrnamelen = FI_NAME_MAX;
	res = fi_getname(reinterpret_cast<fid*>(provider_.ep_), addrname, &addrnamelen);
	TMPI_ASSERT_FATAL(res == 0, "fi_getname() returned " << res, __FILE__, __LINE__);

	char* alladdrnames = (char*) malloc(FI_NAME_MAX * num_endpoints_);
	res = utility::external::allgather(addrname, FI_NAME_MAX, MPI_BYTE, alladdrnames, FI_NAME_MAX,
			MPI_BYTE, MPI_COMM_WORLD);
	TMPI_ASSERT_FATAL(res == MPI_SUCCESS, "allgather() returned " << res, __FILE__, __LINE__);

	fi_addr_t* addr = new fi_addr_t[num_endpoints_];
	global_id_t ins_res = fi_av_insert(provider_.av_, alladdrnames, num_endpoints_, addr, 0,
			nullptr);
	TMPI_ASSERT_FATAL(ins_res == num_endpoints_, "fi_av_insert() returned " << res, __FILE__,
			__LINE__);

	for (global_id_t i = 0; i < num_endpoints_; ++i) {
		if (i != node_rank) {
			endpoints_[i] = new OfiEndpoint(provider_.ep_, addr[i]);
		} else {
			endpoints_[i] = nullptr;
		}
	}

	free(alladdrnames);
}
void OfiProvider::destroy_endpoints() {
	for (global_id_t i = 0; i < num_endpoints_; ++i) {
		if (i != env::node_id)
			delete endpoints_[i];
	}
}

} // END NAMESPACE impl
} // END NAMESPACE ofi
} // END NAMESPACE tinympi
