// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _OFI_RECVREQUEST_H_635106f3226c46d3d114158785752466
#define _OFI_RECVREQUEST_H_635106f3226c46d3d114158785752466

#include <rdma/fi_domain.h>

#include "global.h"
#include "utility/error.h"

#include "ofi_endpoint.h"
#include "ofi_request.h"
#include "ofi_provider.h"

namespace tinympi {
namespace ofi {
namespace impl {

//tinympi::ofi::impl::recv_request_states
enum class recv_request_states {
	INITIALIZED,
	DONE
};

#define OFI_RECV_REQUEST_MEMBERS \
	OFI_REQUEST_MEMBERS \
	recv_request_states req_state;

//tinympi::ofi::impl::OfiRecvRequest
struct OfiRecvRequest {
	OFI_RECV_REQUEST_MEMBERS
};
REQUESTS_CONVERTIBLE(OfiRecvRequest, BaseRequest)
REQUESTS_CONVERTIBLE(OfiRecvRequest, OfiRequest)

static_assert(std::is_standard_layout<OfiRecvRequest>::value, "OfiRecvRequest must be POD!");
ASSERT_OFI_REQUEST_COMPATIBLE(OfiRecvRequest)
#define ASSERT_OFI_RECVREQUEST_COMPATIBLE(TYPE) \
	ASSERT_OFI_REQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(OfiRecvRequest, TYPE, req_state)

namespace ofiRecvRequest {

	namespace impl {

		// converts a state into a string
		inline std::string to_string(recv_request_states state) {
			switch (state) {
			case recv_request_states::INITIALIZED:
				return "recv_request_states::INITIALIZED";
			case recv_request_states::DONE:
				return "recv_request_states::DONE";
			default:
				return std::string("??? ") + std::to_string(static_cast<int>(state));
			}
		}

		inline void init(OfiRecvRequest* req) {
			ofiRequest::init(reinterpret_cast<OfiRequest*>(req), ofi_request_types::OFI_RECV);
			req->req_state = recv_request_states::INITIALIZED;
		}
		inline int initialized_to_recvd(OfiRecvRequest* req, fi_cq_tagged_entry& wc) {
			using namespace tinympi::impl;

			// if we have received the message, we need to update tag and size
			req->buflen = wc.len;
			req->tag = ofiRequest::extract_tag(wc.tag);
			req->context_id = ofiRequest::extract_contextid(wc.tag);

			req->req_state = recv_request_states::DONE;
			baseRequest::request_completed(TO_BaseRequest(req));
			return TMPI_SUCCESS;
		}


	} // END NAMESPACE impl

	inline bool fill_status(OfiRecvRequest* req, TMPI_Status* status) {
		using tinympi::impl::baseRequest::is_terminated;

		if (is_terminated(reinterpret_cast<BaseRequest*>(req))) {
			status->MPI_SOURCE = req->peer_rank;
			status->MPI_TAG = req->tag;
			status->MPI_ERROR = TMPI_SUCCESS;
			status->size = req->buflen;
			return true;

		} else {
			return false;
		}
	}

	inline int handle_poll(OfiRecvRequest* req, fi_cq_tagged_entry& wc) {
		using namespace impl;
		TMPI_DASSERT_FATAL(req->req_state == recv_request_states::INITIALIZED,
				"unexpected request state (" << to_string(req->req_state) << ") at this place!",
				__FILE__, __LINE__);

		if (req->next_anysource_request != nullptr) {
			other_matched_anysource(req->next_anysource_request, TO_BaseRequest(req));
		}
		return initialized_to_recvd(req, wc);
	}

	inline OfiRecvRequest* generate(void* buf, buflen_t buflen, const tasks::Task* src_descr,
			tag_t tag, context_id_t context_id, OfiEndpoint* endpoint) {

		OfiRecvRequest* req = TO_OfiRecvRequest(baseRequest::malloc());

		INIT_BASE_REQUEST( \
			req, \
			src_descr->get_identifier().global_identifier, \
			context_id, \
			tag, \
			buf, \
			buflen, \
			tasks::get_current()
		)

		uint64_t matchbits = ofiRequest::get_tag(context_id, tasks::get_threadunique_id(), tag);
		uint64_t maskbits = ofiRequest::get_mask(tag);

		int res = -FI_EAGAIN;
		while (res == -FI_EAGAIN)
			res = fi_trecv(endpoint->get_ep(), req->buf, req->buflen, NULL, endpoint->get_addr(),
					matchbits, maskbits, &req->fabric_context);
		TMPI_ASSERT_FATAL(res == 0, "fi_trecv() failed", __FILE__, __LINE__);

		return req;
	}
	inline OfiRecvRequest* generate_anysource(void* buf, buflen_t buflen, tag_t tag,
			context_id_t context_id) {

		OfiRecvRequest* req = TO_OfiRecvRequest(baseRequest::malloc());

		INIT_BASE_REQUEST( \
			req, \
			TMPI_ANY_SOURCE, \
			context_id, \
			tag, \
			buf, \
			buflen, \
			tasks::get_current()
		)

		uint64_t matchbits = ofiRequest::get_tag(context_id, tasks::get_threadunique_id(), tag);
		uint64_t maskbits = ofiRequest::get_mask(tag);

		int res = -FI_EAGAIN;
		while (res == -FI_EAGAIN)
			res = fi_trecv(provider_.get_ep(), req->buf, req->buflen, NULL, FI_ADDR_UNSPEC,
					matchbits, maskbits, &req->fabric_context);
		TMPI_ASSERT_FATAL(res == 0, "fi_trecv() failed", __FILE__, __LINE__);

		return req;
	}


} // END NAMESPACE ofiSendRequest
} // END NAMESPACE impl
} // END NAMESPACE ofi
} // END NAMESPACE tinympi

#endif //_OFI_RECVREQUEST_H_
