// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_EAGERENDPOINT_H_635106f3226c46d3d114158785752466
#define _VERBS_EAGERENDPOINT_H_635106f3226c46d3d114158785752466

#include <malloc.h>
#include <array>

#include "verbs_eagermessage.h"
#include "verbs_peerinfo.h"
#include "verbs_global.h"
#include "verbs_config.h"

#include "utility/error.h"
#include "tasks.h"

namespace tinympi {
namespace verbs {
namespace impl {
namespace eager_channel_slot_status {
	enum eager_channel_slot_status {
		free,
		reserved,
		nackldg
	};
}

struct VerbsEagerEndpoint {
public:
	//////////////////////////////
	// initialization functions
	void setup(ibv_pd* pd);
	void destroy();

	void eager_set_peer_info(uint32_t, uint64_t, uint32_t, uint64_t);
	void eager_get_recv_slots_info(uint32_t&, uint64_t&) const;
	void eager_get_remote_info(uint32_t&, uint64_t&) const;

	//////////////////////////////
	// piggyback functions
	bool test_explicit_piggyback() const;
	std::pair<piggyback_t, piggybackmsg_t> get_explicit_piggyback_message();
	piggybackmsg_t get_piggyback_message();
	void recvd_piggyback_message(piggybackmsg_t cnt);

	//////////////////////////////
	// eager remote slot functions
	int eager_reserve_remote_slot();
	bool eager_has_remote_slot() const;
	const PeerInfo& eager_get_remote_status_peer_info() const {
		return eager_remote_status_peer_info_;
	}

	//////////////////////////////
	// eager recv slot functions
	void eager_free_recv_slot(int i);
	const EagerMessage* eager_get_recv_slot(int i) const;
	const PeerInfo& eager_get_recv_slots_peer_info() const {
		return eager_recv_slots_peer_info_;
	}
	bool eager_poll_and_lock_recv_slots(rid_t expected_rid, int& pos, std::unique_lock<std::mutex>&);

private:
	//////////////////////////////
	// sender
	// state of the remote recv slots
	eager_message_status* eager_remote_status_;
	// memory region of the remote status slots
	ibv_mr* eager_remote_status_mr_;
	// peer info of the remote slot (rkey / addr)
	PeerInfo eager_remote_status_peer_info_;
	// the next element where we can send to
	int eager_remote_head_;
	// the last remote slot that is reserved
	int eager_remote_tail_;

	//////////////////////////////
	// receiver
	// recv slots
	EagerMessage* eager_recv_slots_;
	// memory region of eager_recv_slots_
	ibv_mr* eager_recv_slots_mr_;
	// memory region of the remote recv slots
	PeerInfo eager_recv_slots_peer_info_;
	// the first nackldg element (<= reserved_tail_)
	int eager_recv_nackldg_tail_;
	// the first reserved element
	// where we recv the next element (used for polling)
	std::atomic<int> eager_recv_head_;

};

//////////////////////////////
// INITIALIZATION FUNCTIONS
inline void VerbsEagerEndpoint::setup(ibv_pd* pd) {
	//////////////////////////////
	// send
	// we have EAGER_CHMESSAGE_LIMIT remote status slots
	eager_remote_status_ = (eager_message_status*) memalign(64,
			EAGER_CHMESSAGE_LIMIT*sizeof(eager_message_status));
	for (unsigned i = 0; i < EAGER_CHMESSAGE_LIMIT; ++i)
		eager_remote_status_[i] = eager_message_status::free;

	// register memory region
	eager_remote_status_mr_ = ibv_reg_mr(pd, eager_remote_status_,
			EAGER_CHMESSAGE_LIMIT*sizeof(eager_message_status),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(eager_remote_status_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);

	// our first send goes to the first position
	eager_remote_head_ = 0;
	eager_remote_tail_ = 0;

	//////////////////////////////
	// receive
	// we have EAGER_CHMESSAGE_LIMIT recv slots
	eager_recv_slots_ = (EagerMessage*) memalign(64, EAGER_CHMESSAGE_LIMIT*sizeof(EagerMessage));
	std::memset(eager_recv_slots_, 0, EAGER_CHMESSAGE_LIMIT*sizeof(EagerMessage));
	for (unsigned i = 0; i < EAGER_CHMESSAGE_LIMIT; ++i)
		eager_recv_slots_[i].status = eager_message_status::free;

	// register memory region
	eager_recv_slots_mr_ = ibv_reg_mr(pd, eager_recv_slots_,
			EAGER_CHMESSAGE_LIMIT*sizeof(EagerMessage),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(eager_recv_slots_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__, __LINE__);

	// no elements are reserved / free
	eager_recv_nackldg_tail_ = 0;
	eager_recv_head_ = 0;
}
inline void VerbsEagerEndpoint::destroy() {
	int res;

	//////////////////////////////
	// send
	res = ibv_dereg_mr(eager_remote_status_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);

	free(eager_remote_status_);

	//////////////////////////////
	// receive
	res = ibv_dereg_mr(eager_recv_slots_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);

	free(eager_recv_slots_);
}
inline void VerbsEagerEndpoint::eager_get_recv_slots_info(uint32_t& rkey, uint64_t& addr) const {
	// no lock required
	rkey = eager_recv_slots_mr_->rkey;
	addr = reinterpret_cast<uint64_t>(eager_recv_slots_);
}

inline void VerbsEagerEndpoint::eager_get_remote_info(uint32_t& rkey, uint64_t& addr) const {
	// no lock required
	rkey = eager_remote_status_mr_ ->rkey;
	addr = reinterpret_cast<uint64_t>(eager_remote_status_);
}
inline void VerbsEagerEndpoint::eager_set_peer_info(
		uint32_t eager_recv_rkey, uint64_t eager_recv_addr,
		uint32_t eager_remote_rkey, uint64_t eager_remote_addr) {
	// no lock required
	eager_recv_slots_peer_info_.rkey    = eager_recv_rkey;
	eager_recv_slots_peer_info_.addr    = eager_recv_addr;
	eager_remote_status_peer_info_.rkey = eager_remote_rkey;
	eager_remote_status_peer_info_.addr = eager_remote_addr;
}

//////////////////////////////
// EAGER REMOTE SLOT FUNCTIONS
inline int VerbsEagerEndpoint::eager_reserve_remote_slot() {
	// endpoint lock required
	TMPI_LOG_LEVEL(4,"VerbsEagerEndpoint", "reserve remote slot " << eager_remote_head_)
	TMPI_DASSERT_FATAL(eager_remote_status_[eager_remote_head_] == eager_message_status::free,
			"remote slot is not free", __FILE__, __LINE__);

	// check if we still have a remote slot
	if (!eager_has_remote_slot())
		return -1;

	// when there was an explicit piggyback we need first to forward the tail, because otherwise
	// we will have a problem when we set a message to reserved which was not yet observed as
	// free by the tail
	// the second test is to ensure that if the whole array is free, we will not hang up
	while (eager_remote_status_[eager_remote_tail_] == eager_message_status::free
			&& eager_remote_tail_ != eager_remote_head_) {
		RING_INCREMENT(eager_remote_tail_, EAGER_CHMESSAGE_LIMIT);
	}

	eager_remote_status_[eager_remote_head_] = eager_message_status::retrieved;
	int res = eager_remote_head_;
	RING_INCREMENT(eager_remote_head_, EAGER_CHMESSAGE_LIMIT);
	return res;
}
inline bool VerbsEagerEndpoint::eager_has_remote_slot() const {
	// endpoint lock required
	// an eager slot is free, if the next message is also free (ring buffer - we need one
	// free element)
	return eager_remote_status_[(eager_remote_head_+1)&(EAGER_CHMESSAGE_LIMIT-1)]
			== eager_message_status::free;
}
//////////////////////////////
// EAGER RECV SLOT FUNCTIONS

// frees a recv slot, i.e. sets it to nackldgd and forwards the reserved tail to the
// point where the next reserved message is
inline void VerbsEagerEndpoint::eager_free_recv_slot(int i) {
	// the recv slot should be reserved
	TMPI_DASSERT_FATAL(eager_recv_slots_[i].status == eager_message_status::reserved,
			"double freed recv slot", __FILE__, __LINE__);

	// we set the eager slot to nackldg
	eager_recv_slots_[i].status = eager_message_status::nackldg;
}
inline const EagerMessage* VerbsEagerEndpoint::eager_get_recv_slot(int i) const {
	// no lock required
	return &eager_recv_slots_[i];
}
// polls the receive slots for a message with given request id
inline bool VerbsEagerEndpoint::eager_poll_and_lock_recv_slots(rid_t expected_rid, int& pos,
		std::unique_lock<std::mutex>& verbs_lock) {

	// get the current head with which we try to get the recv slot
	int tmp_head = eager_recv_head_ & (EAGER_CHMESSAGE_LIMIT - 1);
	if (eager_recv_slots_[tmp_head].rid == expected_rid) {
		// if possible, abort without write operation
		if (eager_recv_slots_[tmp_head].status != eager_message_status::retrieved)
			return false;

		// if we have the right rid, we try to reserve it, if and only if the slot is retrieved but
		// not reserved yet
		eager_message_status cmp = eager_message_status::retrieved;
		if (eager_recv_slots_[tmp_head].status
				.compare_exchange_weak(cmp, eager_message_status::reserved)) {

			// now the request is reserved and we can increase the head
			pos = eager_recv_head_.fetch_add(1);
			pos &= (EAGER_CHMESSAGE_LIMIT - 1);

			// then lock the appropriate verbs queue
			tasks::Task& task = tasks::get_task(eager_recv_slots_[pos].dest_id);
			verbs_lock = std::unique_lock<std::mutex>(
					static_cast<tasks::LocalTask*>(&task)->verbs_lock_);

			TMPI_LOG_LEVEL(4,"VerbsEagerEndpoint", "successful polling at position " << pos);

			return true;
		} else {
			return false;
		}
	} else {
		return false;
	}
}

//////////////////////////////
// PIGGYBACK FUNCTIONS

// returns whether we should send an explicit piggyback message, which is the case
// when there at least EAGER_EXPLICIT_PIGGYBACK reserved entries
inline bool VerbsEagerEndpoint::test_explicit_piggyback() const {
	// endpoint lock required
	int diff = (eager_recv_head_ & (EAGER_CHMESSAGE_LIMIT - 1)) - eager_recv_nackldg_tail_;
	if (diff < 0)
		diff += EAGER_CHMESSAGE_LIMIT;

	return (diff >= static_cast<int>(EAGER_EXPLICIT_PIGGYBACK));
}
// when we send the explicit piggyback message, we need to get the start and the number
// of entries
inline std::pair<piggyback_t, piggybackmsg_t> VerbsEagerEndpoint::get_explicit_piggyback_message() {
	// endpoint lock required
	piggyback_t pstart = eager_recv_nackldg_tail_;
	piggybackmsg_t pcnt = get_piggyback_message();
	return {pstart, pcnt};
}
// gets the piggyback message, i.e the size of the piggyback area and frees
// the entries (the message we get MUST be sent!)
inline piggybackmsg_t VerbsEagerEndpoint::get_piggyback_message() {
	// endpoint lock required
	TMPI_LOG_LEVEL(3,"VerbsEagerEndpoint", "finding piggyback entries");

	piggyback_t pcnt = 0;
	while (eager_recv_slots_[eager_recv_nackldg_tail_].status == eager_message_status::nackldg
			&& pcnt < std::numeric_limits<piggybackmsg_t>::max()) {
		// if the nackldg tail is nackldgable, we can piggyback and free the entry
		pcnt += 1;
		eager_recv_slots_[eager_recv_nackldg_tail_].status = eager_message_status::free;

		// then increase tail
		RING_INCREMENT(eager_recv_nackldg_tail_, EAGER_CHMESSAGE_LIMIT);
	}

	// return piggyback entries
	TMPI_LOG_LEVEL(3,"VerbsEagerEndpoint", "    found " << (int)pcnt << " piggyback entries");
	return pcnt;
}
inline void VerbsEagerEndpoint::recvd_piggyback_message(piggybackmsg_t pcnt) {
	// endpoint lock required
	// if we received a piggyback message, we can set the remote status of pcnt messages to free
	TMPI_LOG_LEVEL(4, "VerbsEagerEndpoint", "recvd piggyback " << (int)pcnt);
	// we just set pcnt messages to free. Note that eager_remote_tail_ can point to a free
	// message, if there was an explicit piggyback
	while (pcnt > 0) {
		if (eager_remote_status_[eager_remote_tail_] !=  eager_message_status::free) {
			eager_remote_status_[eager_remote_tail_] = eager_message_status::free;
			pcnt -= 1;
		}
		RING_INCREMENT(eager_remote_tail_, EAGER_CHMESSAGE_LIMIT);
	}
}

}
}
}

#endif //_VERBS_EAGERENDPOINT_H_
