// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_EAGERPROVIDER_H_635106f3226c46d3d114158785752466
#define _VERBS_EAGERPROVIDER_H_635106f3226c46d3d114158785752466

#include <atomic>

#include <malloc.h>

#include "global.h"
#include "debug/debugger.h"
#include "utility/error.h"
#include "tasks.h"

#include "verbs_config.h"

#include "verbs_eagermessage.h"

namespace tinympi {
namespace verbs {
namespace impl {

class VerbsEagerSendRequest;

//tinympi::verbs::impl::VerbsProvider
class EagerProvider {
public:
	//////////////////////////////
	// initialization functions
	static void init(ibv_pd* pd);
	static void post_init();
	static void finalize();

	//////////////////////////////
	// send slot functions
	static void eager_free_send_slot(int);
	static std::pair<int, EagerMessage*> eager_get_send_slot();
	static uint32_t eager_get_send_slots_lkey();

	//////////////////////////////
	// piggyback functions
	static const eager_message_status* get_piggyback_send_buffer();
	static uint32_t get_piggyback_send_buffer_lkey();

private:

	//////////////////////////////
	// send slots

	// we have a limited amount of eager send slots that can be used by
	// all endpoints
	static EagerMessage* eager_send_slots_;
	// and the memory region
	static ibv_mr* eager_send_slots_mr_;
	// next send slot to be used
	static thread_local uint_fast16_t eager_send_slots_head_;
	// send_slot_base
	static thread_local uint_fast16_t eager_send_slots_start_;

	//////////////////////////////
	// piggyback send buffer
	// this is used for explicit piggyback
	static eager_message_status* piggyback_send_buffer_;
	// and its memory region
	static ibv_mr* piggyback_send_buffer_mr_;

};

//////////////////////////////
// INITIALIZATION FUNCTIONS
// called by one thread
inline void EagerProvider::init(ibv_pd* pd) {

	//////////////////////////////
	// send slots
	eager_send_slots_ = (EagerMessage*) memalign(64,
			env::num_threads*EAGER_MESSAGE_LIMIT*sizeof(EagerMessage));
	for (unsigned i = 0; i < env::num_threads*EAGER_MESSAGE_LIMIT; ++i)
		eager_send_slots_[i].status = eager_message_status::free;
	eager_send_slots_mr_ = ibv_reg_mr(pd, eager_send_slots_,
			env::num_threads*EAGER_MESSAGE_LIMIT*sizeof(EagerMessage),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(eager_send_slots_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);
	eager_send_slots_head_ = 0;

	//////////////////////////////
	// piggyback send buffer
	piggyback_send_buffer_ = new eager_message_status[EAGER_CHMESSAGE_LIMIT];
	for (unsigned i = 0; i < EAGER_CHMESSAGE_LIMIT; ++i)
		piggyback_send_buffer_[i] = eager_message_status::free;
	piggyback_send_buffer_mr_ = ibv_reg_mr(pd, piggyback_send_buffer_,
			EAGER_CHMESSAGE_LIMIT*sizeof(eager_message_status), 0);
	TMPI_ASSERT_FATAL(piggyback_send_buffer_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);
}
// called by all threads
inline void EagerProvider::post_init() {
	eager_send_slots_start_ = threads::impl::self->rank * EAGER_MESSAGE_LIMIT;
}
// called by one thread
inline void EagerProvider::finalize() {
	int res;

	//////////////////////////////
	// send slots
	res = ibv_dereg_mr(eager_send_slots_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);
	free(eager_send_slots_);

	//////////////////////////////
	// piggyback send buffer
	res = ibv_dereg_mr(piggyback_send_buffer_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);
}

//////////////////////////////
// SEND SLOTS FUNCTIONS
// free a send slot
// each slot can be freed only by one task at one time. This can be done by a different thread.
// i is the global identifier
inline void EagerProvider::eager_free_send_slot(int i) {
	TMPI_DASSERT_FATAL(eager_send_slots_[i].status != eager_message_status::free,
			"double freed send slot", __FILE__, __LINE__);

	eager_send_slots_[i].status.store(eager_message_status::free);
}

// get a free send slot
inline std::pair<int, EagerMessage*> EagerProvider::eager_get_send_slot() {

	// we need to check whether the slots are full. This is the case when the current slot is
	// reserved.

	while (eager_send_slots_[eager_send_slots_start_ +
			(eager_send_slots_head_ & (EAGER_MESSAGE_LIMIT - 1))].status
			!= eager_message_status::free) {
		eager_send_slots_head_ += 1;
	}
	auto my_slot = eager_send_slots_start_ + (eager_send_slots_head_ & (EAGER_MESSAGE_LIMIT - 1));

	// this thread sets the slot to retrieved
	eager_send_slots_[my_slot].status.store(eager_message_status::retrieved);

	return {my_slot, &eager_send_slots_[my_slot]};
}
// trivially thread safe
inline uint32_t EagerProvider::eager_get_send_slots_lkey() {
	// NO LOCK required
	return eager_send_slots_mr_->lkey;
}

//////////////////////////////
// PIGGYBACK FUNCTIONS
// trivially thread safe
inline const eager_message_status* EagerProvider::get_piggyback_send_buffer() {
	// NO LOCK required
	return piggyback_send_buffer_;
}
// trivially thread safe
inline uint32_t EagerProvider::get_piggyback_send_buffer_lkey() {
	// NO LOCK required
	return piggyback_send_buffer_mr_->lkey;
}

} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif // _VERBS_EAGERPROVIDER_H
