// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_EAGERRECVREQUEST_H_635106f3226c46d3d114158785752466
#define _VERBS_EAGERRECVREQUEST_H_635106f3226c46d3d114158785752466

#include <malloc.h>

#include "global.h"

#include "verbs_endpoint.h"
#include "verbs_request.h"
#include "verbs_provider.h"
#include "verbs_config.h"
#include "verbs_recvrequest.h"

namespace tinympi {
namespace verbs {
namespace impl {

void send_explicit_piggyback(VerbsEndpoint* const ep);

//tinympi::verbs::impl::recv_request_state
enum class eager_recv_request_states {
	WAITING_DATA,
	UNEXPECTED_REQUEST,
	DONE
};

#define VERBS_EAGERRECV_REQUEST_MEMBERS \
	VERBS_RECV_REQUEST_MEMBERS \
	eager_recv_request_states req_state;


//tinympi::verbs::impl::VerbsRecvRequest
struct VerbsEagerRecvRequest {
	VERBS_EAGERRECV_REQUEST_MEMBERS
};
static_assert(std::is_standard_layout<VerbsEagerRecvRequest>::value,
		"VerbsEagerRecvRequest must be POD!");

REQUESTS_CONVERTIBLE(VerbsEagerRecvRequest, VerbsRecvRequest)
REQUESTS_CONVERTIBLE(VerbsEagerRecvRequest, VerbsRequest)
REQUESTS_CONVERTIBLE(VerbsEagerRecvRequest, BaseRequest)

ASSERT_VERBS_RECVREQUEST_COMPATIBLE(VerbsEagerRecvRequest)

#define ASSERT_VERBS_EAGERRECVREQUEST_COMPATIBLE(TYPE) \
	ASSERT_VERBS_RECVREQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(VerbsRecvRequest, TYPE, req_state)

namespace verbsEagerRecvRequest {
	namespace impl {
		// converts a state into a string
		inline std::string to_string(eager_recv_request_states state) {
			// no lock required
			switch (state) {
			case eager_recv_request_states::WAITING_DATA:
				return "eager_recv_request_state::WAITING_DATA";
			case eager_recv_request_states::UNEXPECTED_REQUEST:
				return "eager_recv_request_state::UNEXPECTED_REQUEST";
			case eager_recv_request_states::DONE:
				return "eager_recv_request_state::DONE";
			}
			return std::string("??? ") + std::to_string(static_cast<int>(state));
		}

		// initialize a new request
		inline void init(VerbsEagerRecvRequest* req) {
			// no lock required
			verbsRecvRequest::init(TO_VerbsRecvRequest(req), verbs_request_types::VERBS_EAGERRECV);
		}

		/////////////////////////////
		// primary state changes
		// when we receive data for an existing request, we can copy the data to the right place
		inline int waitingdata_to_done(VerbsEagerRecvRequest* req, const EagerMessage* received_msg) {
			// no lock required
			// when we receive the data, we need to copy and terminate the request

			// check the size of the received data and update local size and tag
			TMPI_ASSERT_FATAL(req->buflen >= received_msg->buflen, "overflow error: "
					<< "recvd buffer has size " << received_msg->buflen
					<< ", expected size: " << req->buflen, __FILE__, __LINE__);
			req->buflen = received_msg->buflen;

			// update our tag (may be ANY_TAG)
			req->tag = received_msg->tag;

			// retrieve data buffer
			const void* dataptr = reinterpret_cast<const void*>(received_msg->buffer
					+ EAGER_MESSAGE_THRESH - received_msg->buflen);
			memcpy(req->buf, dataptr, req->buflen);

			// we have received all data
			req->req_state = eager_recv_request_states::DONE;
			using namespace tinympi::impl;
			baseRequest::request_completed(TO_BaseRequest(req));

			TMPI_LOG_LEVEL(2, "EagerRecvRequest", "WAITINGDATA -> DONE");
			return 1;
		}

		inline int recv_done(VerbsEagerRecvRequest*) {
			// everything is done - so we don't need to do anything
			return 1;
		}
	} // END NAMESPACE impl

	/////////////////////////////
	// transparent function

	// we have two flows:
	//   post to waiting queue --->  retrieve data
	//     ^ in base request

	//   retrieve data         -->   get request

	// this function is called when we already have a waiting request, but now a message has been
	// found in the receive slots. This request is a base request currently!
	inline int retrieved_eager_for_waiting_request(VerbsEagerRecvRequest* req,
			const EagerMessage* message, VerbsEndpoint* endpoint) {
		// LOCK REQUIRED FOR endpoint, e.g., recvd_piggyback_message (OK)
		using namespace impl;

		// we need to cast the base request to an eager recv request
		req->provider_type = provider_types::node2node;
		req->verbs_request_type = verbs_request_types::VERBS_EAGERRECV;
		req->endpoint = endpoint;
		req->rid = message->rid;

		req->req_state = eager_recv_request_states::WAITING_DATA;

		// update peer rank
		req->peer_rank = message->src_id;

		// if the message has some piggyback entries, we free the indicated entries
		endpoint->eager.recvd_piggyback_message(message->piggyback_cnt);

		// then we retrieve the data
		return waitingdata_to_done(req, message);
	}
	// this function is called whenever we poll a work completion event from the completion queue.
	// this should not happen for this request type
	inline int handle_poll(VerbsEagerRecvRequest* const, const ibv_wc&) {
		// if this function is needed at some point, enable it in verbs.h (handle_wc)
		using namespace impl;

		TMPI_FATAL_ERROR("no handle_poll should be called for eager recv requests!",
				__FILE__, __LINE__);
		return 1;
	}

	// when we retrieve an unexpected eager message, we generate an eager recv request from this
	// message, which is then added to the unexpected queue
	inline VerbsEagerRecvRequest* retrieved_unexpected_eager(int message_pos,
			const EagerMessage* message, VerbsEndpoint* endpoint) {
		// LOCK REQUIRED FOR endpoint, recvd_piggyback_message (OK)
		using namespace impl;

		// if the message has some piggyback entries, we free the indicated entries
		endpoint->eager.recvd_piggyback_message(message->piggyback_cnt);

		// try to get a request, otherwise allocate a new one
		VerbsEagerRecvRequest* req = TO_VerbsEagerRecvRequest(baseRequest::malloc());
		init(req);

		INIT_BASE_REQUEST(
			req,
			message->src_id,
			message->context_id,
			message->tag,
			nullptr,
			message_pos, // we initialize the size of the request with the message position
			nullptr
		)

		req->endpoint = endpoint;
		req->rid = message->rid;

		req->req_state = eager_recv_request_states::UNEXPECTED_REQUEST;
		return req;
	}

	// a request has been posted for an eager message. We already know that req is an eager!
	inline int posted_request_for_unexpected_eager(VerbsEagerRecvRequest* req, void* buf,
			buflen_t buflen) {
		// LOCK REQUIRED FOR endpoint, for eager_free_recv_slot, test_explicit_piggyback (OK)
		using namespace impl;

		// we restore our message from the buffer
		int message_pos = req->buflen;
		const EagerMessage* message = req->endpoint->eager.eager_get_recv_slot(message_pos);

		// set the request to the size given by the local recv (overflow test)
		req->buf = buf;
		req->buflen = buflen;
		req->waiting_task = tasks::get_current();

		// we are no longer in the UNEXPECTED_REQUEST state
		req->req_state = eager_recv_request_states::WAITING_DATA;

		waitingdata_to_done(req, message);
		req->endpoint->eager.eager_free_recv_slot(message_pos);

		// when we receive a message, we probably need to send an explicit piggyback
		if (req->endpoint->eager.test_explicit_piggyback()) {
			std::lock_guard<std::mutex> ep_lock(req->endpoint->lock);
			if (req->endpoint->eager.test_explicit_piggyback()) {
				send_explicit_piggyback(req->endpoint);
			}
		}

		return TMPI_SUCCESS;
	}

} // END NAMESPACE verbsEagerRecvRequest
} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif //_VERBS_EAGERRECVREQUEST_H_
