// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_EAGERSENDREQUEST_H_635106f3226c46d3d114158785752466
#define _VERBS_EAGERSENDREQUEST_H_635106f3226c46d3d114158785752466

#include <cstring>
#include <tuple>

#include "utility/error.h"
#include "communicator.h"

#include "verbs_endpoint.h"
#include "verbs_request.h"
#include "verbs_provider.h"
#include "verbs_config.h"
#include "verbs_global.h"

#include "verbs_sendrequest.h"

namespace tinympi {
namespace verbs {
namespace impl {

//tinympi::verbs::impl::send_request_state
enum class eager_send_request_states {
	INITIALIZED,
	SENDINGDATA,
	DONE
};

#define VERBS_EAGERSEND_REQUEST_MEMBERS \
	VERBS_SEND_REQUEST_MEMBERS \
	eager_send_request_states req_state; \
	int send_slot_pos; \
	int rem_message_pos;

//tinympi::verbs::impl::VerbsEagerSendRequest
struct VerbsEagerSendRequest {
	VERBS_EAGERSEND_REQUEST_MEMBERS
};
static_assert(std::is_standard_layout<VerbsEagerSendRequest>::value,
		"VerbsEagerSendRequest must be POD!");

REQUESTS_CONVERTIBLE(VerbsEagerSendRequest, VerbsSendRequest)
REQUESTS_CONVERTIBLE(VerbsEagerSendRequest, VerbsRequest)
REQUESTS_CONVERTIBLE(VerbsEagerSendRequest, BaseRequest)

ASSERT_VERBS_SENDREQUEST_COMPATIBLE(VerbsEagerSendRequest)

#define ASSERT_VERBS_EAGERSENDREQUEST_COMPATIBLE(TYPE) \
	ASSERT_VERBS_SENDREQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(VerbsEagerSendRequest, TYPE, req_state) \
	VERIFY_COMPATIBILITY(VerbsEagerSendRequest, TYPE, send_slot_pos) \
	VERIFY_COMPATIBILITY(VerbsEagerSendRequest, TYPE, rem_message_pos)

namespace verbsEagerSendRequest {
	namespace impl {
		// converts a state into a string
		inline std::string to_string(eager_send_request_states state) {
			// no lock required
			switch (state) {
			case eager_send_request_states::INITIALIZED:
				return "eager_send_request_states::INITIALIZED";
			case eager_send_request_states::SENDINGDATA:
				return "eager_send_request_states::SENDINGDATA";
			case eager_send_request_states::DONE:
				return "eager_send_request_states::DONE";
			default:
				return std::string("??? ") + std::to_string(static_cast<int>(state));
			}
		}
		// initialize a new request
		inline void init(VerbsEagerSendRequest* req) {
			// no lock required
			req->req_state = eager_send_request_states::INITIALIZED;
			verbsSendRequest::init(TO_VerbsSendRequest(req), verbs_request_types::VERBS_EAGERSEND);
		}

		/////////////////////////////
		// primary state changes
		inline int initialized_to_sendingdata(VerbsEagerSendRequest* req, EagerMessage* loc_message) {
			// LOCK REQUIRED for endpoint, for eager_reserve_remote_slot, get_piggyback_message
			int res;

			// if we found send and receive entries, we can send the data

			// we fill the message
			loc_message->buflen = eager_size_t(req->buflen);
			loc_message->src_id = tasks::get_globalunique_id();
			loc_message->dest_id  = tasks::get_threadunique_id(req->peer_rank);
			loc_message->context_id = req->context_id;
			loc_message->rid = req->rid;
			loc_message->tag = req->tag;

			// the data is copied to the end of the buffer
			void* dataptr = loc_message->buffer + EAGER_MESSAGE_THRESH - req->buflen;
			memcpy(dataptr, req->buf, req->buflen);

			// we get a piggyback message
			std::unique_lock<std::mutex> ep_lock(req->endpoint->lock);
			loc_message->piggyback_cnt = req->endpoint->eager.get_piggyback_message();
			ep_lock.unlock();

			// we send the part of the eagermessage that is filled
			ibv_sge sr_sg_lst = {
				reinterpret_cast<uint64_t>(dataptr),                                 // addr
				static_cast<uint32_t>(sizeof(EagerMessage) - EAGER_MESSAGE_THRESH + req->buflen),
						// length
				VerbsProvider::Eager::eager_get_send_slots_lkey()                 // lkey
			};

			// prepare eager send request
			ibv_send_wr sr_wr;
			sr_wr.wr_id = reinterpret_cast<uint64_t>(req);
			sr_wr.opcode = IBV_WR_RDMA_WRITE; // we write directly into rdma buffer
			if(sr_sg_lst.length < provider_.get_max_inline_data())
				// if the data is small enough, then send inline
				// retrieve a signal when sent
				sr_wr.send_flags = IBV_SEND_INLINE | IBV_SEND_SIGNALED;
			else
				// retrieve a signal when sent
				sr_wr.send_flags = IBV_SEND_SIGNALED;

			sr_wr.sg_list = &sr_sg_lst;
			sr_wr.num_sge = 1; // only one message
			sr_wr.next = nullptr;

			// set request state before post_send, such that no other state change can occur now
			req->req_state = eager_send_request_states::SENDINGDATA;

			// set remote data with peer info retrieved from channel
			const PeerInfo& pi = req->endpoint->eager.eager_get_recv_slots_peer_info();
			sr_wr.wr.rdma.rkey = pi.rkey;
			sr_wr.wr.rdma.remote_addr = reinterpret_cast<uint64_t>(pi.addr)
					+ req->rem_message_pos*sizeof(EagerMessage) + EAGER_MESSAGE_THRESH - req->buflen;
			// now we can submit the message to the channel
			ibv_send_wr* bad_wr;
			res = ibv_post_send(req->endpoint->get_qp(), &sr_wr, &bad_wr);
			TMPI_ASSERT_FATAL(res == 0, "ibv_post_send returned " << res, __FILE__, __LINE__);

			// set new state
			TMPI_LOG_LEVEL(2, "EagerSendRequest " << req, "INITIALIZED -> SENDINGDATA");

			return 0;
		}

		// when we read the completion event, the request is done
		inline int sendingdata_to_senddone(VerbsEagerSendRequest* req) {
			// no lock required

			// we can free our send slot, the remote slot will be freed by piggyback messages
			VerbsProvider::Eager::eager_free_send_slot(req->send_slot_pos);

			req->req_state = eager_send_request_states::DONE;

			using namespace tinympi::impl;
			baseRequest::request_completed(TO_BaseRequest(req));

			TMPI_LOG_LEVEL(2, "EagerSendRequest", "SENDING_DATA -> DONE");

			return 1;
		}

	} // END NAMESPACE impl

	/////////////////////////////
	// transparent function

	// this function is called whenever we poll a work completion event from the completion queue.
	// this is only in the event that we have sent the data
	inline int handle_poll(VerbsEagerSendRequest* const req, const ibv_wc&) {
		// no lock required
		using namespace impl;
		TMPI_DASSERT_FATAL(req->req_state == eager_send_request_states::SENDINGDATA,
				"unexpected request state (" << to_string(req->req_state) << ") for request "
				<< req << " at this place!", __FILE__, __LINE__);

		return sendingdata_to_senddone(req);
	}

	// generate a new send request
	inline VerbsEagerSendRequest* generate(void* buf, buflen_t buflen, tag_t tag,
			context_id_t context_id, const tasks::Task& dest_descr, VerbsEndpoint* endpoint) {
		// LOCK REQUIRED FOR endpoint, for eager_has_remote_slot, initialized_to_sendingdata (OK)
		using namespace impl;

		// check if there are free slots available
		if (!endpoint->eager.eager_has_remote_slot()) {
			TMPI_LOG_LEVEL(3, "VerbsEagerSendRequest", "slots full: switch to rdv");
			return nullptr;
		}

		// try to get a send slot
		int send_slot_pos;
		EagerMessage* loc_message;
		std::tie(send_slot_pos, loc_message) = VerbsProvider::Eager::eager_get_send_slot();
		if (loc_message == nullptr) {
			TMPI_LOG_LEVEL(3, "VerbsEagerSendRequest", "slots full: switch to rdv");
			return nullptr;
		}

		// if we were successful, now try to get a remote slot
		std::unique_lock<std::mutex> ep_lock(endpoint->lock);
		int rem_message_pos = endpoint->eager.eager_reserve_remote_slot();
		if (rem_message_pos == -1) {
			ep_lock.unlock();
			// this may fail because we did not lock before when testing
			VerbsProvider::Eager::eager_free_send_slot(send_slot_pos);
			TMPI_LOG_LEVEL(3, "VerbsEagerSendRequest", "slots full: switch to rdv");
			return nullptr;
		}
		rid_t rid = endpoint->get_send_rid();
		ep_lock.unlock();

		// try to get a request, otherwise allocate a new one
		VerbsEagerSendRequest* req = TO_VerbsEagerSendRequest(baseRequest::malloc());
		req->send_slot_pos = send_slot_pos;

		// fill data
		INIT_BASE_REQUEST( \
			req, \
			dest_descr.get_identifier().global_identifier, \
			context_id, \
			tag, \
			buf, \
			buflen, \
			tasks::get_current() \
		)

		req->endpoint = endpoint;
		// update the request with the information we already retrieved
		req->rid = rid;
		req->rem_message_pos = rem_message_pos;

		req->req_state = eager_send_request_states::INITIALIZED;

		init(req);
		// since we also have the data, we just send it
		initialized_to_sendingdata(req, loc_message);

		return req;
	}


} // END NAMESPACE verbsEagerSendRequest
} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif //_VERBS_EAGERSENDREQUEST_H_
