// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_MRALLOCATOR_H_635106f3226c46d3d114158785752466
#define _VERBS_MRALLOCATOR_H_635106f3226c46d3d114158785752466
// adapted from https://gist.github.com/donny-dont/1471329

#include <dlfcn.h>
#include <malloc_overwrite.h>

namespace tinympi {

template <typename T>
class mr_allocator {
public:
	using pointer = T*;
	using const_pointer = const T*;
	using reference = T&;
	using const_reference = T&;
	using value_type = T;
	using size_type = std::size_t;
	using difference_type = ptrdiff_t;

	T * address(T& r) const {
		return &r;
	}

	const T * address(const T& s) const {
		return &s;
	}

	std::size_t max_size() const {
		return (static_cast<std::size_t>(0) - static_cast<std::size_t>(1)) / sizeof(T);
	}

	template <typename U>
	struct rebind {
		using other = mr_allocator<U>;
	} ;

	bool operator!=(const mr_allocator& other) const {
		return !(*this == other);
	}

	void construct(T * const p, const T& t) const {
		void * const pv = static_cast<void *>(p);
		new (pv) T(t);
	}

	void destroy(T * const p) const {
		p->~T();
	}

	bool operator==(const mr_allocator&) const {
		return true;
	}

	mr_allocator() { }
	mr_allocator(const mr_allocator&) { }
	template <typename U> mr_allocator(const mr_allocator<U>&) { }
	~mr_allocator() { }


	T* allocate(const std::size_t n) const {
		// All allocators should contain an integer overflow check.
		// The Standardization Committee recommends that std::length_error
		// be thrown in the case of integer overflow.
		if (n > max_size()) {
			throw std::length_error("mr_allocator<T>::allocate() - Integer overflow.");
		}

		void* pv = malloc_(n * sizeof(T));

		// Allocators should throw std::bad_alloc in the case of memory allocation failure.
		if (pv == NULL) {
			throw std::bad_alloc();
		}

		return static_cast<T*>(pv);
	}

	void deallocate(T * const p, const std::size_t) const {
		free(p);
	}


	template <typename U>
	T * allocate(const std::size_t n, const U * /* const hint */) const {
		return allocate(n);
	}


	// Allocators are not required to be assignable, so
	// all allocators should have a private unimplemented
	// assignment operator. Note that this will trigger the
	// off-by-default (enabled under /Wall) warning C4626
	// "assignment operator could not be generated because a
	// base class assignment operator is inaccessible" within
	// the STL headers, but that warning is useless.
	mr_allocator& operator=(const mr_allocator&) = delete;

private:

};
} // END NAMESPACE tinympi
#endif //_VERBS_MRALLOCATOR_H_635106f3226c46d3d114158785752466
