// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_PROVIDER_H_635106f3226c46d3d114158785752466
#define _VERBS_PROVIDER_H_635106f3226c46d3d114158785752466

#include <stack>
#include <vector>
#include <memory>
#include <infiniband/verbs.h>
#include <chrono>

#include <malloc_overwrite.h>

#include "global.h"

#include "verbs_config.h"
#include "verbs_endpoint.h"
#include "verbs_rdvprovider.h"
#include "verbs_eagerprovider.h"
#include "verbs_mrwrapper.h"
#include "verbs_mrallocator.h"

#include "debug/debugger.h"

namespace tinympi {
namespace verbs {
namespace impl {

class RdvRemoteMessage;
class EagerMessage;
class VerbsRecvRequest;

//tinympi::verbs::impl::VerbsProvider
class VerbsProvider {
public:
	VerbsProvider() : last_fast_progress_(std::chrono::high_resolution_clock::now()) {
	}
	void init();
	void post_init();
	void finalize();

	using Rdv = RdvProvider;
	using Eager = EagerProvider;

private:
	ibv_context* context_;
	ibv_pd* protection_domain_;

	// map for registered memory regions
	std::mutex memory_regions_lock_;
	std::map<
			uint64_t,
			std::shared_ptr<mr_wrapper>,
			std::less<uint64_t>,
			mr_allocator<std::pair<const uint64_t, std::shared_ptr<mr_wrapper> > >
		> memory_regions_;

	const size_t max_inline_data_ = MAX_INLINE_DATA;

	std::atomic<std::chrono::time_point<std::chrono::high_resolution_clock>> last_fast_progress_;

	// as ibv_reg_mr and ibv_dereg_mr call malloc, we need to break the recursion
	static thread_local bool recursive_ibvregmr_call;

public:
	size_t get_max_inline_data() const { return max_inline_data_; }
	ibv_pd* get_pd() const { return protection_domain_; }

	//////////////////////////////
	std::shared_ptr<mr_wrapper> register_memory(void* ptr, size_t sz);
	void deregister_memory(ibv_mr* mr);
	void malloc_hook(void* ptr, size_t sz);

	//////////////////////////////
	bool countdown() {
// 		return true;

		// turned out not to improve performance currently TODO
		using namespace std::chrono;
		auto now = high_resolution_clock::now();
		auto last = last_fast_progress_.load();
		constexpr uint64_t limit = 1500; // parameter
		if (duration_cast<nanoseconds>(now - last) > nanoseconds(limit)) {
			if (last_fast_progress_.compare_exchange_strong(last, now))
				return true;
		}
		return false;
	}

////////////
// STATIC
public:
	static void setup_endpoints();
	static void destroy_endpoints();
private:
	static void pair_endpoints(VerbsEndpoint*, uint16_t, uint32_t,
			uint32_t*, uint64_t*, uint32_t*, uint64_t*);
};

////////////
// malloc_hook is called whenever malloc or memalign are called. It is not called with
// malloc_/memalign_
inline void VerbsProvider::malloc_hook(void* ptr, size_t sz) {
	// in this case we for sure don't need to degister memory (we are just registering ...)
	if (recursive_ibvregmr_call)
		return;

	// when we call malloc, we need to deregister the memory if it was already registered
	using memory_regions_t = decltype(memory_regions_);
	std::lock_guard<std::mutex> lk(memory_regions_lock_);

	auto get_end = [](const memory_regions_t::value_type& v) {
		return v.first + v.second->get_sz();
	};

	uint64_t start = reinterpret_cast<uint64_t>(ptr);
	uint64_t end = start + sz;

	// finds the first element that starts after my start
	auto lower_it = memory_regions_.upper_bound(start);
	if (lower_it != memory_regions_.begin()) {
		lower_it = std::prev(lower_it);
		// now we have the first element which starts at or before my start

		// we then need to check whether it ends at or after my end
		size_t lower_end = get_end(*lower_it);
		if (lower_end >= end) {
			// remove this memory region
			bool tmp = recursive_ibvregmr_call;
			recursive_ibvregmr_call = true;
			memory_regions_.erase(lower_it);
			recursive_ibvregmr_call = tmp;
			return;
		}

		// if the lower interval ends at or after my start, it is the first we delete, otherwise
		// we keep it the way it is
		// and, whether it ends at or after my start, in which case we combine the intervals
		if (lower_end < start) {
			// otherwise we modify lower_it such that it is again the first elemnt which intersects
			lower_it = std::next(lower_it);
		}
	}

	// finds the first element that starts after my end
	auto upper_it = memory_regions_.upper_bound(end);
	if (lower_it != memory_regions_.end()) {
		bool tmp = recursive_ibvregmr_call;
		recursive_ibvregmr_call = true;
		memory_regions_.erase(lower_it, upper_it);
		recursive_ibvregmr_call = tmp;
	}
}

////////////
// check if memory is already registered, and if not then register new block of memory
inline std::shared_ptr<mr_wrapper> VerbsProvider::register_memory(void* ptr, size_t sz) {
// if it breaks ...
// 	return std::move(std::make_shared<mr_wrapper>(protection_domain_,
// 			reinterpret_cast<uint64_t>(ptr), sz));
	TMPI_LOG_LEVEL(6, "register_memory", "find memory " << ptr << " with size = " << sz);
	using memory_regions_t = decltype(memory_regions_);

	uint64_t start = reinterpret_cast<uint64_t>(ptr);
	std::lock_guard<std::mutex> lk(memory_regions_lock_);

	auto get_end = [](const memory_regions_t::value_type& v) {
		return v.first + v.second->get_sz();
	};

	size_t end = start + sz;

	// finds the first element that starts after my start
	auto lower_it = memory_regions_.upper_bound(start);
	if (lower_it != memory_regions_.begin()) {
		lower_it = std::prev(lower_it);
		// now we have the first element which starts at or before my start

		// we then need to check whether it ends at or after my end
		size_t lower_end = get_end(*lower_it);
		if (lower_end >= end) {
			return lower_it->second;
		}

		// and, whether it ends at or after my start, in which case we combine the intervals
		if (lower_end >= start) {
			start = lower_it->first;
		} else {
			// otherwise we modify lower_it such that it is again the first elemnt which intersects
			lower_it = std::next(lower_it);
		}
	}

	// finds the first element that starts after my end
	auto upper_it = memory_regions_.upper_bound(end);
	if (upper_it != memory_regions_.begin()) {
		upper_it = std::prev(upper_it);
		// now we have the last element which starts at or before my end

		// we need to check whether it ends after my end, in which we combine the intervals
		size_t upper_end = get_end(*upper_it);
		if (upper_end >= end) {
			end = upper_end;
		}

		// we then set upper_it back to the first element which does not intersect anymore
		upper_it = std::next(upper_it);
	}
	if (lower_it != memory_regions_.end()) {
		bool tmp = recursive_ibvregmr_call;
		recursive_ibvregmr_call = true;
		memory_regions_.erase(lower_it, upper_it);
		recursive_ibvregmr_call = tmp;
	}

	// register the memory (note that we need to make sure not to deadlock with the
	// memory_regions_lock)
	mr_allocator<mr_wrapper> mra;
	bool tmp = recursive_ibvregmr_call;
	recursive_ibvregmr_call = true;
	std::shared_ptr<mr_wrapper> mrw = std::allocate_shared<mr_wrapper>(mra, protection_domain_,
			start, end-start);
	recursive_ibvregmr_call = tmp;

	// and insert for later reuse
	memory_regions_.insert(memory_regions_t::value_type{start, mrw});
	return std::move(mrw);
}
inline void VerbsProvider::deregister_memory(ibv_mr* mr) {
	bool tmp = recursive_ibvregmr_call;
	recursive_ibvregmr_call = true;
	ibv_dereg_mr(mr);
	recursive_ibvregmr_call = tmp;
}


extern VerbsProvider provider_;
extern VerbsEndpoint** endpoints_;
extern global_id_t num_endpoints_;

inline VerbsEndpoint* get_endpoint(node_id_t node_id) {
	unsigned ep_num = tasks::get_current()->get_identifier().local.thread_id * ENDPOINTSPERNODE
			/ env::num_threads;

	return endpoints_[ep_num * num_endpoints_/ENDPOINTSPERNODE + node_id];
}

} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif // _VERBS_PROVIDER_H
