// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_QUEUES_H_635106f3226c46d3d114158785752466
#define _VERBS_QUEUES_H_635106f3226c46d3d114158785752466

#include <mutex>

#include "tasks.h"
#include "threads.h"
#include "verbs_request.h"

#include "provider_impl/queue_functions.h"

namespace tinympi {
namespace verbs {
namespace impl {
namespace recvqueues {
	using namespace tinympi::impl::recvqueues;

	extern std::mutex queue_mutex_;

	////////////////////////////
	// FIND REQUEST
	// finds a request in the unexpected queue
	template<typename dequeue_request_>
	inline VerbsRequest* find_unexpected_request(const global_id_t& source, tag_t tag,
			context_id_t context_id, VerbsRequest*& prev) {
		// LOCK REQUIRED FOR node queue (OK)

		return tinympi::impl::recvqueues::find_unexpected_request(source, tag, context_id,
			tasks::get_current()->verbs_unexpected_queue_head_,
			tasks::get_current()->verbs_unexpected_queue_tail_, prev, dequeue_request_());
	}

	// finds a request in the unexpected queue
	template<typename dequeue_request_>
	inline VerbsRequest* find_unexpected_anysource_request(tag_t tag, context_id_t context_id,
			VerbsRequest*& prev) {
		// LOCK REQUIRED FOR node queue (OK)

		return tinympi::impl::recvqueues::find_unexpected_anysource_request(tag, context_id,
			tasks::get_current()->verbs_unexpected_queue_head_,
			tasks::get_current()->verbs_unexpected_queue_tail_, prev, dequeue_request_());
	}

	// finds a request in the waiting queue
	template<typename dequeue_request_>
	inline VerbsRequest* find_waiting_request(const global_id_t& source, tag_t tag,
			context_id_t context_id, tasks::Task& dest_task) {
		// LOCK REQUIRED FOR node queue (OK)

		tasks::LocalTask& ltask = static_cast<tasks::LocalTask&>(dest_task);
		VerbsRequest* prev;
		return tinympi::impl::recvqueues::find_waiting_request(source, tag, context_id,
			ltask.verbs_waiting_queue_head_, ltask.verbs_waiting_queue_tail_, prev,
			dequeue_request_());
	}

	////////////////////////////
	// ADD REQUEST
	inline void add_to_waiting_queue(VerbsRequest* req) {
		// LOCK REQUIRED FOR node queue (OK)
		add_to_queue(req, tasks::get_current()->verbs_waiting_queue_head_,
				tasks::get_current()->verbs_waiting_queue_tail_);
	}
	// add to an unexpected queue of another thread
	inline void add_to_unexpected_queue(VerbsRequest* req, tasks::Task& dest_task) {
		// LOCK REQUIRED FOR node queue (OK)
		tasks::LocalTask& ltask = static_cast<tasks::LocalTask&>(dest_task);
		add_to_queue(req, ltask.verbs_unexpected_queue_head_, ltask.verbs_unexpected_queue_tail_);
	}

	////////////////////////////
	// REMOVE REQUEST
	inline bool remove_waiting_request(VerbsRequest* req) {
		// LOCK REQUIRED FOR node queue (OK)

		using namespace baseRequest::impl;
		TMPI_LOG_LEVEL(1, "remove from task", "req->waiting_task");
		return tinympi::impl::recvqueues::remove_waiting_request(req,
			req->waiting_task->verbs_waiting_queue_head_,
			req->waiting_task->verbs_waiting_queue_tail_);
	}

	////////////////////////////
	// DEQUEUE REQUEST
	inline void dequeue_unexpected_request(VerbsRequest* req, VerbsRequest*& prev) {
		// LOCK REQUIRED FOR verbs queue

		tasks::LocalTask& ltask = *tasks::get_current();

		// if prev is nullptr, we found a new head
		if (prev == nullptr) {
			ltask.verbs_unexpected_queue_head_ = TO_VerbsRequest(req->next_request);
		} else {
			// otherwise we are somewhere in between or at the tail
			prev->next_request = req->next_request;
			// the tail can be recognized because next_request is set to nullptr
			if (prev->next_request == nullptr) {
				ltask.verbs_unexpected_queue_tail_ = prev;
			}
		}
	}

} // END NAMESPACE recvqueues
} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif // _VERBS_QUEUES_H_
