// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_RDVENDPOINT_H_635106f3226c46d3d114158785752466
#define _VERBS_RDVENDPOINT_H_635106f3226c46d3d114158785752466

#include <malloc.h>

#include <atomic>

#include "verbs_rdvsreadymessage.h"
#include "verbs_rdvrreadymessage.h"
#include "verbs_global.h"

#include "verbs_peerinfo.h"

namespace tinympi {
namespace verbs {
namespace impl {

namespace rdv_channel_slot_status {
	enum rdv_channel_slot_status {
		free,
		reserved
	};
}

class VerbsRdvRecvRequest;

struct VerbsRdvEndpoint {
public:
	//////////////////////////////
	// initialization functions
	void setup(ibv_pd* pd);
	void destroy();

	//////////////////////////////
	// sready functions

	///////////
	// sready initialization functions
	void sready_set_peer_info(uint32_t, uint64_t, uint32_t, uint64_t);
	void sready_get_recv_slots_info(uint32_t&, uint64_t&) const;
	void sready_get_remote_info(uint32_t&, uint64_t&) const;

	///////////
	// sready remote slot functions
	int sready_reserve_remote_slot();
	void sready_free_remote_slot(int i);

	///////////
	// sready recv slot functions
	const RdvSReadyMessage* sready_get_recv_slot(int i) const;
	void sready_free_recv_slot(int i);
	bool sready_poll_and_lock_recv_slots(rid_t expected_rid, int& pos,
			std::unique_lock<std::mutex>&);
	const PeerInfo& sready_get_recv_slots_peer_info() const {
		return sready_recv_slots_peer_info_;
	}

	//////////////////////////////
	// rready functions

	///////////
	// rready initialization functions
	void rready_set_peer_info(uint32_t, uint64_t, uint32_t, uint64_t);
	void rready_get_recv_slots_info(uint32_t&, uint64_t&) const;
	void rready_get_remote_info(uint32_t&, uint64_t&) const;

	///////////
	// rready remote slot functions
	int rready_reserve_remote_slot();
	void rready_free_remote_slot(int i);

	///////////
	// rready recv slot functions
	bool rready_poll_recv_slots(int& pos);
	const RdvRReadyMessage* rready_get_recv_slot(int i) const;
	void rready_free_recv_slot(int i);
	const PeerInfo& rready_get_recv_slots_peer_info() const {
		return rready_recv_slots_peer_info_;
	}

	//////////////////////////////
	// waiting request functions
	recvreqid_t add_waitingdata_request(VerbsRdvRecvRequest* ptr);
	VerbsRdvRecvRequest* get_waitingdata_request(recvreqid_t i);

private:
	//////////////////////////////
	// SREADY MESSAGES
	///////////
	// sender
	// state of the remote recv slots
	sready_message_status* sready_remote_status_;
	// memory region of the remote status slots
	ibv_mr* sready_remote_status_mr_;
	// peer info of the remote slot (rkey / addr)
	PeerInfo sready_remote_status_peer_info_;
	// the next element where we can send to
	int sready_remote_head_;

	///////////
	// receiver
	// recv slots
	RdvSReadyMessage* sready_recv_slots_;
	// memory region of sready_recv_slots_
	ibv_mr* sready_recv_slots_mr_;
	// memory region of the remote recv slots
	PeerInfo sready_recv_slots_peer_info_;
	// where we recv the next element (used for polling)
	std::atomic<int> sready_recv_head_;
	// the last reserved sready
	std::atomic<int> sready_recv_tail_;

	//////////////////////////////
	// RREADY MESSAGES
	///////////
	// sender
	// state of the remote recv slots
	rready_message_status* rready_remote_status_;
	// memory region of the remote status slots
	ibv_mr* rready_remote_status_mr_;
	// peer info of the remote slot (rkey / addr)
	PeerInfo rready_remote_status_peer_info_;
	// the next element where we can send to
	int rready_remote_head_;

	///////////
	// receiver
	// recv slots
	RdvRReadyMessage* rready_recv_slots_;
	// memory region of rready_recv_slots_
	ibv_mr* rready_recv_slots_mr_;
	// memory region of the remote recv slots
	PeerInfo rready_recv_slots_peer_info_;
	// where we recv the next element (used for polling)
	std::atomic<int> rready_recv_head_;

	//////////////////////////////
	VerbsRdvRecvRequest** waiting_requests_;
	uint16_t waiting_requests_head_;
};

//////////////////////////////
// INITIALIZATION FUNCTIONS
inline void VerbsRdvEndpoint::setup(ibv_pd* pd) {

	//////////////////////////////
	// SREADY

	/////////
	// send
	// we have RDV_CHMESSAGE_LIMIT remote status slots
	sready_remote_status_ = (sready_message_status*) memalign(64,
			RDV_CHMESSAGE_LIMIT*sizeof(sready_message_status));
	for (unsigned i = 0; i < RDV_CHMESSAGE_LIMIT; ++i)
		sready_remote_status_[i] = sready_message_status::free;

	// register memory region
	sready_remote_status_mr_ = ibv_reg_mr(pd, sready_remote_status_,
			RDV_CHMESSAGE_LIMIT*sizeof(sready_message_status),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(sready_remote_status_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);

	// our first send goes to the first position
	sready_remote_head_ = 0;

	/////////
	// receive
	// we have RDV_CHMESSAGE_LIMIT recv slots
	sready_recv_slots_ = (RdvSReadyMessage*) memalign(64,
			RDV_CHMESSAGE_LIMIT*sizeof(RdvSReadyMessage));
	std::memset(sready_recv_slots_, 0, RDV_CHMESSAGE_LIMIT*sizeof(RdvSReadyMessage));
	for (unsigned i = 0; i < RDV_CHMESSAGE_LIMIT; ++i)
		sready_recv_slots_[i].status = sready_message_status::free;

	// register memory region
	sready_recv_slots_mr_ = ibv_reg_mr(pd, sready_recv_slots_,
			RDV_CHMESSAGE_LIMIT*sizeof(RdvSReadyMessage),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(sready_recv_slots_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);

	sready_recv_head_ = 0;
	sready_recv_tail_ = 0;

	//////////////////////////////
	// RREADY

	/////////
	// send
	// we have RDV_CHMESSAGE_LIMIT remote status slots
	rready_remote_status_ = (rready_message_status*) memalign(64,
			RDV_CHMESSAGE_LIMIT*sizeof(rready_message_status));
	for (unsigned i = 0; i < RDV_CHMESSAGE_LIMIT; ++i)
		rready_remote_status_[i] = rready_message_status::free;

	// register memory region
	rready_remote_status_mr_ = ibv_reg_mr(pd, rready_remote_status_,
			RDV_CHMESSAGE_LIMIT*sizeof(rready_message_status),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(rready_remote_status_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);

	// our first send goes to the first position
	rready_remote_head_ = 0;

	/////////
	// receive
	// we have RDV_CHMESSAGE_LIMIT recv slots
	rready_recv_slots_ = (RdvRReadyMessage*) memalign(64,
			RDV_CHMESSAGE_LIMIT*sizeof(RdvRReadyMessage));
	std::memset(rready_recv_slots_, 0, RDV_CHMESSAGE_LIMIT*sizeof(RdvRReadyMessage));
	for (unsigned i = 0; i < RDV_CHMESSAGE_LIMIT; ++i)
		rready_recv_slots_[i].status = rready_message_status::free;

	// register memory region
	rready_recv_slots_mr_ = ibv_reg_mr(pd, rready_recv_slots_,
			RDV_CHMESSAGE_LIMIT*sizeof(RdvRReadyMessage),
			IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_READ);
	TMPI_ASSERT_FATAL(rready_recv_slots_mr_ != nullptr, "ibv_reg_mr() returned 0", __FILE__,
			__LINE__);

	rready_recv_head_ = 0;

	//////////////////////////////
	// waiting requests
	waiting_requests_ = new VerbsRdvRecvRequest*[RDV_WAITINGREQUEST_LIMIT];
	for (unsigned i = 0; i < RDV_WAITINGREQUEST_LIMIT; ++i)
		waiting_requests_[i] = nullptr;
	waiting_requests_head_ = 0;

}
inline void VerbsRdvEndpoint::destroy() {
	int res;

	//////////////////////////////
	// SREADY

	/////////
	// send
	res = ibv_dereg_mr(sready_remote_status_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);

	free(sready_remote_status_);

	/////////
	// receive
	res = ibv_dereg_mr(sready_recv_slots_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);

	free(sready_recv_slots_);

	//////////////////////////////
	// RREADY

	/////////
	// send
	res = ibv_dereg_mr(rready_remote_status_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);

	free(rready_remote_status_);

	/////////
	// receive
	res = ibv_dereg_mr(rready_recv_slots_mr_);
	TMPI_ASSERT_FATAL(res == 0, "ibv_dereg_mr() returned " << res, __FILE__, __LINE__);

	free(rready_recv_slots_);

	//////////////////////////////
	// waiting requests
	delete[] waiting_requests_;
}
//////////////////////////////
// SREADY FUNCTIONS
///////////
// SREADY INITIALIZATION FUNCTIONS
inline void VerbsRdvEndpoint::sready_get_recv_slots_info(uint32_t& rkey, uint64_t& addr) const {
	rkey = sready_recv_slots_mr_->rkey;
	addr = reinterpret_cast<uint64_t>(sready_recv_slots_);
}

inline void VerbsRdvEndpoint::sready_get_remote_info(uint32_t& rkey, uint64_t& addr) const {
	rkey = sready_remote_status_mr_ ->rkey;
	addr = reinterpret_cast<uint64_t>(sready_remote_status_);
}
inline void VerbsRdvEndpoint::sready_set_peer_info(
		uint32_t sready_recv_rkey, uint64_t sready_recv_addr,
		uint32_t sready_remote_rkey, uint64_t sready_remote_addr) {
	sready_recv_slots_peer_info_.rkey    = sready_recv_rkey;
	sready_recv_slots_peer_info_.addr    = sready_recv_addr;
	sready_remote_status_peer_info_.rkey = sready_remote_rkey;
	sready_remote_status_peer_info_.addr = sready_remote_addr;
}
///////////
// SREADY REMOTE SLOT FUNCTIONS
inline int VerbsRdvEndpoint::sready_reserve_remote_slot() {
	TMPI_LOG_LEVEL(4,"VerbsRdvEndpoint", "reserve sready remote slot " << sready_remote_head_)
	TMPI_DASSERT_FATAL(sready_remote_status_[sready_remote_head_] == sready_message_status::free,
			"remote slot is not free. Increase RDV_CHMESSAGE_LIMIT", __FILE__, __LINE__);

	sready_remote_status_[sready_remote_head_] = sready_message_status::reserved;
	int res = sready_remote_head_;
	RING_INCREMENT(sready_remote_head_, RDV_CHMESSAGE_LIMIT);
	return res;
}
inline void VerbsRdvEndpoint::sready_free_remote_slot(int i) {
	TMPI_LOG_LEVEL(4,"VerbsRdvEndpoint", "free sready remote slot " << i)
	TMPI_DASSERT_FATAL(sready_remote_status_[i] != sready_message_status::free,
			"sready remote slot " << i << " double freed", __FILE__, __LINE__);

	sready_remote_status_[i] = sready_message_status::free;
}
///////////
// SREADY RECV SLOT FUNCTIONS
inline const RdvSReadyMessage* VerbsRdvEndpoint::sready_get_recv_slot(int i) const {
	return &sready_recv_slots_[i];
}
// frees a recv slot, i.e. sets it to tobefreed, if we are the tail, we can forward the tail
// and free the next recv slots
inline void VerbsRdvEndpoint::sready_free_recv_slot(int i) {
	TMPI_DASSERT_FATAL(sready_recv_slots_[i].status == sready_message_status::reserved,
			"double freed recv slot", __FILE__, __LINE__);

	sready_recv_slots_[i].status = sready_message_status::tobefreed;
	int tmp = sready_recv_tail_ & (RDV_CHMESSAGE_LIMIT - 1);
	if (i == tmp) {
		// if we are the last reserved message, we can continue the tail to the next reserved position
		// NOTE: we need to test status == tobefreed and not status != reserved because there may be
		// no reserved but only tobefreed messages
		// At this point, multiple threads may be present, but since all do the same, this is not a
		// problem

		sready_message_status cmp = sready_message_status::tobefreed;
		// TODO which is faster? readonly + read/write or read/write
		while (sready_recv_slots_[tmp].status == sready_message_status::tobefreed
				&& (bool)__atomic_compare_exchange_n(&sready_recv_slots_[tmp].status, &cmp,
					(uint8_t) sready_message_status::free, true, __ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST)) {
			// in each iteration, only one thread survives, because all others will get a tmp-value
			// which has already been done, e.g. one thread may get to here with tmp=3, then is set
			// to 4. Another thread may get here, setting tmp=4. The first thread will not survive
			// this iteration
			tmp = (sready_recv_tail_.fetch_add(1) + 1) & (RDV_CHMESSAGE_LIMIT - 1);
		}
	}
}
// polls the receive slots for a message with given request id
inline bool VerbsRdvEndpoint::sready_poll_and_lock_recv_slots(rid_t expected_rid, int& pos,
		std::unique_lock<std::mutex>& verbs_lock) {

	// get the current head with which we try to get the recv slot
	int tmp_head = sready_recv_head_ & (RDV_CHMESSAGE_LIMIT - 1);
	if (sready_recv_slots_[tmp_head].rid == expected_rid) {
		// if possible, abort without write operation
		if (sready_recv_slots_[tmp_head].status != sready_message_status::retrieved)
			return false;

		// if we have the right rid, we try to reserve it, if and only if the slot is retrieved, but
		// not reserved yet
		sready_message_status cmp = sready_message_status::retrieved;
		if ((bool)__atomic_compare_exchange_n(&sready_recv_slots_[tmp_head].status, &cmp,
				(uint8_t) sready_message_status::reserved, true, __ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST)) {

			// there may be a very unlikely case where rid was read, then we wait for an extremely
			// long time, and then we cas the status, which is already in the next iteration through
			// the array. therefore we should test again the rid, which - when not matching - is set
			// back to reserved. This does not break anything, because in this case all other threads
			// would just wait.
			if (sready_recv_slots_[tmp_head].rid != expected_rid) {
				sready_recv_slots_[tmp_head].status = cmp;
				return false;
			}

			// now the request is reserved and we can increase the head
			pos = sready_recv_head_.fetch_add(1);
			pos &= (RDV_CHMESSAGE_LIMIT - 1);

			// then lock the appropriate verbs queue
			tasks::Task& task = tasks::get_task(sready_recv_slots_[pos].dest_id);
			verbs_lock = std::unique_lock<std::mutex>(
					static_cast<tasks::LocalTask*>(&task)->verbs_lock_);

			TMPI_LOG_LEVEL(4,"VerbsRdvEndpoint", "successful polling sready at position " << pos);
			return true;

		} else {
			return false;
		}

	} else {
		return false;
	}
}

//////////////////////////////
// RREADY FUNCTIONS
///////////
// RREADY INITIALIZATION FUNCTIONS
inline void VerbsRdvEndpoint::rready_get_recv_slots_info(uint32_t& rkey, uint64_t& addr) const {
	rkey = rready_recv_slots_mr_->rkey;
	addr = reinterpret_cast<uint64_t>(rready_recv_slots_);
}

inline void VerbsRdvEndpoint::rready_get_remote_info(uint32_t& rkey, uint64_t& addr) const {
	rkey = rready_remote_status_mr_ ->rkey;
	addr = reinterpret_cast<uint64_t>(rready_remote_status_);
}
inline void VerbsRdvEndpoint::rready_set_peer_info(
		uint32_t rready_recv_rkey, uint64_t rready_recv_addr,
		uint32_t rready_remote_rkey, uint64_t rready_remote_addr) {
	rready_recv_slots_peer_info_.rkey    = rready_recv_rkey;
	rready_recv_slots_peer_info_.addr    = rready_recv_addr;
	rready_remote_status_peer_info_.rkey = rready_remote_rkey;
	rready_remote_status_peer_info_.addr = rready_remote_addr;
}
///////////
// RREADY REMOTE SLOT FUNCTIONS
inline int VerbsRdvEndpoint::rready_reserve_remote_slot() {
	TMPI_LOG_LEVEL(4,"VerbsRdvEndpoint", "reserve rready remote slot " << rready_remote_head_)
	TMPI_DASSERT_FATAL(rready_remote_status_[rready_remote_head_] == rready_message_status::free,
			"remote slot is not free", __FILE__, __LINE__);

	rready_remote_status_[rready_remote_head_] = rready_message_status::reserved;
	int res = rready_remote_head_;
	RING_INCREMENT(rready_remote_head_, RDV_CHMESSAGE_LIMIT);
	return res;
}
inline void VerbsRdvEndpoint::rready_free_remote_slot(int i) {
	TMPI_LOG_LEVEL(4,"VerbsRdvEndpoint", "free rready remote slot " << i)
	TMPI_DASSERT_FATAL(rready_remote_status_[i] != rready_message_status::free,
			"rready remote slot " << i << " double freed", __FILE__, __LINE__);

	rready_remote_status_[i] = rready_message_status::free;
}
///////////
// RREADY RECV SLOT FUNCTIONS
inline bool VerbsRdvEndpoint::rready_poll_recv_slots(int& pos) {
	// get the current head
	int tmp_head = rready_recv_head_ & (RDV_CHMESSAGE_LIMIT - 1);

	// if possible, abort without write operation
	if (rready_recv_slots_[tmp_head].status != rready_message_status::retrieved)
		return false;

	rready_message_status cmp = rready_message_status::retrieved;
	if ((bool)__atomic_compare_exchange_n(&rready_recv_slots_[tmp_head].status, &cmp,
			(uint8_t) rready_message_status::reserved, true, __ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST)) {

		pos = rready_recv_head_.fetch_add(1);
		pos &= (RDV_CHMESSAGE_LIMIT - 1);

		TMPI_LOG_LEVEL(4,"VerbsRdvEndpoint", "successful polling rready at position " << pos);

		return true;

	} else {
		return false;
	}
}
// freeing rready slots is trivial
inline void VerbsRdvEndpoint::rready_free_recv_slot(int i) {
	TMPI_DASSERT_FATAL(rready_recv_slots_[i].status == rready_message_status::reserved,
			"double freed recv slot", __FILE__, __LINE__);

	rready_recv_slots_[i].status = rready_message_status::free;
}
inline const RdvRReadyMessage* VerbsRdvEndpoint::rready_get_recv_slot(int i) const {
	return &rready_recv_slots_[i];
}

//////////////////////////////
// WAITING REQUEST FUNCTIONS
// adding a waiting request is just adding it to the head
inline recvreqid_t VerbsRdvEndpoint::add_waitingdata_request(VerbsRdvRecvRequest* const ptr) {

	while (waiting_requests_[waiting_requests_head_] != nullptr) {
		RING_INCREMENT(waiting_requests_head_, RDV_WAITINGREQUEST_LIMIT);
	}
	recvreqid_t tmp = waiting_requests_head_;
	waiting_requests_[tmp] = ptr;

	RING_INCREMENT(waiting_requests_head_, RDV_WAITINGREQUEST_LIMIT);

	// the identifier must be > 0
	return tmp+1;
}
inline VerbsRdvRecvRequest* VerbsRdvEndpoint::get_waitingdata_request(recvreqid_t i) {
	TMPI_DASSERT_FATAL(waiting_requests_[i-1] != nullptr,
			"waiting_request does not exist", __FILE__, __LINE__);

	VerbsRdvRecvRequest* tmp = waiting_requests_[i-1];
	waiting_requests_[i-1] = nullptr;
	return tmp;
}

} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif //_VERBS_RDVENDPOINT_H_
