// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _VERBS_RDVSENDREQUEST_H_635106f3226c46d3d114158785752466
#define _VERBS_RDVSENDREQUEST_H_635106f3226c46d3d114158785752466

#include <cstring>
#include <memory>

#include "utility/error.h"

#include "verbs_endpoint.h"
#include "verbs_request.h"
#include "verbs_provider.h"
#include "verbs_config.h"
#include "verbs_global.h"
#include "verbs_sendrequest.h"
#include "verbs_mrwrapper.h"

namespace tinympi {
namespace verbs {
namespace impl {

//tinympi::verbs::impl::send_request_state
enum class rdv_send_request_states : uint_fast8_t {
	INITIALIZED,
	SENDING_SREADY,
	SENDING_DATA,
	DONE // = SENT_SREADY_SENT_DATA
};

#define VERBS_RDVSEND_REQUEST_MEMBERS \
	VERBS_SEND_REQUEST_MEMBERS           /* 76 */ \
	std::shared_ptr<mr_wrapper> mr;      /*  8 */ \
	rdv_send_request_states req_state;   /* >1 */ \
	std::atomic<uint8_t> req_state_cntr; /*  1 */ \
	int send_sready_slot_pos;            /* 4-8*/ \
	int rem_sready_message_pos;          /* 4-8*/

//tinympi::verbs::impl::VerbsRdvSendRequest
struct VerbsRdvSendRequest {
	VERBS_RDVSEND_REQUEST_MEMBERS
};
static_assert(std::is_standard_layout<VerbsRdvSendRequest>::value,
		"VerbsRdvSendRequest must be POD!");

REQUESTS_CONVERTIBLE(VerbsRdvSendRequest, VerbsSendRequest)
REQUESTS_CONVERTIBLE(VerbsRdvSendRequest, VerbsRequest)
REQUESTS_CONVERTIBLE(VerbsRdvSendRequest, BaseRequest)

ASSERT_VERBS_SENDREQUEST_COMPATIBLE(VerbsRdvSendRequest)

#define ASSERT_VERBS_RDVSENDREQUEST_COMPATIBLE(TYPE) \
	ASSERT_VERBS_SENDREQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(VerbsRdvSendRequest, TYPE, req_state) \
	VERIFY_COMPATIBILITY(VerbsRdvSendRequest, TYPE, req_state_cntr) \
	VERIFY_COMPATIBILITY(VerbsRdvSendRequest, TYPE, mr) \
	VERIFY_COMPATIBILITY(VerbsRdvSendRequest, TYPE, send_sready_slot_pos) \
	VERIFY_COMPATIBILITY(VerbsRdvSendRequest, TYPE, rem_sready_message_pos)

namespace verbsRdvSendRequest {
	namespace impl {
		// converts a state into a string
		inline std::string to_string(rdv_send_request_states state) {
			switch (state) {
			case rdv_send_request_states::INITIALIZED:
				return "rdv_send_request_states::INITIALIZED";
			case rdv_send_request_states::SENDING_SREADY:
				return "rdv_send_request_states::SENDING_SREADY";
			case rdv_send_request_states::SENDING_DATA:
				return "rdv_send_request_states::SENDING_DATA";
			case rdv_send_request_states::DONE:
				return "rdv_send_request_states::DONE";
			}
			return std::string("UNKNOWN STATE: ") + std::to_string(static_cast<int>(state));
		}
		// initialize a new request
		inline void init(VerbsRdvSendRequest* req) {
			// no lock required
			////  req->req_state = rdv_send_request_states::INITIALIZED;
			std::memset(&req->mr, 0, sizeof(req->mr));
			verbsSendRequest::init(TO_VerbsSendRequest(req), verbs_request_types::VERBS_RDVSEND);
		}


		/////////////////////////////
		// secondary state changes
		inline int sendingsready_to_sentsready(VerbsRdvSendRequest* req) {
			// no lock required
			// if we have sent the sready, we can free the send slot. The remote slot is freed
			// when we receive the ready
			VerbsProvider::Rdv::sready_free_send_slot(req->send_sready_slot_pos);
			return 0;
		}
		inline int sendingsready_to_sendingdata(VerbsRdvSendRequest* req,
				const RdvRReadyMessage* rready_message) {
			// if we received the rready, we can send the data

			int res;

			// the send ready has been received and the slot can be freed
			req->endpoint->rdv.sready_free_remote_slot(req->rem_sready_message_pos);

			// we send our buffer as specified in the request
			ibv_sge sr_sg_lst = {
				reinterpret_cast<uint64_t>(req->buf),              // addr
				static_cast<uint32_t>(req->buflen),                // length
				req->mr ? req->mr->get_mr()->lkey : 0              // lkey
			};

			// prepare eager send request
			ibv_send_wr sr_wr;
			sr_wr.wr_id = reinterpret_cast<uint64_t>(req);
			sr_wr.opcode = IBV_WR_RDMA_WRITE_WITH_IMM; // we write directly into rdma buffer with imm
			sr_wr.send_flags = IBV_SEND_SIGNALED;
			sr_wr.imm_data = rready_message->recvreq_identifier;

			sr_wr.sg_list = &sr_sg_lst;
			sr_wr.num_sge = req->buflen > 0 ? 1 : 0; // only one message
			sr_wr.next = nullptr;

			sr_wr.wr.rdma.rkey = rready_message->rkey;
			sr_wr.wr.rdma.remote_addr = rready_message->addr;

			// set new state before posting
			req->req_state = rdv_send_request_states::SENDING_DATA;

			// now we can submit the message to the channel
			ibv_send_wr* bad_wr;
			res = ibv_post_send(req->endpoint->get_qp(), &sr_wr, &bad_wr);
			TMPI_ASSERT_FATAL(res == 0, "ibv_post_send() returned " << res, __FILE__, __LINE__);
			TMPI_LOG_LEVEL(3, "ibv_post_send", "with imm_data = " << sr_wr.imm_data);

			return 0;
		}
		inline int sendingdata_to_sentdata(VerbsRdvSendRequest* req) {
			// no lock required
			// if the data has been sent, we can deregister the memory region

			req->mr.reset();
			return 0;
		}

		/////////////////////////////
		// primary state changes
		inline int initialized_to_sendingsready(VerbsRdvSendRequest* req) {
			// register memory
			TMPI_LOG_LEVEL(4, "VerbsRdvSendRequest", "try to register " << req->buf << " with size "
					<< req->buflen);
			if (req->buflen > 0) {
				req->mr = provider_.register_memory(req->buf, req->buflen);
			}

			// prepare the sready send slot
			RdvSReadyMessage* loc_sready_message;

			// get a send slot
			int res;
			// this also sets the status to reserved
			std::tie(req->send_sready_slot_pos, loc_sready_message)
					= VerbsProvider::Rdv::sready_get_send_slot();
			TMPI_ASSERT_FATAL(loc_sready_message != nullptr, "No empty sready send slots! "
					"Increase RDV_MESSAGE_LIMIT!", __FILE__, __LINE__);

			// reserve a remote position
			{
				std::lock_guard<std::mutex> ep_lock(req->endpoint->lock);
				req->rid = req->endpoint->get_send_rid();
				req->rem_sready_message_pos = req->endpoint->rdv.sready_reserve_remote_slot();
			}
			TMPI_ASSERT_FATAL(req->rem_sready_message_pos != -1,
					"Did not find a sready remote slot!", __FILE__, __LINE__);

			// fill data that has been generated when creating the request
			loc_sready_message->send_request = reinterpret_cast<uint64_t>(req);
			loc_sready_message->tag = req->tag;
			loc_sready_message->context_id = req->context_id;
			loc_sready_message->src_id = tasks::get_globalunique_id();
			loc_sready_message->dest_id= tasks::get_threadunique_id(req->peer_rank);
			loc_sready_message->rid = req->rid;
			loc_sready_message->buflen = req->buflen;

			// we send the complete sready message
			ibv_sge sr_sg_lst = {
				reinterpret_cast<uint64_t>(loc_sready_message),    // addr
				rdvSReadyMessageSendSize,                          // length
				VerbsProvider::Rdv::sready_get_send_slots_lkey()   // lkey
			};

			// prepare send request
			ibv_send_wr sr_wr;
			sr_wr.wr_id = reinterpret_cast<uint64_t>(req);
			sr_wr.opcode = IBV_WR_RDMA_WRITE; // we write directly into rdma buffer
			sr_wr.send_flags = IBV_SEND_SIGNALED;

			sr_wr.sg_list = &sr_sg_lst;
			sr_wr.num_sge = 1; // only one message
			sr_wr.next = nullptr;

			const PeerInfo& pi = req->endpoint->rdv.sready_get_recv_slots_peer_info();
			sr_wr.wr.rdma.rkey = pi.rkey;
			sr_wr.wr.rdma.remote_addr = reinterpret_cast<uint64_t>(pi.addr)
					+ req->rem_sready_message_pos*sizeof(RdvSReadyMessage);

			// set new state before posting
			req->req_state = rdv_send_request_states::SENDING_SREADY;
			req->req_state_cntr = 0;

			// now we can submit the message to the channel
			ibv_send_wr* bad_wr;
			res = ibv_post_send(req->endpoint->get_qp(), &sr_wr, &bad_wr);
			TMPI_ASSERT_FATAL(res == 0, "ibv_post_send() returned " << res, __FILE__, __LINE__);
			TMPI_LOG_LEVEL(3, "ibv_post_send", "with imm_data = " << sr_wr.imm_data);

			return 0;
		}

		// both messages have been sent
		inline int sendingrequests_to_done(VerbsRdvSendRequest* req) {
			// no lock required
			// data has been arrived, so the request has been done
			sendingsready_to_sentsready(req);
			sendingdata_to_sentdata(req);

			using namespace tinympi::impl;
			req->req_state = rdv_send_request_states::DONE;
			baseRequest::request_completed(TO_BaseRequest(req));

			return 1;
		}
	} // END NAMESPACE impl

	/////////////////////////////
	// transparent function

	// this function is called when we received an rready message by polling the rready message
	// slots
	inline int retrieved_rready(VerbsRdvSendRequest* req, const RdvRReadyMessage* message) {
		using namespace impl;
		// endpoint lock required, e.g. for sentsreadywaitingrready_to_sentsreadysendingdata
		TMPI_DASSERT_FATAL(req->req_state == rdv_send_request_states::SENDING_SREADY,
				"unexpected request state (" << to_string(req->req_state) << ") at this place!",
				__FILE__, __LINE__);

		sendingsready_to_sendingdata(req, message);
		return 0;
	}

	// this function is called whenever we poll a work completion event from the completion queue.
	// this can be sready send completion event or a data send completion event
	inline int handle_poll(VerbsRdvSendRequest* const req, const ibv_wc&) {
		// no lock required
		using namespace impl;

		if (++req->req_state_cntr == 2) {
			// NO LOCK required
			return sendingrequests_to_done(req);
		}
		return 0;
	}

	// generate a new send request
	inline VerbsRdvSendRequest* generate(void* buf, buflen_t buflen, tag_t tag,
			context_id_t context_id, const tasks::Task& dest_pos, VerbsEndpoint* endpoint) {
		// endpoint lock required, e.g., for initialized_to_sendingsready
		using namespace impl;

		// try to get a request, otherwise allocate a new one
		VerbsRdvSendRequest* req = TO_VerbsRdvSendRequest(baseRequest::malloc());
		init(req);
		TMPI_LOG_LEVEL(3, "VerbsRdvSendRequest", "created new request (" << req << ")");

		// fill data
		INIT_BASE_REQUEST( \
			req, \
			dest_pos.get_identifier().global_identifier, \
			context_id, \
			tag, \
			buf, \
			buflen, \
			tasks::get_current() \
		)

		req->endpoint = endpoint;

		// start sending sready
		initialized_to_sendingsready(req);

		return req;
	}

} // END NAMESPACE verbsRdvSendRequest
} // END NAMESPACE impl
} // END NAMESPACE verbs
} // END NAMESPACE tinympi

#endif //_VERBS_RDVSENDREQUEST_H_
