// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _THREAD_H_635106f3226c46d3d114158785752466
#define _THREAD_H_635106f3226c46d3d114158785752466

#include <vector>

#include "status.h"
#include "global.h"
#include "utility/error.h"

#include "thread_queues.h"
#include "thread_sendrequest.h"
#include "thread_recvrequest.h"

namespace tinympi {

namespace thread2thread {

//tinympi::thread::self_wrapper
struct thread_wrapper {
private:
	// import some classes from the impl class to the namespace of the struct
	// we should not move them outside of impl, because they are not directly
	// availbale from the outside
	using ThreadRequest = impl::ThreadRequest;
	using BaseRequest = tinympi::impl::BaseRequest;
	using dispatch_methods = tinympi::impl::dispatch_methods;

public:
	// this will be used by request_dispatch
	using ProviderBaseRequest = ThreadRequest;

	static int progress();
	static int isend(void*, buflen_t, tasks::Task*, tag_t, context_id_t, BaseRequest**);
	static int posted_request(void*, buflen_t, const tasks::Task*, tag_t, context_id_t,
			BaseRequest**);
	static bool recv_or_post_anysource(void*, buflen_t, tag_t, context_id_t, BaseRequest**);

	static constexpr int max_request_size = max(
			sizeof(impl::ThreadSendRequest), sizeof(impl::ThreadRecvRequest)
	);

	template<dispatch_methods dispatch_method, typename... Args>
	static int request_dispatch(ThreadRequest*, Args...) {
		TMPI_FATAL_ERROR("Invalid call!", __FILE__, __LINE__);
	}
};

// this is an internal function, which is called by posted_request, when we got a matching recv for
// an unexpected message
template<>
inline int thread_wrapper::request_dispatch<
		thread_wrapper::dispatch_methods::posted_request_for_unexpected>(ThreadRequest* req,
		void* buf, buflen_t buflen) {
	// NO LOCK required

	using namespace impl;
	TMPI_LOG_LEVEL(1, "ThreadWrapper", "posted_request_for_unexpected");

	// we can then recv the data. We then copy [req->buf / req->buflen] to [buf / buflen]
	return threadRecvRequest::posted_irecv_for_unexpected_request(TO_ThreadRecvRequest(req), buf,
			buflen);
}
inline int thread_wrapper::progress() {
	// NO LOCK required

	// a progress function is not needed
	return 0;
}

template<>
inline int thread_wrapper::request_dispatch<thread_wrapper::dispatch_methods::fill_status>(
		ThreadRequest* req, TMPI_Status* status) {
	// NO LOCK required

	using namespace impl;

	// just dispatch fill_status to the correct function
	switch (req->thread_request_type) {
	case thread_request_types::THREAD_SEND:
		return threadSendRequest::fill_status(TO_ThreadSendRequest(req), status);
	case thread_request_types::THREAD_RECV:
		return threadRecvRequest::fill_status(TO_ThreadRecvRequest(req), status);
	default:
		TMPI_FATAL_ERROR("unknown request type", __FILE__, __LINE__);
		return -1;
	}
}
template<>
inline int thread_wrapper::request_dispatch<thread_wrapper::dispatch_methods::delete_request>(
		ThreadRequest* treq) {
	// NO LOCK required

	// just free the request
	TMPI_LOG_LEVEL(2, "thread_wrapper", "delete " << treq);
	using namespace impl;
	baseRequest::free(TO_BaseRequest(treq));
	return TMPI_SUCCESS;
}
template<>
inline int thread_wrapper::request_dispatch<thread_wrapper::dispatch_methods::try_cancel_request>(
		ThreadRequest* treq) {
	// LOCK REQUIRED FOR thread queue (OK)
	using namespace impl;
	LSB_RECORD(THREADWRAPPER_CANCELREQUEST_STARTED);

	// we currently only support cancellation of waiting requests (UnknownRequest)
	TMPI_ASSERT_FATAL(treq->thread_request_type == thread_request_types::THREAD_RECV,
			"should not cancel send request", __FILE__, __LINE__);

	threadRecvRequest::cancel(TO_ThreadRecvRequest(treq));
	LSB_RECORD(THREADWRAPPER_CANCELREQUEST_DONE);
	return 1;
}
inline int thread_wrapper::isend(void* buf, buflen_t buflen, tasks::Task* dest_descr,
		tag_t tag, context_id_t context_id, BaseRequest** reqptr) {
	// no lock required. MUST NOT lock thread queue, node queue
	using namespace impl::recvqueues;
	using namespace impl;
	LSB_RECORD(THREADWRAPPER_ISEND_STARTED);

	TMPI_LOG_LEVEL(1, "ThreadWrapper", "isend");

	// in any case we generate a send request, because we will send thread2thread
	ThreadSendRequest* sreq = threadSendRequest::generate(buf, buflen, tag, context_id, *dest_descr);

	// we search, whether there has already been posted a corresponding recv request. This recv
	// request can be thread2thread or node2node+anysource. In some cases, there may be multiple
	// iterations.
	bool done = false;
	while (!done) {
		// lock the thread queue
		// REQUIRED LOCK FOR find_waiting_request, add_to_unexpected_queue, threadRecvRequest::cancel
		std::unique_lock<std::mutex> thread_lock(
				static_cast<tasks::LocalTask*>(dest_descr)->thread_lock_);

		ThreadRequest* prev;
		// find in waiting queue of destination task, but do not dequeue yet
		ThreadRequest* rtreq = find_waiting_request<dequeue_request::no>(
				tasks::get_globalunique_id(), tag, context_id, *dest_descr, prev);

		if (rtreq != nullptr) {
			//  if we find a thread request, it depends whether the request that has been found is
			//  from ANY_SOURCE
			LSB_RECORD(THREADWRAPPER_ISEND_FOUNDWAITINGREQUEST);
			if (rtreq->peer_rank == (global_id_t)TMPI_ANY_SOURCE) {
				// we have found an any source request. In this case we also need to lock the queue
				// on node level, because there are requests we need to cancel. We cannot enforce this
				// lock.
				std::unique_lock<std::mutex> node_lock(
						static_cast<tasks::LocalTask*>(dest_descr)->verbs_lock_, std::defer_lock);
				if (node_lock.try_lock()) {

					// if we were successful, we can continue. First we need to dequeue the request.
					dequeue_waiting_request(rtreq, prev, *dest_descr);

					// Then we need to check whether another provider has already retrieved the
					// message. For libfabric this request cannot be cancelled then.
					TMPI_LOG_LEVEL(1, "ThreadWrapper", "found matching any source request " << rtreq);
					// requires both thread and node lock
					if (!other_request_already_retrieved(TO_BaseRequest(rtreq))) {

						LSB_RECORD(THREADWRAPPER_ISEND_NOTRETRIEVEDFROMOTHER);
						thread_lock.unlock();

						// if it was not yet retrieved, we take it
						TMPI_LOG_LEVEL(1, "ThreadWrapper", "forward to waiting request");
						threadRecvRequest::retrieved_isend_for_waiting_request(
								TO_ThreadRecvRequest(rtreq), buf, buflen, sreq);
						done = true;
						LSB_RECORD(THREADWRAPPER_ISEND_ANYSOURCERETRIEVED);

					} else {
						LSB_RECORD(THREADWRAPPER_ISEND_RETRIEVEDFROMOTHER);

						TMPI_LOG_LEVEL(1, "ThreadWrapper", "cancel request");
						// otherwise we cancel the local ThreadRecvRequest.
						threadRecvRequest::cancel(TO_ThreadRecvRequest(rtreq));

						LSB_RECORD(THREADWRAPPER_ISEND_CANCELLEDREQUEST);
					}
				} else {
					// if we were not successful getting the lock, we need to try again finding
					// a request
					continue;
				}
			} else {
				TMPI_LOG_LEVEL(1, "ThreadWrapper", "found waiting request");
				// if we have found a recv request which is not any source, we just put the data
				// there. No pointers can be broken and the thread lock is enough.

				// first we dequeue the request
				dequeue_waiting_request(rtreq, prev, *dest_descr);

				// after dequeueing the lcok is nomore needed
				thread_lock.unlock();

				// and we forward the isend to the request
				threadRecvRequest::retrieved_isend_for_waiting_request(TO_ThreadRecvRequest(rtreq),
						buf, buflen, sreq);
				done = true;

				LSB_RECORD(THREADWRAPPER_ISEND_RETRIEVED);
			}
		} else {
			// if no request has been found, we can add to the unexpected request queue. This only
			// affects thread-level data
			rtreq = TO_ThreadRequest(threadRecvRequest::retrieved_unexpected(buf, buflen, tag,
					context_id, sreq));

			TMPI_LOG_LEVEL(1, "ThreadWrapper", "add to unexpected queue (req = " << rtreq << ")");
			// add to unexpected queue of destination task
			add_to_unexpected_queue(rtreq, *dest_descr);
			done = true;

			LSB_RECORD(THREADWRAPPER_ISEND_ADDEDTOUNEXPECTED);
		}

		// when we are not yet done, this is because another request has retrieved this send request.
		// In this case just look for another request that matches
	}

	// then set the reqptr to the new send request
	*reqptr = TO_BaseRequest(sreq);
	TMPI_LOG_LEVEL(1, "ThreadWrapper", "isend done");
	return TMPI_SUCCESS;
}

// this function is called when a request was posted for a given source.
inline int thread_wrapper::posted_request(void* buf, buflen_t buflen,
		const tasks::Task* src_descr, tag_t tag, context_id_t context_id, BaseRequest** req) {
	// no lock required. MUST NOT lock thread queue
	using namespace impl;

	LSB_RECORD(THREADWRAPPER_POSTEDREQUEST_STARTED);

	// we need to lock the thread queue because we will look through this queue now. No other lock
	// is needed because no node queues are considered
	// REQUIRED LOCK FOR find_unexpected_request, add_to_waiting_queue
	std::unique_lock<std::mutex> lock_thread(tasks::get_current()->thread_lock_);
	LSB_RECORD(THREADWRAPPER_POSTEDREQUEST_ACQUIREDTHREADLOCK);

	// search request in my unexpected queue and deque if found.
	*req = TO_BaseRequest(recvqueues::find_unexpected_request<recvqueues::dequeue_request::yes>(
			src_descr->get_identifier().global_identifier, tag, context_id));
	if (*req != nullptr) {
		TMPI_LOG_LEVEL(2, "ThreadWrapper", "found in unexpected queue " << *req);
		LSB_RECORD(THREADWRAPPER_POSTEDREQUEST_FOUNDUNEXPECTED);

		// if we have found it, we forward our data to the request and retrieve the data
		int res = request_dispatch<dispatch_methods::posted_request_for_unexpected>(
				TO_ThreadRequest(*req), buf, buflen);

		LSB_RECORD(THREADWRAPPER_POSTEDREQUEST_DISPATCHED);
		return res;

	} else {
		TMPI_LOG_LEVEL(2, "ThreadWrapper", "did not find in unexpected queue");

		// otherwise, we enqueue a new request to the waiting queue
		*req = TO_BaseRequest(threadRecvRequest::generate_waiting_request(*src_descr, tag,
				context_id, buf, buflen));

		recvqueues::add_to_waiting_queue(TO_ThreadRequest(*req));
		LSB_RECORD(THREADWRAPPER_POSTEDREQUEST_ADDEDTOWAITING);
		return TMPI_SUCCESS;
	}
}
// this function is called when a request was posted with ANY_SOURCE as source
inline bool thread_wrapper::recv_or_post_anysource(void* buf, buflen_t buflen, tag_t tag,
		context_id_t context_id, BaseRequest** req) {
	// LOCK REQUIRED FOR thread queue, add_to_waiting_queue, dequeue_request (OK)
	using namespace impl;
	LSB_RECORD(THREADWRAPPER_RECVORPOSTANYSOURCE_STARTED);

	// we first find the request in the unexpected queue
	ThreadRequest* treq = recvqueues::find_unexpected_anysource_request
			<recvqueues::dequeue_request::yes>(tag, context_id);
	if (treq != nullptr) {
		LSB_RECORD(THREADWRAPPER_RECVORPOSTANYSOURCE_FOUNDUNEXPECTED);

		TMPI_LOG_LEVEL(2, "ThreadWrapper", "found in unexpected request");

		// if this was successful, then retrieve it
		request_dispatch<dispatch_methods::posted_request_for_unexpected>(treq, buf, buflen);

		// add old any source to new request
		treq->next_anysource_request = *req;
		*req = TO_BaseRequest(treq);

		LSB_RECORD(THREADWRAPPER_RECVORPOSTANYSOURCE_DISPATCHED);

		// signal that we have recvd the message
		return true;

	} else {
		TMPI_LOG_LEVEL(2, "ThreadWrapper", "did not find unexpected request");

		// otherwise, we enqueue a new request to the waiting queue
		treq = TO_ThreadRequest(threadRecvRequest::generate_anysource_waiting_request(
				tag, context_id, buf, buflen));
		recvqueues::add_to_waiting_queue(treq);

		// add old any source to new request
		treq->next_anysource_request = *req;
		*req = TO_BaseRequest(treq);

		// signal that we have not recvd the message
		LSB_RECORD(THREADWRAPPER_RECVORPOSTANYSOURCE_ADDEDTOWAITING);
		return false;
	}
}

} // END NAMESPACE thread2thread
} // END NAMESPACE tinympi

#endif //_THREAD_H_
