// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _THREAD_QUEUES_H_635106f3226c46d3d114158785752466
#define _THREAD_QUEUES_H_635106f3226c46d3d114158785752466

#include <mutex>

#include "tasks.h"
#include "threads.h"
#include "thread_request.h"

#include "provider_impl/queue_functions.h"

namespace tinympi {
namespace thread2thread {
namespace impl {
namespace recvqueues {
	using namespace tinympi::impl::recvqueues;

	extern std::mutex queue_mutex_;

	////////////////////////////
	// FIND REQUEST
	// finds a request in the unexpected queue
	template<typename dequeue_request_>
	inline ThreadRequest* find_unexpected_request(const global_id_t& source, tag_t tag,
			context_id_t context_id) {
		// LOCK REQUIRED FOR thread queue (OK)

		ThreadRequest* prev;
		return tinympi::impl::recvqueues::find_unexpected_request(source, tag, context_id,
			tasks::get_current()->thread_unexpected_queue_head_,
			tasks::get_current()->thread_unexpected_queue_tail_, prev ,dequeue_request_());
	}

	// finds a request in the unexpected queue
	template<typename dequeue_request_>
	inline ThreadRequest* find_unexpected_anysource_request(tag_t tag, context_id_t context_id) {
		// LOCK REQUIRED FOR thread queue (OK)

		ThreadRequest* prev;
		return tinympi::impl::recvqueues::find_unexpected_anysource_request(tag, context_id,
			tasks::get_current()->thread_unexpected_queue_head_,
			tasks::get_current()->thread_unexpected_queue_tail_, prev, dequeue_request_());
	}

	// finds a request in the waiting queue
	template<typename dequeue_request_>
	inline ThreadRequest* find_waiting_request(const global_id_t& source, tag_t tag,
			context_id_t context_id, tasks::Task& dest_task, ThreadRequest*& prev) {
		// LOCK REQUIRED FOR thread queue (OK)

		tasks::LocalTask& ltask = static_cast<tasks::LocalTask&>(dest_task);
		return tinympi::impl::recvqueues::find_waiting_request(source, tag, context_id,
			ltask.thread_waiting_queue_head_, ltask.thread_waiting_queue_tail_, prev,
			dequeue_request_());
	}

	////////////////////////////
	// ADD REQUEST
	inline void add_to_waiting_queue(ThreadRequest* req) {
		// LOCK REQUIRED FOR thread queue (OK)

		add_to_queue(req, tasks::get_current()->thread_waiting_queue_head_,
				tasks::get_current()->thread_waiting_queue_tail_);
		TMPI_LOG_LEVEL(3, "ThreadQueues", "*head = " <<
				tasks::get_current()->thread_waiting_queue_head_);
	}
	// add to an unexpected queue of another thread
	inline void add_to_unexpected_queue(ThreadRequest* req, tasks::Task& dest_task) {
		// LOCK REQUIRED FOR thread queue (OK)

		tasks::LocalTask& ltask = static_cast<tasks::LocalTask&>(dest_task);
		add_to_queue(req, ltask.thread_unexpected_queue_head_, ltask.thread_unexpected_queue_tail_);
	}

	////////////////////////////
	// REMOVE REQUEST
	// this functions only work locally!
	inline bool remove_waiting_request(ThreadRequest* req) {
		// LOCK REQUIRED FOR thread queue (OK)

		using namespace baseRequest::impl;
		TMPI_LOG_LEVEL(3, "ThreadQueues", "try remove req " << req
				<< " from waiting queue (head = "
				<< &req->waiting_task->thread_waiting_queue_head_ << ")");
		TMPI_LOG_LEVEL(3, "ThreadQueues", "*head = " <<
				req->waiting_task->thread_waiting_queue_head_);
		return tinympi::impl::recvqueues::remove_waiting_request(req,
			req->waiting_task->thread_waiting_queue_head_,
			req->waiting_task->thread_waiting_queue_tail_);
	}
	inline bool remove_unexpected_request(ThreadRequest* req) {
		// LOCK REQUIRED FOR thread queue
		using namespace baseRequest::impl;
		return tinympi::impl::recvqueues::remove_unexpected_request(req,
			req->waiting_task->thread_unexpected_queue_head_,
			req->waiting_task->thread_unexpected_queue_tail_);
	}

	////////////////////////////
	// DEQUEUE REQUEST
	inline void dequeue_waiting_request(ThreadRequest*& req, ThreadRequest*& prev,
			tasks::Task& dest_task) {
		// LOCK REQUIRED FOR thread queue

		tasks::LocalTask& ltask = static_cast<tasks::LocalTask&>(dest_task);

		// if prev is nullptr, we found a new head
		if (prev == nullptr) {
			TMPI_ASSERT_FATAL(ltask.thread_waiting_queue_head_ == req, "asdf", __FILE__, __LINE__);
			ltask.thread_waiting_queue_head_ = TO_ThreadRequest(req->next_request);
		} else {
			// otherwise we are somewhere in between or at the tail
			prev->next_request = req->next_request;
			// the tail can be recognized because next_request is set to nullptr
			if (prev->next_request == nullptr) {
				ltask.thread_waiting_queue_tail_ = prev;
			}
		}
	}
} // END NAMESPACE recvqueues
} // END NAMESPACE impl
} // END NAMESPACE thread2thread
} // END NAMESPACE tinympi

#endif // _THREAD_QUEUES_H_
