// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _THREAD_RECVREQUEST_H_635106f3226c46d3d114158785752466
#define _THREAD_RECVREQUEST_H_635106f3226c46d3d114158785752466

#include <cstring>

#include "global.h"
#include "status.h"

#include "thread_sendrequest.h"
#include "thread_request.h"

namespace tinympi {
namespace thread2thread {
namespace impl {

#define THREAD_RECV_REQUEST_MEMBERS \
	THREAD_REQUEST_MEMBERS \
	ThreadSendRequest* send_req_ptr;

//tinympi::thread::impl::ThreadRecvRequest
struct ThreadRecvRequest {
	THREAD_RECV_REQUEST_MEMBERS
};
static_assert(std::is_standard_layout<ThreadRecvRequest>::value, "ThreadRecvRequest must be POD!");

REQUESTS_CONVERTIBLE(ThreadRecvRequest, ThreadRequest)
REQUESTS_CONVERTIBLE(ThreadRecvRequest, BaseRequest)

ASSERT_THREAD_REQUEST_COMPATIBLE(ThreadRecvRequest)

#define ASSERT_THREAD_RECVREQUEST_COMPATIBLE(TYPE) \
	ASSERT_THREAD_REQUEST_COMPATIBLE(TYPE) \
	VERIFY_COMPATIBILITY(ThreadRecvRequest, TYPE, send_req_ptr) \

namespace threadRecvRequest {

	namespace impl {
		inline void init(ThreadRecvRequest* req) {
			// NO LOCK required
			threadRequest::init(TO_ThreadRequest(req), thread_request_types::THREAD_RECV);
		}

		inline int retrieve_data(ThreadRecvRequest* req, void* sendbuf, buflen_t sendbuflen,
				void* recvbuf, buflen_t recvbuflen) {
			// NO LOCK required

			TMPI_ASSERT_FATAL(sendbuflen <= recvbuflen, "overflow error: "
					<< "recvd buffer has size " << sendbuflen
					<< ", expected size: " << recvbuflen, __FILE__, __LINE__);

			// Copy the data.
			memcpy(recvbuf, sendbuf, sendbuflen);

			// then the request and the corresponding send request are completed
			using namespace tinympi::impl;
			baseRequest::request_completed(TO_BaseRequest(req->send_req_ptr));
			baseRequest::request_completed(TO_BaseRequest(req));

			return TMPI_SUCCESS;
		}
	} // END NAMESPACE impl

	inline void cancel(ThreadRecvRequest* req) {
		// LOCK REQUIRED FOR thread queue (OK)

		// A request can be cancelled by removing from the waiting / unexpected queue

		// we first look in the waiting queue, because ANY SOURCE requests will be found there
		if (recvqueues::remove_waiting_request(TO_ThreadRequest(req))) {
			// we have successfully cancelled the request
			baseRequest::request_cancelled(TO_BaseRequest(req));

		// otherwise try to cancel unexpected queue
		} else if (recvqueues::remove_unexpected_request(TO_ThreadRequest(req))) {
			// we have successfully cancelled the request
			baseRequest::request_cancelled(TO_BaseRequest(req));
		}
	}

	inline bool fill_status(ThreadRecvRequest* req, TMPI_Status* status) {
		// NO LOCK required

		using tinympi::impl::baseRequest::is_terminated;

		// only fill if this specific request is completed
		if (is_terminated(TO_BaseRequest(req))) {
			// TODO if we support multiple communicators, we should add this here
			TMPI_ASSERT_FATAL_EFF(req->context_id == comm_world_.get_id(),
					"unsupported communicator", __FILE__, __LINE__);
			status->MPI_SOURCE = comm_world_.get_commrank(req->peer_rank);
			TMPI_LOG_LEVEL(1, "status", status->MPI_SOURCE)
			status->MPI_TAG = req->tag;
			status->MPI_ERROR = TMPI_SUCCESS;
			status->size = req->buflen;
			return true;

		} else {
			return false;
		}
	}
	// this function is called for an isend where no matching irecv has been posted yet
	inline ThreadRecvRequest* retrieved_unexpected(void* buf, buflen_t buflen, tag_t tag,
			context_id_t context_id, ThreadSendRequest* sreq) {
		// NO LOCK required

		using namespace impl;

		// in the case we retrieved an unexpected send, i.e. a send without a posted recv,
		// we just create a send request and wait for the isend
		ThreadRecvRequest* req = TO_ThreadRecvRequest(baseRequest::malloc());
		init(req);
		INIT_BASE_REQUEST( \
			req, \
			tasks::get_globalunique_id(), \
			context_id, \
			tag, \
			buf, \
			buflen, \
			nullptr
		)
		req->send_req_ptr = sreq;

		return req;
	}
	// this function is called when the matching irecv is then posted
	inline int posted_irecv_for_unexpected_request(ThreadRecvRequest* req, void* buf,
			size_t buflen) {
		// NO LOCK required

		using namespace impl;
		req->waiting_task = tasks::get_current();
		retrieve_data(req, req->buf, req->buflen, buf, buflen);

		// tag and size are already correctly initialized
		return TMPI_SUCCESS;
	}

	// this function is called when for a request the isend is finally posted
	inline int retrieved_isend_for_waiting_request(ThreadRecvRequest* req,
			void* buf, size_t buflen, ThreadSendRequest* sreq) {
		// NO LOCK required

		using namespace impl;
		using namespace tinympi::impl;

		// add send request pointer
		req->send_req_ptr = sreq;

		// we then also need to update our tag and our size according to the send request
		req->buflen = sreq->buflen;
		req->tag = sreq->tag;
		req->peer_rank = tasks::get_globalunique_id();

		// then retrieve data
		TMPI_LOG_LEVEL(2, "ThreadRecvRequest", "retrieve data");
		retrieve_data(req, buf, buflen, req->buf, req->buflen);

		return TMPI_SUCCESS;
	}
	inline ThreadRecvRequest* generate_anysource_waiting_request(tag_t tag,
			context_id_t context_id, void* buf, buflen_t buflen) {
		// NO LOCK required

		using namespace impl;

		// allocate a new request
		ThreadRecvRequest* req = TO_ThreadRecvRequest(baseRequest::malloc());
		TMPI_LOG_LEVEL(3, "ThreadRecvRequest", "generate waiting (req = " << req << ")");
		init(req);

		INIT_BASE_REQUEST(
			req,
			TMPI_ANY_SOURCE,
			context_id,
			tag,
			buf,
			buflen,
			tasks::get_current()
		)
		return req;
	}

	inline ThreadRecvRequest* generate_waiting_request(const tasks::Task& pos, tag_t tag,
			context_id_t context_id, void* buf, buflen_t buflen) {
		// NO LOCK required
		using namespace impl;

		// allocate a new request
		ThreadRecvRequest* req = TO_ThreadRecvRequest(baseRequest::malloc());
		TMPI_LOG_LEVEL(3, "ThreadRecvRequest", "generate waiting (req = " << req << ")");
		init(req);

		INIT_BASE_REQUEST(
			req,
			pos.get_identifier().global_identifier,
			context_id,
			tag,
			buf,
			buflen,
			tasks::get_current()
		)
		return req;
	}
} // END NAMESPACE threadRecvRequest
} // END NAMESPACE impl
} // END NAMESPACE thread2thread
} // END NAMESPACE tinympi

#endif //_THREAD_RECVREQUEST_H_
