// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _PT2PT_H_635106f3226c46d3d114158785752466
#define _PT2PT_H_635106f3226c46d3d114158785752466

#include <cstring>

#include "global.h"

// everything related to requests (all providers)
#include "provider.h"

// additional things
#include "communicator.h"
#include "status.h"

// TMPI_LOG
#include "debug/debugger.h"

// tasks
#include "tasks.h"

namespace tinympi {
namespace pt2pt {

namespace impl {

	// deletes a request with all related requests (ANY SOURCE)
	inline void delete_request(TMPI_Request req) {
		LSB_RECORD(PT2PT_DELETEREQUEST_STARTED);
		// no lock required
		using namespace tinympi;
		using namespace tinympi::impl;
		if (req->next_anysource_request == nullptr) {
			// if we are not an anysource request, just delete
			request_dispatch<dispatch_methods::delete_request>(&req);
		} else {
			// otherwise start with the second request
			TMPI_Request tmp_req = req->next_anysource_request;
			while (tmp_req != req) {
				TMPI_LOG_LEVEL(1, "delete request", tmp_req);
				TMPI_Request next_request = tmp_req->next_anysource_request;
				TMPI_LOG_LEVEL(1, "del_req->next_anysource_request", next_request);
				request_dispatch<dispatch_methods::delete_request>(&tmp_req);
				tmp_req = next_request;
			}
			// and finally delete the main request
			request_dispatch<dispatch_methods::delete_request>(&req);
		}
		LSB_RECORD(PT2PT_DELETEREQUEST_DONE);
	}

	inline int irecv_anysource(void* buf, buflen_t buflen, tag_t tag, context_id_t context_id,
			TMPI_Request* req) {
		LSB_RECORD(PT2PT_IRECVANYSOURCE_STARTED);
		// no lock required. MUST NOT lock thread queue, node queue
		TMPI_LOG_LEVEL(1, "irecv_anysource", req);
		using namespace tinympi::impl;
		// reset request
		*req = nullptr;

		// either we find a matching send, or we post the message to all queues. When we post
		// the message to multiple queues, we should be able to loop through all requests with
		// req->next_anysource_req. We always prepend new requests to the head, which is stored
		// in req. We temporally store tail_req to complete the ring in the end

		// first lock the thread queue and the node queue
		std::unique_lock<std::mutex> node_lock(tasks::get_current()->verbs_lock_);
		std::unique_lock<std::mutex> thread_lock(tasks::get_current()->thread_lock_);
		LSB_RECORD(PT2PT_IRECVANYSOURCE_ACQUIREDLOCKS);
		TMPI_LOG_LEVEL(1, "irecv_anysource", "lock acquired");

		// if we find a local message, we just use this request as corresponding send request.
		// Other requests have not yet been posted. Required thread queue lock
		if (ThreadWrapper::recv_or_post_anysource(buf, buflen, tag, context_id, req)) {
			LSB_RECORD(PT2PT_IRECVANYSOURCE_THREADRECVD);
			// then everything is done. No other locks are needed.
			TMPI_LOG_LEVEL(1, "irecv_anysource", "found in ThreadWrapper queue");

		}  else {
			LSB_RECORD(PT2PT_IRECVANYSOURCE_THREADPOSTED);
			// keep the tail
			TMPI_Request tail_req = *req;

			// in recv_or_post_anysourcewe need to get the endpoint lock. If we don't get it,
			// we need to try again
			// try to recv from another node. Required node lock
			using namespace tinympi::verbs;
			if (NodeWrapper::recv_or_post_anysource(buf, buflen, tag, context_id, req)) {
				LSB_RECORD(PT2PT_IRECVANYSOURCE_NODERECVD);

				TMPI_LOG_LEVEL(1, "irecv_anysource", "found in NodeWrapper queue");

				// if that was successful, complete the ring
				tail_req->next_anysource_request = *req;

				// and delete the already existing requests. Required node + thread locks
				other_matched_anysource((*req)->next_anysource_request, *req);

				LSB_RECORD(PT2PT_IRECVANYSOURCE_NODEDELETEDOTHERS);

				// unlock thread and node lock
				thread_lock.unlock();
				node_lock.unlock();

				// if we have found it, we forward our data to the request and retrieve the data
				request_dispatch<dispatch_methods::posted_request_for_unexpected>(req, buf, buflen);

				TMPI_ASSERT_FATAL((*req)->next_anysource_request == nullptr,
						"should not have next pointer", __FILE__, __LINE__);

				LSB_RECORD(PT2PT_IRECVANYSOURCE_NODEDONE);

			} else {
				// when the request is posted on thread and node queue, we only complete the
				// ring
				tail_req->next_anysource_request = *req;
				TMPI_LOG_LEVEL(1, "irecv_anysource", "did not find in queues");
				LSB_RECORD(PT2PT_IRECVANYSOURCE_NODEPOSTED);
			}

		}

		// req points to the last request that has been added. Therefore, any request which is
		// cancelled will not be the head and req points to a valid request
		return TMPI_SUCCESS;

	}
	inline int irecv_knownsource(void* buf, buflen_t buflen, rank_type src, tag_t tag,
			context_id_t context_id, const TMPI_Comm comm, TMPI_Request* req) {
		// no lock required. MUST NOT lock anything
		TMPI_LOG_LEVEL(1, "irecv_knownsource", req);

		// find task descriptor
		const tasks::Task& src_task = comm->get_task(src);

		// and post request to the corresponding task
		return src_task.post_knownsource_recv_request(buf, buflen, &src_task, tag, context_id, req);

	}

} // END NAMESPACE impl

// CONVENTION: Each input signature takes Requests of type TMPI_Request*
//             Everything in the ProviderWrapper and request_dispatch takes BaseRequest**
// wait until one specific request is terminated
inline static int wait(TMPI_Request* req) {
	LSB_Rec(999);
	LSB_Set_Rparam_int("type", 1);

	using namespace impl;
	using namespace tinympi::impl;
	using namespace tinympi;

	// try to make progress
	verbs::fast_progress();

	// check if request has terminated
	if (baseRequest::is_terminated(*req)) {
		LSB_Set_Rparam_int("type", 0);
		return 0;
	}

	credit_t cnt = 0;
	// note we use 10 as an arbitrary number which is large enough that we cannot get credit = 0
	// while setting the tasks as waiting. Otherwise the task could be scheduled while it is still
	// still active
	tasks::set_credit(10);
	cnt += baseRequest::set_task_as_waiting(req);
	TMPI_LOG_LEVEL(1, "wait", "wait for " << *req << "(" << cnt << ")");
	// then we atomically decrease the credit to the right value. If it gets 0, all requests are
	// already done, otherwise the credit has the right number
	if (tasks::dec_credit(10 - cnt) != 0) {
		int res = 0;
		LSB_Rec(3);

		while (tasks::yield() == 0)
			res += progress();

		LSB_Set_Rparam_int("type", -res);
		LSB_Rec(tasks::has_active_task() ? 20 : 21);
		LSB_Set_Rparam_int("type", 1);
	}

	// when we get back, the request should be terminated
	TMPI_ASSERT_FATAL(baseRequest::is_terminated(*req), "request is not terminated!", __FILE__,
			__LINE__);
	LSB_Rec(3);
	LSB_Set_Rparam_int("type", 0);

	return 0;
}

inline int isend(void* buf, buflen_t buflen, rank_type dest, tag_t tag, context_id_t context_id,
		const TMPI_Comm comm, TMPI_Request* req) {
	// no lock required. MUST NOT lock thread queue, node queue (OK)
	TMPI_LOG_LEVEL(0, "pt2pt", "isend");

	using namespace tinympi;
	using namespace tinympi::impl;

	// find process
	tasks::Task& dest_task = comm->get_task(dest);

	// and continue with the corresponding isend function
	return dest_task.isend(buf, buflen, &dest_task, tag, context_id, req);
}

inline int irecv(void* buf, buflen_t buflen, rank_type src, tag_t tag, context_id_t context_id,
		const TMPI_Comm comm, TMPI_Request* req) {
	// we dispatch the recv requests to either anysource or non-anysource requests
	using namespace impl;
	TMPI_LOG_LEVEL(0, "pt2pt", "irecv");

	if (src == TMPI_ANY_SOURCE) {
		return irecv_anysource(buf, buflen, tag, context_id, req);

	} else {
		return irecv_knownsource(buf, buflen, src, tag, context_id, comm, req);
	}
}

inline int waitall(int count, TMPI_Request requests[], TMPI_Status statuses[]) {
	LSB_Rec(999);
	LSB_Set_Rparam_int("type", 1);

	TMPI_LOG_LEVEL(0, "pt2pt", "waitall");
	LSB_RECORD(PT2PT_WAITALL_STARTED);
	verbs::fast_progress();

	using namespace tinympi;
	using namespace tinympi::impl;
	using namespace impl;

	// first check if all requests are done
	int unterminated = 0;
	for (int i = 0; i < count; ++i) {
		if (requests[i] != nullptr) {
			if (!baseRequest::is_terminated(requests[i])) {
				unterminated += 1;
			} else {
				// if a request has terminated, we can delete it and fill the status
				if (statuses != TMPI_STATUSES_IGNORE)
					fill_status(&requests[i], &statuses[i]);
				delete_request(requests[i]);
				requests[i] = nullptr;
			}
		}
	}

	// if yes, we can terminate
	if (unterminated == 0) {
		TMPI_LOG_LEVEL(0, "waitall", "Finished");
		LSB_RECORD(PT2PT_WAITALL_DONE);
		LSB_Set_Rparam_int("type", 0);
		return TMPI_SUCCESS;
	}

	// otherwise we need to set the requests to waiting. We first set the credit to
	// 3*count + 1 which is large enough that the credit cannot get 0 while setting the requests to
	// waiting
	credit_t cnt = 0;
	tasks::set_credit(3 * count + 1);
	for (int i = 0; i < count; ++i) {
		if (requests[i] != nullptr) {
			// we then try to set the tasks to waiting if they are not completed yet. We can also
			// have added > 1 when it is an any source request
			credit_t added = baseRequest::set_task_as_waiting(&requests[i]);
			if (added > 0) {
				TMPI_LOG_LEVEL(1, "wait_all", "wait for " << requests[i] << "(" << added << ")");
				cnt += added;
			} else {
				// if a request has terminated, we can delete it and fill the status
				if (statuses != TMPI_STATUSES_IGNORE)
					fill_status(&requests[i], &statuses[i]);
				delete_request(requests[i]);
				requests[i] = nullptr;
			}
		}
	}
	LSB_RECORD(PT2PT_WAITALL_CREDITSET);
	// then we set the credit to the right number such that it gets 0 as soon as we are done.
	if (tasks::dec_credit(3 * count + 1 - cnt) != 0) {
		LSB_Rec(3);
		int res = 0;

		// try to yield until I was successful and data arrives
		while (tasks::yield() == 0)
			res += progress();

		LSB_Set_Rparam_int("type", -res);
		LSB_Rec(tasks::has_active_task() ? 20 : 21);
		LSB_Set_Rparam_int("type", 1);
	}

	// now all requests should have finished. fill the statuses. If a request is not done yet,
	// we have an error.
	unterminated = 0;
	for (int i = 0; i < count; ++i) {
		if (requests[i] != nullptr) {
			if (!baseRequest::is_terminated(requests[i])) {
				TMPI_FATAL_ERROR("yield came back before all requests done! "
						<< "(" << requests[i] << ")", __FILE__, __LINE__);
			} else {
				// if a request has finished, then delete it and get the status
				if (statuses != TMPI_STATUSES_IGNORE)
					fill_status(&requests[i], &statuses[i]);
				delete_request(requests[i]);
				requests[i] = nullptr;
			}
		}
	}

	TMPI_LOG_LEVEL(0, "waitall", "Finished");
	LSB_RECORD(PT2PT_WAITALL_DONE);
	LSB_Rec(3);
	LSB_Set_Rparam_int("type", 0);
	return TMPI_SUCCESS;
}
inline int wait(TMPI_Request* request, TMPI_Status* status) {
	LSB_RECORD(PT2PT_WAIT_STARTED);
	TMPI_LOG_LEVEL(0, "pt2pt", "wait");

	using namespace tinympi;
	using namespace tinympi::impl;
	using namespace impl;

	wait(request);
	if (status !=  TMPI_STATUS_IGNORE)
		fill_status(request, status);
	TMPI_LOG_LEVEL(1, "pt2pt", "delete request");
	delete_request(*request);
	*request = nullptr;
	LSB_RECORD(PT2PT_WAIT_DONE);
	return TMPI_SUCCESS;
}

inline TMPI_ALWAYSINLINE int recv(void* buf, buflen_t buflen, rank_type src, tag_t tag,
		context_id_t context_id, const TMPI_Comm comm, TMPI_Status* status) {
	TMPI_LOG_LEVEL(0, "pt2pt", "recv");
	using namespace tinympi;
	using namespace tinympi::impl;
	using namespace impl;
	TMPI_Request req;

	// create an irecv
	int res = irecv(buf, buflen, src, tag, context_id, comm, &req);
	TMPI_ASSERT_FATAL_EFF(res == TMPI_SUCCESS, "recv failed", __FILE__, __LINE__);
	// wait for completion
	wait(&req);
	// fill the status
	fill_status(&req, status);
	// delete the request
	delete_request(req);

	return res;
}
inline TMPI_ALWAYSINLINE int send(void* buf, buflen_t buflen, rank_type dest, tag_t tag,
		context_id_t context_id, const TMPI_Comm comm) {
	TMPI_LOG_LEVEL(0, "pt2pt", "send");
	using namespace tinympi;
	using namespace tinympi::impl;
	using namespace impl;
	TMPI_Request req;

	// create an isend
	int res = isend(buf, buflen, dest, tag, context_id, comm, &req);
	TMPI_ASSERT_FATAL_EFF(res == TMPI_SUCCESS, "send failed", __FILE__, __LINE__);
	// wait for completion
	wait(&req);
	// delete the request
	delete_request(req);

	return res;
}
inline TMPI_ALWAYSINLINE int sendrecv(void* sendbuf, buflen_t sendlength, rank_type dest,
		tag_t sendtag, void* recvbuf, buflen_t recvlength, rank_type src, tag_t recvtag,
		context_id_t context_id, const TMPI_Comm comm, TMPI_Status* status) {

	STATISTICS_ADD(dest, sendlength);

	TMPI_LOG_LEVEL(0, "pt2pt", "sendrecv (src " << src << ", dest " << dest << ")");
	int res;
	TMPI_Request req[2];

	res = isend(sendbuf, sendlength, dest, sendtag, context_id, comm, &req[0]);
	TMPI_ASSERT_FATAL_EFF(res == TMPI_SUCCESS, "send failed", __FILE__, __LINE__);

	res = irecv(recvbuf, recvlength, src , recvtag, context_id, comm, &req[1]);
	TMPI_ASSERT_FATAL_EFF(res == TMPI_SUCCESS, "recv failed", __FILE__, __LINE__);

	if (status != TMPI_STATUS_IGNORE) {
		TMPI_Status statuses[2];
		res = waitall(2, req, statuses);
		*status = statuses[1];
		return res;
	} else {
		res = waitall(2, req, TMPI_STATUSES_IGNORE);
		return res;
	}
}

} // END NAMESPACE pt2pt
} // END NAMESPACE tinympi

#endif // _PT2PT_H_
