// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _TASKS_INIT_H_635106f3226c46d3d114158785752466
#define _TASKS_INIT_H_635106f3226c46d3d114158785752466

#include "utility/error.h"

#include <cstring>
#include <vector>
#include <iostream>
#include <type_traits>

#include <malloc.h>
#include <pthread.h>

#include "global.h"
#include "utility/external.h"

#include "threads.h"
#include "stack.h"
#include "scheduler.h"
#include "task.h"

namespace tinympi {

namespace impl {
	int progress();
} // END NAMESPACE impl

namespace tasks {

// will be defined in ../tasks.h
template<typename keep_active = thread_inactive>
int yield();

namespace impl {

	using thread_active = std::true_type;
	using thread_inactive = std::false_type;

	// which is the main task (this is important for terminating tasks, because only the main_task
	// will survive)
	extern thread_local LocalTask* main_task_;

}

inline void terminate() {
	using namespace impl;
	LocalTask* task = LocalTask::get_current_task();
	task->terminate();

	if (task == main_task_) {
		// the main thread waits until all tasks have terminated. If not, we yield if there are other
		// active tasks
		while (!LocalTask::alltasks_terminated()) {
			if (scheduler_.has_active_task()) {
				yield();
			}
			tinympi::impl::progress();
		}
		// then we terminate
		TMPI_LOG_LEVEL(0, "TMPI_Finalize", "all tasks done");

	} else {
		// if this is not the main task: if there is another active thread, we activate it,
		// otherwise: if there is a task which is still alive, we wait for it to become active,
		// if all have terminated, we go to the main task
		while (!LocalTask::alltasks_terminated()) {
			if (scheduler_.has_active_task()) {
				yield();
			}
			tinympi::impl::progress();
		}
		LSB_RESET();
		LocalTask::run_task(main_task_);

		// and never come back ...
		throw std::logic_error("came back from terminate!");
	}
}

inline void pre_init() {
	using namespace impl;
	all_tasks_ = new LocalTask[env::num_tasks * env::num_threads];
}

// this function returns true, if we are the main thread, and false otherwise
inline bool startup() {
	using namespace impl;

	// retrieve stack top from threads
	uint64_t* base_stack_top = reinterpret_cast<uint64_t*>(threads::impl::self->stack_top);

	// we create a copy of the current stack, including registers and return address.
	void* state_copy = malloc(stack_size);
	uint64_t registers_copy[8];
	void* retaddr_copy;
	void* stack_pointer_copy;
	int is_root_task = copy_stack(state_copy, registers_copy, &retaddr_copy,
			&stack_pointer_copy, base_stack_top);
	if (is_root_task != 0) {
		// this is not the root task

		// when we are a new task
		return false;
	}

	// this is the root task
	LocalTask* local_tasks = &all_tasks_[env::num_tasks * env::thread_id];

	// then start all tasks.
	for (task_id_t task = 1; task < env::num_tasks; ++task) {

		// create one task context for each task and set its values
		local_tasks[task].init({task, env::thread_id, env::node_id}, state_copy, registers_copy,
				retaddr_copy, stack_pointer_copy);
		// we need to initialize the scheduler from here because task has no access
		local_tasks[task].set_scheduler(&impl::scheduler_);

		// this task is ready to be executed, therefore we add it to the scheduler
		scheduler_.add_task(local_tasks[task]);
	}

	// initialize self
	local_tasks[0].init_root({0, env::thread_id, env::node_id});
	local_tasks[0].set_scheduler(&impl::scheduler_);
	main_task_ = &local_tasks[0];

	// all threads have started and copied the state. No barrier is needed for tasks!
	free(state_copy);

	return true;
}
inline task_id_t get_taskrank() {
	return LocalTask::get_current_task()->get_identifier().local.task_id;
}



} // END NAMESPACE threads
} // END NAMESPACE tinympi

#endif // _TASKS_INIT_H_

