// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _TASKS_LOCAL_TASK_H_635106f3226c46d3d114158785752466
#define _TASKS_LOCAL_TASK_H_635106f3226c46d3d114158785752466

#include <mutex>
#include <atomic>

#include "global.h"

#include "task_identifier.h"
#include "task.h"

namespace tinympi {

// forward declare a few things
namespace thread2thread {
namespace impl {
	struct ThreadRequest;
} // END NAMESPACE impl
} // END NAMESPACE thread2thread
#if PROVIDER == PROVIDER_VERBS
namespace verbs {
namespace impl {
	struct VerbsRequest;
} // END NAMESPACE impl
} // END NAMESPACE verbs
#endif

namespace tasks {
namespace impl {
	class Scheduler;
	// import Requests for simplicity
	using ThreadRequest = tinympi::thread2thread::impl::ThreadRequest;
	using VerbsRequest = tinympi::verbs::impl::VerbsRequest;
} // END NAMESPACE impl
} // END NAMESPACE tasks

////////////////////////////
namespace tasks {
namespace impl {
	// array of all tasks on this node
	extern LocalTask* all_tasks_;

} // END NAMESPACE impl

class LocalTask : public Task {

////////////////////////////
// VARIABLES
private:
	// variables needed to activate thread
	uint64_t registers_[8];
	void* stack_pointer_;
	void* return_address_;
	// variables needed for the scheduler
	impl::Scheduler* scheduler_;
	std::atomic<credit_t> credit_;

public:
	// variables needed for thread to thread queues. Make them public for simplicity
	impl::ThreadRequest* thread_unexpected_queue_head_;
	impl::ThreadRequest* thread_unexpected_queue_tail_;
	impl::ThreadRequest* thread_waiting_queue_head_;
	impl::ThreadRequest* thread_waiting_queue_tail_;
	std::mutex thread_lock_;
	// variables needed for node to node queues. Make them public for simplicity
#if PROVIDER == PROVIDER_VERBS
	// Only the verbs provider needs task-local data!
	impl::VerbsRequest* verbs_unexpected_queue_head_;
	impl::VerbsRequest* verbs_unexpected_queue_tail_;
	impl::VerbsRequest* verbs_waiting_queue_head_;
	impl::VerbsRequest* verbs_waiting_queue_tail_;
	std::mutex verbs_lock_;
#endif
	// also next_task_ is used by the scheduler and is made transparent
	std::atomic<LocalTask*> next_task_;
private:
	// variable needed for task implementation
	void* stack_base_;

////////////////////////////
// FUNCTIONS
public:
	void init(task_identifier, void*, uint64_t*, void*, void*);
	void init_root(task_identifier task_desc);
	int run();

	credit_t dec_credit(credit_t, bool);
	int set_credit(credit_t);
	inline void terminate()                         { num_alive_tasks_ -= 1; }

	credit_t get_credit() const                     { return credit_; }
	void set_scheduler(impl::Scheduler* sch)        { scheduler_ = sch; }
	impl::Scheduler* get_scheduler() const          { return scheduler_; }

////////////////////////////
// STATIC VARIABLES
private:
	static thread_local LocalTask* current_task_;
	static thread_local unsigned num_alive_tasks_;

////////////////////////////
// STATIC FUNCTIONS
public:
	static void run_task(LocalTask* task);
	static inline LocalTask* get_current_task()     { return current_task_; }
	static inline bool alltasks_terminated()        { return num_alive_tasks_ == 0; }
};
inline void LocalTask::run_task(LocalTask* task) {
	using namespace impl;

	TMPI_LOG_LEVEL(1, "Tasking Engine", "switch from "
			<< (int)current_task_->identifier_.local.task_id << " to "
			<< (int)task->identifier_.local.task_id );

	EXTERNAL_HOOK_RUN_OTHER_TASK_START()
	LocalTask* old_task = current_task_;
	current_task_ = task;
	gototask(&old_task->stack_pointer_, &task->stack_pointer_, (void**)&old_task->registers_,
			(void**)&task->registers_, &old_task->return_address_, &task->return_address_);
	EXTERNAL_HOOK_RUN_OTHER_TASK_STOP()
}

inline void LocalTask::init(task_identifier task_desc, void* state, uint64_t* old_registers,
		void* old_return_address, void* old_stack_ptr) {

	identifier_ = task_desc;
	stack_base_ = memalign(64, stack_size);

	// state_copy stores first the size of the stack it has copied, then the old stack pointer,
	// and finally the copy itself
	uint64_t stack_src_size = reinterpret_cast<uint64_t*>(state)[0]; // in bytes
	uint64_t* stack_src_start = &reinterpret_cast<uint64_t*>(state)[1];

	// we need to copy the stack to the end of our new stack
	uint64_t* stack_dest_start = reinterpret_cast<uint64_t*>(
			reinterpret_cast<uint64_t>(stack_base_) + stack_size - stack_src_size);

	// then copy the stack
	memcpy(stack_dest_start, stack_src_start, stack_src_size);

	// we need to rewrite the stack. All references to the old stack need to point to the new
	// stack. Without this step we get into trouble.
	uint64_t stack_start_diff = reinterpret_cast<uint64_t>(stack_dest_start)
			- reinterpret_cast<uint64_t>(old_stack_ptr);
	for (unsigned i = 0; i < stack_src_size / 8; ++i) {
		// if we between the old stack bottom and the old stack top, we rewrite the reference.
		if (stack_dest_start[i] >= reinterpret_cast<uint64_t>(old_stack_ptr)
				&& stack_dest_start[i] < reinterpret_cast<uint64_t>(old_stack_ptr)
				+ stack_src_size) {
			stack_dest_start[i] = stack_dest_start[i] + stack_start_diff;
		}
	}

	// we also need to copy and rewrite the registers
	for (unsigned i = 0; i < 8; ++i) {
		if (old_registers[i] >= reinterpret_cast<uint64_t>(old_stack_ptr)
				&& old_registers[i] < reinterpret_cast<uint64_t>(old_stack_ptr) + stack_src_size) {
			registers_[i] = old_registers[i] + stack_start_diff;
		} else {
			registers_[i] = old_registers[i];
		}
	}

	// and we need to copy the return address and the stack pointer
	return_address_ = old_return_address;
	stack_pointer_ = stack_dest_start;

	// then we need to initialize all remaining data
	credit_ = 0;
	next_task_ = nullptr;
	thread_unexpected_queue_head_ = nullptr;
	thread_unexpected_queue_tail_ = nullptr;
	thread_waiting_queue_head_ = nullptr;
	thread_waiting_queue_tail_ = nullptr;
	verbs_unexpected_queue_head_ = nullptr;
	verbs_unexpected_queue_tail_ = nullptr;
	verbs_waiting_queue_head_ = nullptr;
	verbs_waiting_queue_tail_ = nullptr;

	num_alive_tasks_ += 1;

}
inline void LocalTask::init_root(task_identifier task_desc) {
	identifier_ = task_desc;

	// the root can be initialized very easily. Just make clear that we are the root (base=0)
	stack_base_ = 0;

	// the data needed for task switch is left out, because that is assigned when we switch the task

	// then set this task as active
	current_task_ = this;

	// the remaining data
	credit_ = 0;
	next_task_ = 0;
	thread_unexpected_queue_head_ = nullptr;
	thread_unexpected_queue_tail_ = nullptr;
	thread_waiting_queue_head_ = nullptr;
	thread_waiting_queue_tail_ = nullptr;
	verbs_unexpected_queue_head_ = nullptr;
	verbs_unexpected_queue_tail_ = nullptr;
	verbs_waiting_queue_head_ = nullptr;
	verbs_waiting_queue_tail_ = nullptr;

	num_alive_tasks_ += 1;
}
// switches from the current task to this task
inline int LocalTask::run() {
	using namespace impl;
	if (this != current_task_) {
		// update the current task
		LocalTask* old_task = current_task_;
		current_task_ = this;

		// then change to that task
		gototask(&old_task->stack_pointer_, &current_task_->stack_pointer_,
				(void**)&old_task->registers_, (void**)&current_task_->registers_,
				&old_task->return_address_, &current_task_->return_address_);
	}
	return 0;
}
// TODO overload dec_credit such that we can possibly directly jump to this task
inline credit_t LocalTask::dec_credit(credit_t d, bool add_to_scheduler_ = true) {
	using namespace impl;

	credit_t res = credit_.fetch_sub(d);
	TMPI_LOG_LEVEL(1, "dec new credit for " << this, credit_);
	// add to scheduler if we are ready
	if (res == d && add_to_scheduler_) {
		add_to_scheduler(this);
	}

	return res - d;
}
inline int LocalTask::set_credit(credit_t d) {
	TMPI_LOG_LEVEL(1, "set new credit for " << this, d);
	credit_.exchange(d);
	return 0;
}

} // END NAMESPACE tasks
} // END NAMESPACE tinympi

#endif // _TASKS_LOCAL_TASK_H_

