// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _TASKS_SCHEDULER_H_635106f3226c46d3d114158785752466
#define _TASKS_SCHEDULER_H_635106f3226c46d3d114158785752466

#include <type_traits>
#include <mutex>
#include <atomic>

#include "utility/error.h"

#include "task_global.h"
#include "task.h"
#include "local_task.h"

namespace tinympi {
namespace tasks {

namespace impl {

using add_to_tail = std::true_type;
using add_to_head = std::false_type;

using default_add_to = add_to_head;

struct alignas(16) Anchor {
	LocalTask* head_;
	LocalTask* tail_;
};

class Scheduler {
public:
	LocalTask* get_task();

	// yields a task and returns another task that is either ready or running
	template<typename keep_active = thread_inactive>
	void yield_task();

	// adds a new task
	template<typename T = default_add_to>
	void add_task(LocalTask&);

	// tests whether we still have tasks
	bool has_active_task();

	// prints the current queue of tasks
	void debug_print();

	Scheduler()
			: anchor_ (Anchor{nullptr, nullptr}) {
	}

private:
	std::atomic<Anchor> anchor_;
};

extern thread_local Scheduler scheduler_;

inline void Scheduler::debug_print() {
#ifdef SCHEDULER_DEBUG
	TMPI_LOG_LEVEL(5, "Scheduler::LOG", "");
	if (task_queue_head_ == nullptr) {
		TMPI_LOG_LEVEL(5, "", "tasks_queues   = empty");
	} else {
		LocalTask* ptr = task_queue_head_;
		TMPI_LOG_LEVEL(5, "", "tasks_queues   = ");
		while (ptr != nullptr) {
			TMPI_LOG_LEVEL(5, "", "                 " << ptr->get_identifier().local.task_id);
			ptr = ptr->get_next_task();
		}
	}
#endif
}

template<>
inline void Scheduler::add_task<add_to_head>(LocalTask& t) {
	debug_print();

	Anchor anchor;
	do {
		// load the current anchor
		anchor = anchor_.load();

		// if the queue is empty
		if (anchor.head_ == nullptr) {
			// try to set t as head and as tail
			t.next_task_ = nullptr;
			// when it is able to exchange the values, the queue is in a valid state immediately
			if (anchor_.compare_exchange_weak(anchor, Anchor{&t, &t}))
				return;
		} else {
			// otherwise set the head of the new node to the current head
			t.next_task_ = anchor.head_;
			// and try to set new anchor. Afterwards the queue is in a valid state immediately
			// if not, the external state is not changed
			if (anchor_.compare_exchange_weak(anchor, Anchor{&t, anchor.tail_}))
				return;
		}
	} while(true);

	debug_print();
}
template<>
inline void Scheduler::add_task<add_to_tail>(LocalTask& t) {
	debug_print();

	do {
		// load the current anchor
		Anchor anchor = anchor_.load();
		t.next_task_ = nullptr;

		// if the queue is empty
		if (anchor.head_ == nullptr) {
			// when it is able to exchange the values, the queue is in a valid state immediately
			if (anchor_.compare_exchange_weak(anchor, Anchor{&t, &t}))
				return;
		} else {
			// otherwise we need to set the next pointer of the tail of the queue, and the tail
			// pointer.
			LocalTask* expected = nullptr;
			// this
			if (anchor.tail_->next_task_.compare_exchange_weak(expected, &t)) {
				if (anchor_.compare_exchange_weak(anchor, Anchor{&t, anchor.tail_}))
					return;
				else
					anchor.tail_ = nullptr;
			}
		}
	} while (true);
	debug_print();
}

template<>
inline void Scheduler::yield_task<thread_inactive>() {
}
template<>
inline void Scheduler::yield_task<thread_active>() {
	add_task<add_to_head>(*LocalTask::get_current_task());
}
inline bool Scheduler::has_active_task() {
	return (anchor_.load().head_ != nullptr);
}
inline LocalTask* Scheduler::get_task() {
	debug_print();

	do {
		// load the current anchor
		Anchor anchor = anchor_.load();

		// if the queue is empty, then we are done
		if (anchor.head_ == nullptr) {
			return nullptr;

		// if there is only one task
		} else if (anchor.head_ == anchor.tail_) {
			// then we try to set head and tailptr to nullptr
			if (anchor_.compare_exchange_weak(anchor, Anchor{nullptr, nullptr})) {
				// when we succeed, we have successfully dequeued one element from the front
				return anchor.head_;
			}

		} else {
			// otherwise there are multiple elements
			if (anchor_.compare_exchange_weak(anchor, {anchor.head_->next_task_, anchor.tail_})) {
				return anchor.head_;
			}
		}
	} while (true);
}

inline void add_to_scheduler(LocalTask* task) {
	TMPI_LOG_LEVEL(2, "add to scheduler", task);
	task->get_scheduler()->add_task<add_to_head>(*task);
}


} // END NAMESPACE impl
} // END NAMESPACE tasks
} // END NAMESPACE tinympi

#endif // _TASKS_SCHEDULER_H_
