// Copyright (c) 2016 ETH-Zurich. All rights reserved.
// Use of this source code is governed by a BSD-style license that
// can be found in the LICENSE file.

#ifndef _TMPI_H_635106f3226c46d3d114158785752466
#define _TMPI_H_635106f3226c46d3d114158785752466

#include "utility/external.h"
#include "communicator.h"
#include "collectives.h"
#include "pt2pt.h"
#include "status.h"
#include "threads.h"
#include "tasks.h"

#include "debug/debugger.h"

namespace tinympi {
namespace impl {

	inline buflen_t TMPI_ALWAYSINLINE get_length(const int count, TMPI_Datatype type) {
		if (type == TMPI_INT)
			return count * sizeof(int);
		else if (type == TMPI_DOUBLE)
			return count * sizeof(double);
		else if (type == TMPI_BYTE)
			return count * sizeof(char);
		else if (type == TMPI_DOUBLE_INT)
			return count * sizeof(pair_type<double, int>);
		else {
			TMPI_FATAL_ERROR("unsupported datatype", __FILE__, __LINE__);
			return 0;
		}
	}
}
}

inline int TMPI_Init(int* argc, char*** argv) {
	using namespace tinympi;
	using namespace tinympi::impl;

	// initialize external tools (e.g., MPI)
	utility::external::pre_init(argc, argv);
	// retrieve number of nodes
	env::num_nodes = utility::external::get_numnodes();
	env::node_id = utility::external::get_noderank();

	// at this points, num_nodes, num_threads and num_tasks have valid data

	// initialize debugger (open file for preinit phase)
	preinit_debugger();

	///////////////////////////////////
	// PREINITIALIZATION PHASE
	// in this phase we set up everything that is globally shared among all tasks on this node

	// setup global environment

	// the communicator has a global comm_world
	GlobalCommunicator::pre_init();
	// the tasks have a global object with all task objects
	tasks::pre_init();
	// providers are allowed to be global
	NodeWrapper::pre_init();

	// on each processor we spawn N threads (currently fixed amount!)
	threads::startup();
	EXTERNAL_HOOK_RUN_INIT_THREADS()

	// return from N threads
	///////////////////////////////////
	// INITIALIZATION PHASE
	// initialize thread-local external environment
	utility::external::init();

	// providers are allowed to use thread local data
	NodeWrapper::init();

	tasks::startup();

	// we return M times from each task
	///////////////////////////////////
	// POSTINITIALIZATION PHASE

	// communicators can be set up when all tasks are running because we need these structures
	Communicator::post_init();

	// open file for postinit phase
	postinit_debugger();

	// id = 0 is the first task that gets here. then we block
	if (tasks::get_taskunique_id() == 0) {
		threads::barrier();
	}

	TMPI_LOG_LEVEL(0, "TMPI_Init", "done");

	return TMPI_SUCCESS;
}

inline int TMPI_Finalize() {
	using namespace tinympi;
	using namespace tinympi::impl;
	TMPI_LOG_LEVEL(0, "TMPI_Finalize", "started");

	// terminate all tasks
	tasks::terminate();

	// at this point we have only one task per thread. We need to sync all threads now, before
	// destroying global objects
	threads::barrier();

	///////////////////////////////////
	// TERMINATION PHASE

	// destroy communicators
	Communicator::finalize();

	// finalize thread-local external environment
	utility::external::finalize();

	// terminate all threads (kill pthreads)
	threads::terminate();

	///////////////////////////////////
	// POST-TERMINATION PHASE

	// then destroy global environment
	MPI_Barrier(MPI_COMM_WORLD);

	// finalize node wrapper
	NodeWrapper::post_finalize();
	// finalize global communicator
	GlobalCommunicator::post_finalize();

	// finalize external environment
	utility::external::post_finalize();
	TMPI_LOG("finalize", "complete");

	return TMPI_SUCCESS;
}
inline int TMPI_Comm_rank(TMPI_Comm comm, int* rank) {
	*rank = comm->get_rank();
	return TMPI_SUCCESS;
}
inline int TMPI_Comm_size(TMPI_Comm comm, int* rank) {
	*rank = comm->get_size();
	return TMPI_SUCCESS;
}

inline int TMPI_Irecv(void* buf, int count, TMPI_Datatype type, int src, int tag, TMPI_Comm comm,
	TMPI_Request* req) {
	LSB_RESET();

	TMPI_LOG_LEVEL(-1, "TMPI_Irecv", "Starting");
	using namespace tinympi;

	if (src == TMPI_PROC_NULL)
		return TMPI_SUCCESS;

	int length = impl::get_length(count, type);

	int res = pt2pt::irecv(buf, length, src, tag, comm->get_id(), comm, req);
	LSB_RECORD(TMPI_IRECV_DONE);

	return res;

}

inline int TMPI_Isend(void* buf, int count, TMPI_Datatype type, int dest, int tag,
		TMPI_Comm comm, TMPI_Request* req) {
	LSB_RESET();

	TMPI_LOG_LEVEL(-1, "TMPI_Isend", "Starting");
	using namespace tinympi;

	if (dest == TMPI_PROC_NULL)
		return TMPI_SUCCESS;

	int length = impl::get_length(count, type);

	int res = pt2pt::isend(buf, length, dest, tag, comm->get_id(), comm, req);
	LSB_RECORD(TMPI_ISEND_DONE);

	return res;

}

inline int TMPI_Recv(void* buf, int count, TMPI_Datatype type, int src, int tag,
		TMPI_Comm comm, TMPI_Status* status) {
	LSB_RESET();

	TMPI_LOG_LEVEL(-1, "TMPI_Recv", "Starting");
	using namespace tinympi;

	if (src == TMPI_PROC_NULL)
		return TMPI_SUCCESS;

	int length = impl::get_length(count, type);

	int res = pt2pt::recv(buf, length, src, tag, comm->get_id(), comm, status);
	LSB_RECORD(TMPI_RECV_DONE);

	return res;
}
inline int TMPI_Send(void* buf, int count, TMPI_Datatype type, int dest, int tag,
		TMPI_Comm comm) {
	LSB_RESET();
	TMPI_LOG_LEVEL(-1, "TMPI_Send", "Starting");
	using namespace tinympi;

	if (dest == TMPI_PROC_NULL)
		return TMPI_SUCCESS;

	int length = impl::get_length(count, type);

	int res = pt2pt::send(buf, length, dest, tag, comm->get_id(), comm);
	LSB_RECORD(TMPI_SEND_DONE);
	return res;
}
inline int TMPI_ALWAYSINLINE TMPI_Sendrecv(void* sendbuf, int sendcount, TMPI_Datatype sendtype,
		int dest, int sendtag, void* recvbuf, int recvcount, TMPI_Datatype recvtype, int src,
		int recvtag, TMPI_Comm comm, TMPI_Status* status) {
	LSB_RESET();

	TMPI_LOG_LEVEL(-1, "TMPI_Sendrecv", "Starting");
	using namespace tinympi;

	int sendlength = impl::get_length(sendcount, sendtype);
	int recvlength = impl::get_length(recvcount, recvtype);

	// forward to point2point module
	int res = pt2pt::sendrecv(sendbuf, sendlength, dest, sendtag, recvbuf, recvlength, src, recvtag,
			comm->get_id(), comm, status);
	LSB_RECORD(TMPI_SENDRECV_DONE);
	return res;
}
inline int TMPI_Get_count(const TMPI_Status* status, TMPI_Datatype datatype, int* count) {
	LSB_RESET();
	using namespace tinympi;
	using namespace tinympi::impl;
	TMPI_LOG_LEVEL(-1, "TMPI_Get_count", "Starting");
	int size = status->size;
	int size_one = get_length(1, datatype);

	TMPI_ASSERT_FATAL(size % size_one == 0, "datatype mismatch", __FILE__, __LINE__);

	*count = size / size_one;
	LSB_RECORD(TMPI_GETCOUNT_DONE);
	return TMPI_SUCCESS;
}

inline int TMPI_Bcast(void* buffer, int count, TMPI_Datatype datatype, int root, TMPI_Comm comm) {
	TMPI_LOG_LEVEL(-1, "TMPI_Bcast", "Starting");
	using namespace tinympi;
	using namespace tinympi::impl;

	int length = get_length(count, datatype);

	int res = collectives::bcast(buffer, length, root, comm);
	LSB_RECORD(TMPI_BCAST_DONE);
	return res;
}

inline int TMPI_Waitall(int count, TMPI_Request requests[], TMPI_Status statuses[]) {
	using namespace tinympi;
	LSB_RESET();

	int res = pt2pt::waitall(count, requests, statuses);
	LSB_RECORD(TMPI_WAITALL_DONE);
	return res;
}
inline int TMPI_Wait(TMPI_Request* request, TMPI_Status* status) {
	LSB_RESET();
	using namespace tinympi;

	int res = pt2pt::wait(request, status);
	LSB_RECORD(TMPI_WAIT_DONE);
	return res;
}

inline int TMPI_Allreduce(void* sendbuf, void *recvbuf, int count, TMPI_Datatype type,
		TMPI_Op op, const TMPI_Comm comm) {
	LSB_RESET();
	TMPI_LOG_LEVEL(-1, "TMPI_AllReduce", "Starting");
	using namespace tinympi;
	TMPI_ASSERT_FATAL(sendbuf != recvbuf, "send and recv buffer should not overlap", __FILE__,
			__LINE__);

	int length = impl::get_length(count, type);
	collectives::reduction_function_t reduction_function = collectives::get_reduction_function(op,
			type, count);

	int res = collectives::allreduce(sendbuf, recvbuf, length, reduction_function, comm);
	LSB_RECORD(TMPI_ALLREDUCE_DONE);
	return res;

}

inline int TMPI_Barrier(TMPI_Comm comm) {
	TMPI_LOG_LEVEL(-1, "TMPI_Barrier", "Starting");
	LSB_RESET();
	using namespace tinympi;

	int res = collectives::barrier(comm);
	LSB_RECORD(TMPI_BARRIER_DONE);
	return res;
}

inline int TMPI_Request_get_status(TMPI_Request& req, int* flag, TMPI_Status* status) {
	TMPI_LOG_LEVEL(-1, "TMPI_Request_get_status", "Starting");
	LSB_RESET();
	using namespace tinympi;

	*flag = fill_status(&req, status);
	return TMPI_SUCCESS;
}

#endif //_TMPI_H_
